/* SPDX-FileCopyrightText: © 2024-2025 Decompollaborate */
/* SPDX-License-Identifier: MIT */

#[cfg(feature = "R4000ALLEGREX")]
mod common;

#[cfg(feature = "R4000ALLEGREX")]
use common::{check_test_entries, TestEntry};

#[cfg(feature = "R4000ALLEGREX")]
use rabbitizer::{
    abi::Abi,
    display_flags::InstructionDisplayFlags,
    instr::{Instruction, InstructionFlags},
    isa::IsaExtension,
    opcodes::Opcode,
    vram::Vram,
};

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_instructions() {
    const ENTRIES: &[TestEntry] = &[
        // todo
        // 0x00041182 -> "srl         $v0, $a0, 6"

        TestEntry {
            instr: Instruction::new(
                0x00801017,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "clo         $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_clo,
            opcode_str: "clo",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00801016,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "clz         $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_clz,
            opcode_str: "clz",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00C7001C,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "madd        $a2, $a3",
            expected_opcode: Opcode::r4000allegrex_madd,
            opcode_str: "madd",
            operands_str: [Some("$a2"), Some("$a3"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00C7001D,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "maddu       $a2, $a3",
            expected_opcode: Opcode::r4000allegrex_maddu,
            opcode_str: "maddu",
            operands_str: [Some("$a2"), Some("$a3"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00C7002E,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "msub        $a2, $a3",
            expected_opcode: Opcode::r4000allegrex_msub,
            opcode_str: "msub",
            operands_str: [Some("$a2"), Some("$a3"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00C7002F,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "msubu       $a2, $a3",
            expected_opcode: Opcode::r4000allegrex_msubu,
            opcode_str: "msubu",
            operands_str: [Some("$a2"), Some("$a3"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x0085102C,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "max         $v0, $a0, $a1",
            expected_opcode: Opcode::r4000allegrex_max,
            opcode_str: "max",
            operands_str: [Some("$v0"), Some("$a0"), Some("$a1"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x0085102D,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "min         $v0, $a0, $a1",
            expected_opcode: Opcode::r4000allegrex_min,
            opcode_str: "min",
            operands_str: [Some("$v0"), Some("$a0"), Some("$a1"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x0085100B,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "movn        $v0, $a0, $a1",
            expected_opcode: Opcode::r4000allegrex_movn,
            opcode_str: "movn",
            operands_str: [Some("$v0"), Some("$a0"), Some("$a1"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x0085100A,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "movz        $v0, $a0, $a1",
            expected_opcode: Opcode::r4000allegrex_movz,
            opcode_str: "movz",
            operands_str: [Some("$v0"), Some("$a0"), Some("$a1"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C822080,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "ext         $v0, $a0, 2, 5",
            expected_opcode: Opcode::r4000allegrex_ext,
            opcode_str: "ext",
            operands_str: [Some("$v0"), Some("$a0"), Some("2"), Some("5"), None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C8221C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "ext         $v0, $a0, 7, 5",
            expected_opcode: Opcode::r4000allegrex_ext,
            opcode_str: "ext",
            operands_str: [Some("$v0"), Some("$a0"), Some("7"), Some("5"), None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C823084,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "ins         $v0, $a0, 2, 5",
            expected_opcode: Opcode::r4000allegrex_ins,
            opcode_str: "ins",
            operands_str: [Some("$v0"), Some("$a0"), Some("2"), Some("5"), None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C8259C4,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "ins         $v0, $a0, 7, 5",
            expected_opcode: Opcode::r4000allegrex_ins,
            opcode_str: "ins",
            operands_str: [Some("$v0"), Some("$a0"), Some("7"), Some("5"), None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C041420,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "seb         $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_seb,
            opcode_str: "seb",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C041620,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "seh         $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_seh,
            opcode_str: "seh",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C041520,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "bitrev      $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_bitrev,
            opcode_str: "bitrev",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00241182,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "rotr        $v0, $a0, 6",
            expected_opcode: Opcode::r4000allegrex_rotr,
            opcode_str: "rotr",
            operands_str: [Some("$v0"), Some("$a0"), Some("6"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x002414C2,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "rotr        $v0, $a0, 19",
            expected_opcode: Opcode::r4000allegrex_rotr,
            opcode_str: "rotr",
            operands_str: [Some("$v0"), Some("$a0"), Some("19"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x00A41046,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "rotrv       $v0, $a0, $a1",
            expected_opcode: Opcode::r4000allegrex_rotrv,
            opcode_str: "rotrv",
            operands_str: [Some("$v0"), Some("$a0"), Some("$a1"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C0410A0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "wsbh        $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_wsbh,
            opcode_str: "wsbh",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x7C0410E0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "wsbw        $v0, $a0",
            expected_opcode: Opcode::r4000allegrex_wsbw,
            opcode_str: "wsbw",
            operands_str: [Some("$v0"), Some("$a0"), None, None, None],
        },

        TestEntry {
            instr: Instruction::new(
                0xBC840000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x04, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x04"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC860000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x06, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x06"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC880000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x08, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x08"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC8A0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x0A, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x0A"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC8B0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x0B, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x0B"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC940000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x14, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x14"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC960000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x16, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x16"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC980000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x18, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x18"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC990000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x19, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x19"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC9A0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x1A, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x1A"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC9B0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x1B, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x1B"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC9C0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x1C, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x1C"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC9E0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x1E, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x1E"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC9F0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x1F, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x1F"), Some("0x0($a0)"), None, None, None],
        },

        TestEntry {
            instr: Instruction::new(
                0x0000000F,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "sync",
            expected_opcode: Opcode::core_sync,
            opcode_str: "sync",
            operands_str: [Some(""), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xC0820000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "ll          $v0, 0x0($a0)",
            expected_opcode: Opcode::core_ll,
            opcode_str: "ll",
            operands_str: [Some("$v0"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xE0850000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "sc          $a1, 0x0($a0)",
            expected_opcode: Opcode::core_sc,
            opcode_str: "sc",
            operands_str: [Some("$a1"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC800000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x00, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x00"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC810000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x01, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x01"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC820000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x02, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x02"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC830000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x03, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x03"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC850000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x05, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x05"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC870000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x07, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x07"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC890000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x09, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x09"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC8C0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x0C, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x0C"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC8D0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x0D, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x0D"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC8E0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x0E, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x0E"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC8F0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x0F, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x0F"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC900000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x10, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x10"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC910000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x11, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x11"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC920000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x12, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x12"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC930000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x13, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x13"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC950000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x15, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x15"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC970000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x17, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x17"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xBC9D0000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "cache       0x1D, 0x0($a0)",
            expected_opcode: Opcode::core_cache,
            opcode_str: "cache",
            operands_str: [Some("0x1D"), Some("0x0($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x70000000,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "sleep",
            expected_opcode: Opcode::r4000allegrex_sleep,
            opcode_str: "sleep",
            operands_str: [Some(""), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x70020024,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "mfie        $v0",
            expected_opcode: Opcode::r4000allegrex_mfie,
            opcode_str: "mfie",
            operands_str: [Some("$v0"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x70000024,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "mfie        $zero",
            expected_opcode: Opcode::r4000allegrex_mfie,
            opcode_str: "mfie",
            operands_str: [Some("$zero"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x70000026,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "mtie        $zero",
            expected_opcode: Opcode::r4000allegrex_mtie,
            opcode_str: "mtie",
            operands_str: [Some("$zero"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x70040026,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "mtie        $a0",
            expected_opcode: Opcode::r4000allegrex_mtie,
            opcode_str: "mtie",
            operands_str: [Some("$a0"), None, None, None, None],
        },

        TestEntry {
            instr: Instruction::new(
                0x46002085,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "abs.s       $ft0, $ft2",
            expected_opcode: Opcode::core_abs_s,
            opcode_str: "abs.s",
            operands_str: [Some("$ft0"), Some("$ft2"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x46042080,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "add.s       $ft0, $ft2, $ft2",
            expected_opcode: Opcode::core_add_s,
            opcode_str: "add.s",
            operands_str: [Some("$ft0"), Some("$ft2"), Some("$ft2"), None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x45000008,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "bc1f        . + 4 + (0x8 << 2)",
            expected_opcode: Opcode::core_bc1f,
            opcode_str: "bc1f",
            operands_str: [Some(". + 4 + (0x8 << 2)"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x46100030,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "c.f.s       $fv0, $fa4",
            expected_opcode: Opcode::core_c_f_s,
            opcode_str: "c.f.s",
            operands_str: [Some("$fv0"), Some("$fa4"), None, None, None],
        },

        // Allegrex removed 64 bits FPU instructions
        TestEntry {
            instr: Instruction::new(
                0x468020A1,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x468020A1                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUW> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x46002088,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x46002088                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUS> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x46002089,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x46002089                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUS> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x4600208A,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x4600208A                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUS> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x4600208B,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x4600208B                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUS> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x460020A1,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x460020A1                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUS> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x460020A5,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x460020A5                   /* INVALID / 00002080 <OpcodeCategory: R4000ALLEGREX_COP1_FPUS> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vector_zero() {
    const ENTRIES: &[TestEntry] = &[
        /* vector_zero */
        TestEntry::new_r4000allegrex(
            0xD0060000,
            "vzero.s     S000",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060001,
            "vzero.s     S010",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060002,
            "vzero.s     S020",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060003,
            "vzero.s     S030",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060004,
            "vzero.s     S100",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060005,
            "vzero.s     S110",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060006,
            "vzero.s     S120",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060007,
            "vzero.s     S130",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060008,
            "vzero.s     S200",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060009,
            "vzero.s     S210",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006000A,
            "vzero.s     S220",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006000B,
            "vzero.s     S230",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006000C,
            "vzero.s     S300",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006000D,
            "vzero.s     S310",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006000E,
            "vzero.s     S320",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006000F,
            "vzero.s     S330",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060010,
            "vzero.s     S400",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060011,
            "vzero.s     S410",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060012,
            "vzero.s     S420",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060013,
            "vzero.s     S430",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060014,
            "vzero.s     S500",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060015,
            "vzero.s     S510",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060016,
            "vzero.s     S520",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060017,
            "vzero.s     S530",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060018,
            "vzero.s     S600",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060019,
            "vzero.s     S610",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006001A,
            "vzero.s     S620",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006001B,
            "vzero.s     S630",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006001C,
            "vzero.s     S700",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006001D,
            "vzero.s     S710",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006001E,
            "vzero.s     S720",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006001F,
            "vzero.s     S730",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060020,
            "vzero.s     S001",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060021,
            "vzero.s     S011",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S011"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060022,
            "vzero.s     S021",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060023,
            "vzero.s     S031",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S031"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060024,
            "vzero.s     S101",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060025,
            "vzero.s     S111",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S111"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060026,
            "vzero.s     S121",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060027,
            "vzero.s     S131",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S131"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060028,
            "vzero.s     S201",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060029,
            "vzero.s     S211",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S211"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006002A,
            "vzero.s     S221",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006002B,
            "vzero.s     S231",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S231"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006002C,
            "vzero.s     S301",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006002D,
            "vzero.s     S311",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S311"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006002E,
            "vzero.s     S321",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006002F,
            "vzero.s     S331",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S331"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060030,
            "vzero.s     S401",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060031,
            "vzero.s     S411",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S411"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060032,
            "vzero.s     S421",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060033,
            "vzero.s     S431",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S431"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060034,
            "vzero.s     S501",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060035,
            "vzero.s     S511",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S511"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060036,
            "vzero.s     S521",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060037,
            "vzero.s     S531",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S531"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060038,
            "vzero.s     S601",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060039,
            "vzero.s     S611",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S611"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006003A,
            "vzero.s     S621",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006003B,
            "vzero.s     S631",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S631"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006003C,
            "vzero.s     S701",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006003D,
            "vzero.s     S711",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S711"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006003E,
            "vzero.s     S721",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006003F,
            "vzero.s     S731",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S731"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060040,
            "vzero.s     S002",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060041,
            "vzero.s     S012",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060042,
            "vzero.s     S022",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060043,
            "vzero.s     S032",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S032"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060044,
            "vzero.s     S102",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060045,
            "vzero.s     S112",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060046,
            "vzero.s     S122",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060047,
            "vzero.s     S132",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S132"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060048,
            "vzero.s     S202",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060049,
            "vzero.s     S212",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006004A,
            "vzero.s     S222",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006004B,
            "vzero.s     S232",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S232"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006004C,
            "vzero.s     S302",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006004D,
            "vzero.s     S312",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006004E,
            "vzero.s     S322",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006004F,
            "vzero.s     S332",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S332"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060050,
            "vzero.s     S402",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060051,
            "vzero.s     S412",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060052,
            "vzero.s     S422",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060053,
            "vzero.s     S432",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S432"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060054,
            "vzero.s     S502",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060055,
            "vzero.s     S512",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060056,
            "vzero.s     S522",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060057,
            "vzero.s     S532",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S532"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060058,
            "vzero.s     S602",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060059,
            "vzero.s     S612",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006005A,
            "vzero.s     S622",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006005B,
            "vzero.s     S632",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S632"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006005C,
            "vzero.s     S702",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006005D,
            "vzero.s     S712",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006005E,
            "vzero.s     S722",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006005F,
            "vzero.s     S732",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S732"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060060,
            "vzero.s     S003",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060061,
            "vzero.s     S013",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S013"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060062,
            "vzero.s     S023",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S023"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060063,
            "vzero.s     S033",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S033"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060064,
            "vzero.s     S103",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060065,
            "vzero.s     S113",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S113"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060066,
            "vzero.s     S123",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S123"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060067,
            "vzero.s     S133",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S133"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060068,
            "vzero.s     S203",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060069,
            "vzero.s     S213",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S213"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006006A,
            "vzero.s     S223",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S223"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006006B,
            "vzero.s     S233",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S233"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006006C,
            "vzero.s     S303",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006006D,
            "vzero.s     S313",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S313"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006006E,
            "vzero.s     S323",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S323"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006006F,
            "vzero.s     S333",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S333"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060070,
            "vzero.s     S403",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060071,
            "vzero.s     S413",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S413"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060072,
            "vzero.s     S423",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S423"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060073,
            "vzero.s     S433",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S433"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060074,
            "vzero.s     S503",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060075,
            "vzero.s     S513",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S513"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060076,
            "vzero.s     S523",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S523"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060077,
            "vzero.s     S533",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S533"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060078,
            "vzero.s     S603",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060079,
            "vzero.s     S613",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S613"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006007A,
            "vzero.s     S623",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S623"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006007B,
            "vzero.s     S633",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S633"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006007C,
            "vzero.s     S703",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006007D,
            "vzero.s     S713",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S713"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006007E,
            "vzero.s     S723",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S723"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006007F,
            "vzero.s     S733",
            Opcode::r4000allegrex_vzero_s,
            "vzero.s",
            [Some("S733"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060080,
            "vzero.p     C000",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060081,
            "vzero.p     C010",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060082,
            "vzero.p     C020",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060083,
            "vzero.p     C030",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060084,
            "vzero.p     C100",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060085,
            "vzero.p     C110",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060086,
            "vzero.p     C120",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060087,
            "vzero.p     C130",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060088,
            "vzero.p     C200",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060089,
            "vzero.p     C210",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006008A,
            "vzero.p     C220",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006008B,
            "vzero.p     C230",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006008C,
            "vzero.p     C300",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006008D,
            "vzero.p     C310",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006008E,
            "vzero.p     C320",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006008F,
            "vzero.p     C330",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060090,
            "vzero.p     C400",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060091,
            "vzero.p     C410",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060092,
            "vzero.p     C420",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060093,
            "vzero.p     C430",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060094,
            "vzero.p     C500",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060095,
            "vzero.p     C510",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060096,
            "vzero.p     C520",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060097,
            "vzero.p     C530",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060098,
            "vzero.p     C600",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0060099,
            "vzero.p     C610",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006009A,
            "vzero.p     C620",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006009B,
            "vzero.p     C630",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006009C,
            "vzero.p     C700",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006009D,
            "vzero.p     C710",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006009E,
            "vzero.p     C720",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006009F,
            "vzero.p     C730",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A0,
            "vzero.p     R000",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A1,
            "vzero.p     R001",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A2,
            "vzero.p     R002",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A3,
            "vzero.p     R003",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A4,
            "vzero.p     R100",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A5,
            "vzero.p     R101",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A6,
            "vzero.p     R102",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A7,
            "vzero.p     R103",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A8,
            "vzero.p     R200",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600A9,
            "vzero.p     R201",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600AA,
            "vzero.p     R202",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600AB,
            "vzero.p     R203",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600AC,
            "vzero.p     R300",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600AD,
            "vzero.p     R301",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600AE,
            "vzero.p     R302",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600AF,
            "vzero.p     R303",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B0,
            "vzero.p     R400",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B1,
            "vzero.p     R401",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B2,
            "vzero.p     R402",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B3,
            "vzero.p     R403",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B4,
            "vzero.p     R500",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B5,
            "vzero.p     R501",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B6,
            "vzero.p     R502",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B7,
            "vzero.p     R503",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B8,
            "vzero.p     R600",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600B9,
            "vzero.p     R601",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600BA,
            "vzero.p     R602",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600BB,
            "vzero.p     R603",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600BC,
            "vzero.p     R700",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600BD,
            "vzero.p     R701",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600BE,
            "vzero.p     R702",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600BF,
            "vzero.p     R703",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C0,
            "vzero.p     C002",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C1,
            "vzero.p     C012",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C2,
            "vzero.p     C022",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C3,
            "vzero.p     C032",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C032"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C4,
            "vzero.p     C102",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C5,
            "vzero.p     C112",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C6,
            "vzero.p     C122",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C7,
            "vzero.p     C132",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C132"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C8,
            "vzero.p     C202",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600C9,
            "vzero.p     C212",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600CA,
            "vzero.p     C222",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600CB,
            "vzero.p     C232",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C232"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600CC,
            "vzero.p     C302",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600CD,
            "vzero.p     C312",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600CE,
            "vzero.p     C322",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600CF,
            "vzero.p     C332",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C332"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D0,
            "vzero.p     C402",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D1,
            "vzero.p     C412",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D2,
            "vzero.p     C422",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D3,
            "vzero.p     C432",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C432"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D4,
            "vzero.p     C502",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D5,
            "vzero.p     C512",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D6,
            "vzero.p     C522",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D7,
            "vzero.p     C532",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C532"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D8,
            "vzero.p     C602",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600D9,
            "vzero.p     C612",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600DA,
            "vzero.p     C622",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600DB,
            "vzero.p     C632",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C632"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600DC,
            "vzero.p     C702",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600DD,
            "vzero.p     C712",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600DE,
            "vzero.p     C722",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600DF,
            "vzero.p     C732",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("C732"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E0,
            "vzero.p     R020",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E1,
            "vzero.p     R021",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E2,
            "vzero.p     R022",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E3,
            "vzero.p     R023",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R023"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E4,
            "vzero.p     R120",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E5,
            "vzero.p     R121",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E6,
            "vzero.p     R122",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E7,
            "vzero.p     R123",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R123"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E8,
            "vzero.p     R220",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600E9,
            "vzero.p     R221",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600EA,
            "vzero.p     R222",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600EB,
            "vzero.p     R223",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R223"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600EC,
            "vzero.p     R320",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600ED,
            "vzero.p     R321",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600EE,
            "vzero.p     R322",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600EF,
            "vzero.p     R323",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R323"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F0,
            "vzero.p     R420",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F1,
            "vzero.p     R421",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F2,
            "vzero.p     R422",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F3,
            "vzero.p     R423",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R423"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F4,
            "vzero.p     R520",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F5,
            "vzero.p     R521",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F6,
            "vzero.p     R522",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F7,
            "vzero.p     R523",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R523"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F8,
            "vzero.p     R620",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600F9,
            "vzero.p     R621",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600FA,
            "vzero.p     R622",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600FB,
            "vzero.p     R623",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R623"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600FC,
            "vzero.p     R720",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600FD,
            "vzero.p     R721",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600FE,
            "vzero.p     R722",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00600FF,
            "vzero.p     R723",
            Opcode::r4000allegrex_vzero_p,
            "vzero.p",
            [Some("R723"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068000,
            "vzero.t     C000",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068001,
            "vzero.t     C010",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068002,
            "vzero.t     C020",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068003,
            "vzero.t     C030",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068004,
            "vzero.t     C100",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068005,
            "vzero.t     C110",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068006,
            "vzero.t     C120",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068007,
            "vzero.t     C130",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068008,
            "vzero.t     C200",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068009,
            "vzero.t     C210",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006800A,
            "vzero.t     C220",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006800B,
            "vzero.t     C230",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006800C,
            "vzero.t     C300",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006800D,
            "vzero.t     C310",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006800E,
            "vzero.t     C320",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006800F,
            "vzero.t     C330",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068010,
            "vzero.t     C400",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068011,
            "vzero.t     C410",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068012,
            "vzero.t     C420",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068013,
            "vzero.t     C430",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068014,
            "vzero.t     C500",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068015,
            "vzero.t     C510",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068016,
            "vzero.t     C520",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068017,
            "vzero.t     C530",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068018,
            "vzero.t     C600",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068019,
            "vzero.t     C610",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006801A,
            "vzero.t     C620",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006801B,
            "vzero.t     C630",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006801C,
            "vzero.t     C700",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006801D,
            "vzero.t     C710",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006801E,
            "vzero.t     C720",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006801F,
            "vzero.t     C730",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068020,
            "vzero.t     R000",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068021,
            "vzero.t     R001",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068022,
            "vzero.t     R002",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068023,
            "vzero.t     R003",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068024,
            "vzero.t     R100",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068025,
            "vzero.t     R101",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068026,
            "vzero.t     R102",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068027,
            "vzero.t     R103",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068028,
            "vzero.t     R200",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068029,
            "vzero.t     R201",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006802A,
            "vzero.t     R202",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006802B,
            "vzero.t     R203",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006802C,
            "vzero.t     R300",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006802D,
            "vzero.t     R301",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006802E,
            "vzero.t     R302",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006802F,
            "vzero.t     R303",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068030,
            "vzero.t     R400",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068031,
            "vzero.t     R401",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068032,
            "vzero.t     R402",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068033,
            "vzero.t     R403",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068034,
            "vzero.t     R500",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068035,
            "vzero.t     R501",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068036,
            "vzero.t     R502",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068037,
            "vzero.t     R503",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068038,
            "vzero.t     R600",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068039,
            "vzero.t     R601",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006803A,
            "vzero.t     R602",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006803B,
            "vzero.t     R603",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006803C,
            "vzero.t     R700",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006803D,
            "vzero.t     R701",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006803E,
            "vzero.t     R702",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006803F,
            "vzero.t     R703",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068040,
            "vzero.t     C001",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068041,
            "vzero.t     C011",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C011"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068042,
            "vzero.t     C021",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068043,
            "vzero.t     C031",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C031"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068044,
            "vzero.t     C101",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068045,
            "vzero.t     C111",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C111"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068046,
            "vzero.t     C121",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068047,
            "vzero.t     C131",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C131"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068048,
            "vzero.t     C201",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068049,
            "vzero.t     C211",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C211"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006804A,
            "vzero.t     C221",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006804B,
            "vzero.t     C231",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C231"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006804C,
            "vzero.t     C301",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006804D,
            "vzero.t     C311",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C311"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006804E,
            "vzero.t     C321",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006804F,
            "vzero.t     C331",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C331"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068050,
            "vzero.t     C401",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068051,
            "vzero.t     C411",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C411"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068052,
            "vzero.t     C421",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068053,
            "vzero.t     C431",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C431"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068054,
            "vzero.t     C501",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068055,
            "vzero.t     C511",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C511"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068056,
            "vzero.t     C521",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068057,
            "vzero.t     C531",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C531"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068058,
            "vzero.t     C601",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068059,
            "vzero.t     C611",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C611"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006805A,
            "vzero.t     C621",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006805B,
            "vzero.t     C631",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C631"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006805C,
            "vzero.t     C701",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006805D,
            "vzero.t     C711",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C711"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006805E,
            "vzero.t     C721",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006805F,
            "vzero.t     C731",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("C731"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068060,
            "vzero.t     R010",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068061,
            "vzero.t     R011",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R011"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068062,
            "vzero.t     R012",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068063,
            "vzero.t     R013",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R013"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068064,
            "vzero.t     R110",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068065,
            "vzero.t     R111",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R111"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068066,
            "vzero.t     R112",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068067,
            "vzero.t     R113",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R113"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068068,
            "vzero.t     R210",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068069,
            "vzero.t     R211",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R211"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006806A,
            "vzero.t     R212",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006806B,
            "vzero.t     R213",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R213"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006806C,
            "vzero.t     R310",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006806D,
            "vzero.t     R311",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R311"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006806E,
            "vzero.t     R312",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006806F,
            "vzero.t     R313",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R313"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068070,
            "vzero.t     R410",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068071,
            "vzero.t     R411",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R411"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068072,
            "vzero.t     R412",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068073,
            "vzero.t     R413",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R413"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068074,
            "vzero.t     R510",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068075,
            "vzero.t     R511",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R511"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068076,
            "vzero.t     R512",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068077,
            "vzero.t     R513",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R513"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068078,
            "vzero.t     R610",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068079,
            "vzero.t     R611",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R611"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006807A,
            "vzero.t     R612",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006807B,
            "vzero.t     R613",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R613"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006807C,
            "vzero.t     R710",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006807D,
            "vzero.t     R711",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R711"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006807E,
            "vzero.t     R712",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006807F,
            "vzero.t     R713",
            Opcode::r4000allegrex_vzero_t,
            "vzero.t",
            [Some("R713"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068080,
            "vzero.q     C000",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068081,
            "vzero.q     C010",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068082,
            "vzero.q     C020",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068083,
            "vzero.q     C030",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068084,
            "vzero.q     C100",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068085,
            "vzero.q     C110",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068086,
            "vzero.q     C120",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068087,
            "vzero.q     C130",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068088,
            "vzero.q     C200",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068089,
            "vzero.q     C210",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006808A,
            "vzero.q     C220",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006808B,
            "vzero.q     C230",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006808C,
            "vzero.q     C300",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006808D,
            "vzero.q     C310",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006808E,
            "vzero.q     C320",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006808F,
            "vzero.q     C330",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068090,
            "vzero.q     C400",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068091,
            "vzero.q     C410",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068092,
            "vzero.q     C420",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068093,
            "vzero.q     C430",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068094,
            "vzero.q     C500",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068095,
            "vzero.q     C510",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068096,
            "vzero.q     C520",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068097,
            "vzero.q     C530",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068098,
            "vzero.q     C600",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0068099,
            "vzero.q     C610",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006809A,
            "vzero.q     C620",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006809B,
            "vzero.q     C630",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006809C,
            "vzero.q     C700",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006809D,
            "vzero.q     C710",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006809E,
            "vzero.q     C720",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD006809F,
            "vzero.q     C730",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A0,
            "vzero.q     R000",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A1,
            "vzero.q     R001",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A2,
            "vzero.q     R002",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A3,
            "vzero.q     R003",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A4,
            "vzero.q     R100",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A5,
            "vzero.q     R101",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A6,
            "vzero.q     R102",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A7,
            "vzero.q     R103",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A8,
            "vzero.q     R200",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680A9,
            "vzero.q     R201",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680AA,
            "vzero.q     R202",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680AB,
            "vzero.q     R203",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680AC,
            "vzero.q     R300",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680AD,
            "vzero.q     R301",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680AE,
            "vzero.q     R302",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680AF,
            "vzero.q     R303",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B0,
            "vzero.q     R400",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B1,
            "vzero.q     R401",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B2,
            "vzero.q     R402",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B3,
            "vzero.q     R403",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B4,
            "vzero.q     R500",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B5,
            "vzero.q     R501",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B6,
            "vzero.q     R502",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B7,
            "vzero.q     R503",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B8,
            "vzero.q     R600",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680B9,
            "vzero.q     R601",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680BA,
            "vzero.q     R602",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680BB,
            "vzero.q     R603",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680BC,
            "vzero.q     R700",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680BD,
            "vzero.q     R701",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680BE,
            "vzero.q     R702",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680BF,
            "vzero.q     R703",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C0,
            "vzero.q     C002",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C1,
            "vzero.q     C012",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C2,
            "vzero.q     C022",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C3,
            "vzero.q     C032",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C032"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C4,
            "vzero.q     C102",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C5,
            "vzero.q     C112",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C6,
            "vzero.q     C122",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C7,
            "vzero.q     C132",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C132"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C8,
            "vzero.q     C202",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680C9,
            "vzero.q     C212",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680CA,
            "vzero.q     C222",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680CB,
            "vzero.q     C232",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C232"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680CC,
            "vzero.q     C302",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680CD,
            "vzero.q     C312",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680CE,
            "vzero.q     C322",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680CF,
            "vzero.q     C332",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C332"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D0,
            "vzero.q     C402",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D1,
            "vzero.q     C412",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D2,
            "vzero.q     C422",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D3,
            "vzero.q     C432",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C432"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D4,
            "vzero.q     C502",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D5,
            "vzero.q     C512",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D6,
            "vzero.q     C522",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D7,
            "vzero.q     C532",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C532"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D8,
            "vzero.q     C602",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680D9,
            "vzero.q     C612",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680DA,
            "vzero.q     C622",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680DB,
            "vzero.q     C632",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C632"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680DC,
            "vzero.q     C702",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680DD,
            "vzero.q     C712",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680DE,
            "vzero.q     C722",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680DF,
            "vzero.q     C732",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("C732"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E0,
            "vzero.q     R020",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E1,
            "vzero.q     R021",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E2,
            "vzero.q     R022",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E3,
            "vzero.q     R023",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R023"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E4,
            "vzero.q     R120",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E5,
            "vzero.q     R121",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E6,
            "vzero.q     R122",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E7,
            "vzero.q     R123",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R123"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E8,
            "vzero.q     R220",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680E9,
            "vzero.q     R221",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680EA,
            "vzero.q     R222",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680EB,
            "vzero.q     R223",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R223"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680EC,
            "vzero.q     R320",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680ED,
            "vzero.q     R321",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680EE,
            "vzero.q     R322",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680EF,
            "vzero.q     R323",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R323"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F0,
            "vzero.q     R420",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F1,
            "vzero.q     R421",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F2,
            "vzero.q     R422",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F3,
            "vzero.q     R423",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R423"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F4,
            "vzero.q     R520",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F5,
            "vzero.q     R521",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F6,
            "vzero.q     R522",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F7,
            "vzero.q     R523",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R523"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F8,
            "vzero.q     R620",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680F9,
            "vzero.q     R621",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680FA,
            "vzero.q     R622",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680FB,
            "vzero.q     R623",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R623"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680FC,
            "vzero.q     R720",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680FD,
            "vzero.q     R721",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680FE,
            "vzero.q     R722",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00680FF,
            "vzero.q     R723",
            Opcode::r4000allegrex_vzero_q,
            "vzero.q",
            [Some("R723"), None, None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_matrix_zero() {
    const ENTRIES: &[TestEntry] = &[
        /* matrix_zero */
        TestEntry::new_r4000allegrex(
            0xF3860080,
            "vmzero.p    M000",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860081,
            "vmzero.p    M010",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860082,
            "vmzero.p    M020",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860083,
            "vmzero.p    M030",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860084,
            "vmzero.p    M100",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860085,
            "vmzero.p    M110",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860086,
            "vmzero.p    M120",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860087,
            "vmzero.p    M130",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860088,
            "vmzero.p    M200",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860089,
            "vmzero.p    M210",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386008A,
            "vmzero.p    M220",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386008B,
            "vmzero.p    M230",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386008C,
            "vmzero.p    M300",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386008D,
            "vmzero.p    M310",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386008E,
            "vmzero.p    M320",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386008F,
            "vmzero.p    M330",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860090,
            "vmzero.p    M400",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860091,
            "vmzero.p    M410",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860092,
            "vmzero.p    M420",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860093,
            "vmzero.p    M430",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860094,
            "vmzero.p    M500",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860095,
            "vmzero.p    M510",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860096,
            "vmzero.p    M520",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860097,
            "vmzero.p    M530",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860098,
            "vmzero.p    M600",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3860099,
            "vmzero.p    M610",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386009A,
            "vmzero.p    M620",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386009B,
            "vmzero.p    M630",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386009C,
            "vmzero.p    M700",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386009D,
            "vmzero.p    M710",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386009E,
            "vmzero.p    M720",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386009F,
            "vmzero.p    M730",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A0,
            "vmzero.p    E000",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A1,
            "vmzero.p    E001",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A2,
            "vmzero.p    E002",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A3,
            "vmzero.p    E003",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A4,
            "vmzero.p    E100",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A5,
            "vmzero.p    E101",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A6,
            "vmzero.p    E102",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A7,
            "vmzero.p    E103",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A8,
            "vmzero.p    E200",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600A9,
            "vmzero.p    E201",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600AA,
            "vmzero.p    E202",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600AB,
            "vmzero.p    E203",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600AC,
            "vmzero.p    E300",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600AD,
            "vmzero.p    E301",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600AE,
            "vmzero.p    E302",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600AF,
            "vmzero.p    E303",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B0,
            "vmzero.p    E400",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B1,
            "vmzero.p    E401",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B2,
            "vmzero.p    E402",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B3,
            "vmzero.p    E403",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B4,
            "vmzero.p    E500",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B5,
            "vmzero.p    E501",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B6,
            "vmzero.p    E502",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B7,
            "vmzero.p    E503",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B8,
            "vmzero.p    E600",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600B9,
            "vmzero.p    E601",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600BA,
            "vmzero.p    E602",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600BB,
            "vmzero.p    E603",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600BC,
            "vmzero.p    E700",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600BD,
            "vmzero.p    E701",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600BE,
            "vmzero.p    E702",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600BF,
            "vmzero.p    E703",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C0,
            "vmzero.p    M002",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C1,
            "vmzero.p    M012",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C2,
            "vmzero.p    M022",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C3,
            "vmzero.p    M032",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M032"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C4,
            "vmzero.p    M102",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C5,
            "vmzero.p    M112",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C6,
            "vmzero.p    M122",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C7,
            "vmzero.p    M132",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M132"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C8,
            "vmzero.p    M202",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600C9,
            "vmzero.p    M212",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600CA,
            "vmzero.p    M222",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600CB,
            "vmzero.p    M232",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M232"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600CC,
            "vmzero.p    M302",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600CD,
            "vmzero.p    M312",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600CE,
            "vmzero.p    M322",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600CF,
            "vmzero.p    M332",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M332"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D0,
            "vmzero.p    M402",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D1,
            "vmzero.p    M412",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D2,
            "vmzero.p    M422",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D3,
            "vmzero.p    M432",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M432"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D4,
            "vmzero.p    M502",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D5,
            "vmzero.p    M512",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D6,
            "vmzero.p    M522",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D7,
            "vmzero.p    M532",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M532"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D8,
            "vmzero.p    M602",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600D9,
            "vmzero.p    M612",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600DA,
            "vmzero.p    M622",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600DB,
            "vmzero.p    M632",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M632"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600DC,
            "vmzero.p    M702",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600DD,
            "vmzero.p    M712",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600DE,
            "vmzero.p    M722",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600DF,
            "vmzero.p    M732",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("M732"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E0,
            "vmzero.p    E020",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E1,
            "vmzero.p    E021",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E2,
            "vmzero.p    E022",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E3,
            "vmzero.p    E023",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E023"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E4,
            "vmzero.p    E120",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E5,
            "vmzero.p    E121",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E6,
            "vmzero.p    E122",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E7,
            "vmzero.p    E123",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E123"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E8,
            "vmzero.p    E220",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600E9,
            "vmzero.p    E221",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600EA,
            "vmzero.p    E222",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600EB,
            "vmzero.p    E223",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E223"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600EC,
            "vmzero.p    E320",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600ED,
            "vmzero.p    E321",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600EE,
            "vmzero.p    E322",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600EF,
            "vmzero.p    E323",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E323"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F0,
            "vmzero.p    E420",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F1,
            "vmzero.p    E421",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F2,
            "vmzero.p    E422",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F3,
            "vmzero.p    E423",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E423"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F4,
            "vmzero.p    E520",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F5,
            "vmzero.p    E521",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F6,
            "vmzero.p    E522",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F7,
            "vmzero.p    E523",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E523"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F8,
            "vmzero.p    E620",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600F9,
            "vmzero.p    E621",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600FA,
            "vmzero.p    E622",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600FB,
            "vmzero.p    E623",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E623"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600FC,
            "vmzero.p    E720",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600FD,
            "vmzero.p    E721",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600FE,
            "vmzero.p    E722",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38600FF,
            "vmzero.p    E723",
            Opcode::r4000allegrex_vmzero_p,
            "vmzero.p",
            [Some("E723"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868000,
            "vmzero.t    M000",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868001,
            "vmzero.t    M010",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868002,
            "vmzero.t    M020",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868003,
            "vmzero.t    M030",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868004,
            "vmzero.t    M100",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868005,
            "vmzero.t    M110",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868006,
            "vmzero.t    M120",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868007,
            "vmzero.t    M130",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868008,
            "vmzero.t    M200",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868009,
            "vmzero.t    M210",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386800A,
            "vmzero.t    M220",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386800B,
            "vmzero.t    M230",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386800C,
            "vmzero.t    M300",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386800D,
            "vmzero.t    M310",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386800E,
            "vmzero.t    M320",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386800F,
            "vmzero.t    M330",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868010,
            "vmzero.t    M400",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868011,
            "vmzero.t    M410",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868012,
            "vmzero.t    M420",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868013,
            "vmzero.t    M430",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868014,
            "vmzero.t    M500",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868015,
            "vmzero.t    M510",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868016,
            "vmzero.t    M520",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868017,
            "vmzero.t    M530",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868018,
            "vmzero.t    M600",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868019,
            "vmzero.t    M610",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386801A,
            "vmzero.t    M620",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386801B,
            "vmzero.t    M630",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386801C,
            "vmzero.t    M700",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386801D,
            "vmzero.t    M710",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386801E,
            "vmzero.t    M720",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386801F,
            "vmzero.t    M730",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868020,
            "vmzero.t    E000",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868021,
            "vmzero.t    E001",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868022,
            "vmzero.t    E002",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868023,
            "vmzero.t    E003",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868024,
            "vmzero.t    E100",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868025,
            "vmzero.t    E101",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868026,
            "vmzero.t    E102",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868027,
            "vmzero.t    E103",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868028,
            "vmzero.t    E200",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868029,
            "vmzero.t    E201",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386802A,
            "vmzero.t    E202",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386802B,
            "vmzero.t    E203",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386802C,
            "vmzero.t    E300",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386802D,
            "vmzero.t    E301",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386802E,
            "vmzero.t    E302",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386802F,
            "vmzero.t    E303",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868030,
            "vmzero.t    E400",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868031,
            "vmzero.t    E401",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868032,
            "vmzero.t    E402",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868033,
            "vmzero.t    E403",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868034,
            "vmzero.t    E500",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868035,
            "vmzero.t    E501",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868036,
            "vmzero.t    E502",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868037,
            "vmzero.t    E503",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868038,
            "vmzero.t    E600",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868039,
            "vmzero.t    E601",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386803A,
            "vmzero.t    E602",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386803B,
            "vmzero.t    E603",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386803C,
            "vmzero.t    E700",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386803D,
            "vmzero.t    E701",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386803E,
            "vmzero.t    E702",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386803F,
            "vmzero.t    E703",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868040,
            "vmzero.t    M001",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868041,
            "vmzero.t    M011",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M011"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868042,
            "vmzero.t    M021",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868043,
            "vmzero.t    M031",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M031"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868044,
            "vmzero.t    M101",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868045,
            "vmzero.t    M111",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M111"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868046,
            "vmzero.t    M121",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868047,
            "vmzero.t    M131",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M131"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868048,
            "vmzero.t    M201",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868049,
            "vmzero.t    M211",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M211"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386804A,
            "vmzero.t    M221",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386804B,
            "vmzero.t    M231",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M231"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386804C,
            "vmzero.t    M301",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386804D,
            "vmzero.t    M311",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M311"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386804E,
            "vmzero.t    M321",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386804F,
            "vmzero.t    M331",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M331"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868050,
            "vmzero.t    M401",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868051,
            "vmzero.t    M411",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M411"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868052,
            "vmzero.t    M421",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868053,
            "vmzero.t    M431",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M431"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868054,
            "vmzero.t    M501",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868055,
            "vmzero.t    M511",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M511"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868056,
            "vmzero.t    M521",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868057,
            "vmzero.t    M531",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M531"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868058,
            "vmzero.t    M601",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868059,
            "vmzero.t    M611",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M611"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386805A,
            "vmzero.t    M621",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386805B,
            "vmzero.t    M631",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M631"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386805C,
            "vmzero.t    M701",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386805D,
            "vmzero.t    M711",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M711"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386805E,
            "vmzero.t    M721",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386805F,
            "vmzero.t    M731",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("M731"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868060,
            "vmzero.t    E010",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868061,
            "vmzero.t    E011",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E011"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868062,
            "vmzero.t    E012",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868063,
            "vmzero.t    E013",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E013"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868064,
            "vmzero.t    E110",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868065,
            "vmzero.t    E111",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E111"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868066,
            "vmzero.t    E112",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868067,
            "vmzero.t    E113",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E113"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868068,
            "vmzero.t    E210",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868069,
            "vmzero.t    E211",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E211"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386806A,
            "vmzero.t    E212",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386806B,
            "vmzero.t    E213",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E213"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386806C,
            "vmzero.t    E310",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386806D,
            "vmzero.t    E311",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E311"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386806E,
            "vmzero.t    E312",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386806F,
            "vmzero.t    E313",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E313"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868070,
            "vmzero.t    E410",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868071,
            "vmzero.t    E411",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E411"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868072,
            "vmzero.t    E412",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868073,
            "vmzero.t    E413",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E413"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868074,
            "vmzero.t    E510",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868075,
            "vmzero.t    E511",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E511"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868076,
            "vmzero.t    E512",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868077,
            "vmzero.t    E513",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E513"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868078,
            "vmzero.t    E610",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868079,
            "vmzero.t    E611",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E611"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386807A,
            "vmzero.t    E612",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386807B,
            "vmzero.t    E613",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E613"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386807C,
            "vmzero.t    E710",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386807D,
            "vmzero.t    E711",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E711"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386807E,
            "vmzero.t    E712",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386807F,
            "vmzero.t    E713",
            Opcode::r4000allegrex_vmzero_t,
            "vmzero.t",
            [Some("E713"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868080,
            "vmzero.q    M000",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868081,
            "vmzero.q    M010",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M010"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868082,
            "vmzero.q    M020",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868083,
            "vmzero.q    M030",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M030"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868084,
            "vmzero.q    M100",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868085,
            "vmzero.q    M110",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M110"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868086,
            "vmzero.q    M120",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868087,
            "vmzero.q    M130",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M130"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868088,
            "vmzero.q    M200",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868089,
            "vmzero.q    M210",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M210"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386808A,
            "vmzero.q    M220",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386808B,
            "vmzero.q    M230",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M230"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386808C,
            "vmzero.q    M300",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386808D,
            "vmzero.q    M310",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M310"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386808E,
            "vmzero.q    M320",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386808F,
            "vmzero.q    M330",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M330"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868090,
            "vmzero.q    M400",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868091,
            "vmzero.q    M410",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M410"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868092,
            "vmzero.q    M420",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868093,
            "vmzero.q    M430",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M430"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868094,
            "vmzero.q    M500",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868095,
            "vmzero.q    M510",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M510"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868096,
            "vmzero.q    M520",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868097,
            "vmzero.q    M530",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M530"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868098,
            "vmzero.q    M600",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3868099,
            "vmzero.q    M610",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M610"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386809A,
            "vmzero.q    M620",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386809B,
            "vmzero.q    M630",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M630"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386809C,
            "vmzero.q    M700",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386809D,
            "vmzero.q    M710",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M710"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386809E,
            "vmzero.q    M720",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF386809F,
            "vmzero.q    M730",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M730"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A0,
            "vmzero.q    E000",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A1,
            "vmzero.q    E001",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A2,
            "vmzero.q    E002",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A3,
            "vmzero.q    E003",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E003"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A4,
            "vmzero.q    E100",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E100"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A5,
            "vmzero.q    E101",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E101"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A6,
            "vmzero.q    E102",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A7,
            "vmzero.q    E103",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E103"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A8,
            "vmzero.q    E200",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E200"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680A9,
            "vmzero.q    E201",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E201"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680AA,
            "vmzero.q    E202",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680AB,
            "vmzero.q    E203",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E203"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680AC,
            "vmzero.q    E300",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E300"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680AD,
            "vmzero.q    E301",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E301"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680AE,
            "vmzero.q    E302",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680AF,
            "vmzero.q    E303",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E303"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B0,
            "vmzero.q    E400",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E400"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B1,
            "vmzero.q    E401",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E401"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B2,
            "vmzero.q    E402",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B3,
            "vmzero.q    E403",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E403"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B4,
            "vmzero.q    E500",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E500"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B5,
            "vmzero.q    E501",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E501"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B6,
            "vmzero.q    E502",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B7,
            "vmzero.q    E503",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E503"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B8,
            "vmzero.q    E600",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E600"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680B9,
            "vmzero.q    E601",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E601"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680BA,
            "vmzero.q    E602",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680BB,
            "vmzero.q    E603",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E603"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680BC,
            "vmzero.q    E700",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E700"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680BD,
            "vmzero.q    E701",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E701"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680BE,
            "vmzero.q    E702",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680BF,
            "vmzero.q    E703",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E703"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C0,
            "vmzero.q    M002",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C1,
            "vmzero.q    M012",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M012"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C2,
            "vmzero.q    M022",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C3,
            "vmzero.q    M032",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M032"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C4,
            "vmzero.q    M102",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M102"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C5,
            "vmzero.q    M112",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M112"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C6,
            "vmzero.q    M122",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C7,
            "vmzero.q    M132",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M132"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C8,
            "vmzero.q    M202",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M202"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680C9,
            "vmzero.q    M212",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M212"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680CA,
            "vmzero.q    M222",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680CB,
            "vmzero.q    M232",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M232"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680CC,
            "vmzero.q    M302",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M302"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680CD,
            "vmzero.q    M312",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M312"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680CE,
            "vmzero.q    M322",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680CF,
            "vmzero.q    M332",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M332"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D0,
            "vmzero.q    M402",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M402"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D1,
            "vmzero.q    M412",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M412"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D2,
            "vmzero.q    M422",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D3,
            "vmzero.q    M432",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M432"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D4,
            "vmzero.q    M502",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M502"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D5,
            "vmzero.q    M512",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M512"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D6,
            "vmzero.q    M522",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D7,
            "vmzero.q    M532",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M532"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D8,
            "vmzero.q    M602",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M602"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680D9,
            "vmzero.q    M612",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M612"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680DA,
            "vmzero.q    M622",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680DB,
            "vmzero.q    M632",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M632"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680DC,
            "vmzero.q    M702",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M702"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680DD,
            "vmzero.q    M712",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M712"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680DE,
            "vmzero.q    M722",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680DF,
            "vmzero.q    M732",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("M732"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E0,
            "vmzero.q    E020",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E020"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E1,
            "vmzero.q    E021",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E021"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E2,
            "vmzero.q    E022",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E022"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E3,
            "vmzero.q    E023",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E023"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E4,
            "vmzero.q    E120",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E120"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E5,
            "vmzero.q    E121",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E121"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E6,
            "vmzero.q    E122",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E122"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E7,
            "vmzero.q    E123",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E123"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E8,
            "vmzero.q    E220",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E220"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680E9,
            "vmzero.q    E221",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E221"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680EA,
            "vmzero.q    E222",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E222"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680EB,
            "vmzero.q    E223",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E223"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680EC,
            "vmzero.q    E320",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E320"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680ED,
            "vmzero.q    E321",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E321"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680EE,
            "vmzero.q    E322",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E322"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680EF,
            "vmzero.q    E323",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E323"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F0,
            "vmzero.q    E420",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E420"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F1,
            "vmzero.q    E421",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E421"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F2,
            "vmzero.q    E422",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E422"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F3,
            "vmzero.q    E423",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E423"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F4,
            "vmzero.q    E520",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E520"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F5,
            "vmzero.q    E521",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E521"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F6,
            "vmzero.q    E522",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E522"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F7,
            "vmzero.q    E523",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E523"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F8,
            "vmzero.q    E620",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E620"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680F9,
            "vmzero.q    E621",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E621"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680FA,
            "vmzero.q    E622",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E622"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680FB,
            "vmzero.q    E623",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E623"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680FC,
            "vmzero.q    E720",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E720"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680FD,
            "vmzero.q    E721",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E721"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680FE,
            "vmzero.q    E722",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E722"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF38680FF,
            "vmzero.q    E723",
            Opcode::r4000allegrex_vmzero_q,
            "vmzero.q",
            [Some("E723"), None, None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_branches() {
    const ENTRIES: &[TestEntry] = &[
        /* branches */
        TestEntry::new_r4000allegrex(
            0x49000002,
            "bvf         0, . + 4 + (0x2 << 2)",
            Opcode::r4000allegrex_bvf,
            "bvf",
            [Some("0"), Some(". + 4 + (0x2 << 2)"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x49050002,
            "bvt         1, . + 4 + (0x2 << 2)",
            Opcode::r4000allegrex_bvt,
            "bvt",
            [Some("1"), Some(". + 4 + (0x2 << 2)"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x490A0002,
            "bvfl        2, . + 4 + (0x2 << 2)",
            Opcode::r4000allegrex_bvfl,
            "bvfl",
            [Some("2"), Some(". + 4 + (0x2 << 2)"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x490F0002,
            "bvtl        3, . + 4 + (0x2 << 2)",
            Opcode::r4000allegrex_bvtl,
            "bvtl",
            [Some("3"), Some(". + 4 + (0x2 << 2)"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x49100002,
            "bvf         4, . + 4 + (0x2 << 2)",
            Opcode::r4000allegrex_bvf,
            "bvf",
            [Some("4"), Some(". + 4 + (0x2 << 2)"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x49150002,
            "bvt         5, . + 4 + (0x2 << 2)",
            Opcode::r4000allegrex_bvt,
            "bvt",
            [Some("5"), Some(". + 4 + (0x2 << 2)"), None, None, None],
        ),
        TestEntry {
            instr: Instruction::new(
                0x49000002,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: Some("branches_end"),
            display_flags: InstructionDisplayFlags::default(),
            valid: true,
            expected: "bvf         0, branches_end",
            expected_opcode: Opcode::r4000allegrex_bvf,
            opcode_str: "bvf",
            operands_str: [Some("0"), Some("branches_end"), None, None, None],
        },
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_load_move_store() {
    const ENTRIES: &[TestEntry] = &[
        /* load_move_store */
        TestEntry::new_r4000allegrex(0xC8800040, "lv.s        S000, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8810040, "lv.s        S010, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S010"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8820040, "lv.s        S020, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S020"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8830040, "lv.s        S030, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S030"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8840040, "lv.s        S100, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S100"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8850040, "lv.s        S110, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S110"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8860040, "lv.s        S120, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S120"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8870040, "lv.s        S130, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S130"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8880040, "lv.s        S200, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S200"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8890040, "lv.s        S210, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S210"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88A0040, "lv.s        S220, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S220"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88B0040, "lv.s        S230, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S230"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88C0040, "lv.s        S300, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S300"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88D0040, "lv.s        S310, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S310"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88E0040, "lv.s        S320, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S320"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88F0040, "lv.s        S330, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S330"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8900040, "lv.s        S400, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S400"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8910040, "lv.s        S410, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S410"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8920040, "lv.s        S420, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S420"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8930040, "lv.s        S430, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S430"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8940040, "lv.s        S500, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S500"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8950040, "lv.s        S510, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S510"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8960040, "lv.s        S520, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S520"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8970040, "lv.s        S530, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S530"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8980040, "lv.s        S600, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S600"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8990040, "lv.s        S610, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S610"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89A0040, "lv.s        S620, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S620"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89B0040, "lv.s        S630, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S630"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89C0040, "lv.s        S700, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S700"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89D0040, "lv.s        S710, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S710"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89E0040, "lv.s        S720, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S720"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89F0040, "lv.s        S730, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S730"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8800041, "lv.s        S001, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S001"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8810041, "lv.s        S011, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S011"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8820041, "lv.s        S021, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S021"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8830041, "lv.s        S031, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S031"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8840041, "lv.s        S101, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S101"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8850041, "lv.s        S111, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S111"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8860041, "lv.s        S121, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S121"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8870041, "lv.s        S131, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S131"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8880041, "lv.s        S201, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S201"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8890041, "lv.s        S211, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S211"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88A0041, "lv.s        S221, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S221"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88B0041, "lv.s        S231, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S231"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88C0041, "lv.s        S301, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S301"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88D0041, "lv.s        S311, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S311"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88E0041, "lv.s        S321, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S321"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88F0041, "lv.s        S331, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S331"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8900041, "lv.s        S401, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S401"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8910041, "lv.s        S411, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S411"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8920041, "lv.s        S421, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S421"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8930041, "lv.s        S431, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S431"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8940041, "lv.s        S501, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S501"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8950041, "lv.s        S511, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S511"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8960041, "lv.s        S521, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S521"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8970041, "lv.s        S531, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S531"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8980041, "lv.s        S601, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S601"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8990041, "lv.s        S611, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S611"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89A0041, "lv.s        S621, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S621"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89B0041, "lv.s        S631, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S631"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89C0041, "lv.s        S701, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S701"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89D0041, "lv.s        S711, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S711"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89E0041, "lv.s        S721, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S721"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89F0041, "lv.s        S731, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S731"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8800042, "lv.s        S002, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S002"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8810042, "lv.s        S012, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S012"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8820042, "lv.s        S022, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S022"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8830042, "lv.s        S032, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S032"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8840042, "lv.s        S102, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S102"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8850042, "lv.s        S112, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S112"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8860042, "lv.s        S122, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S122"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8870042, "lv.s        S132, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S132"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8880042, "lv.s        S202, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S202"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8890042, "lv.s        S212, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S212"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88A0042, "lv.s        S222, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S222"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88B0042, "lv.s        S232, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S232"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88C0042, "lv.s        S302, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S302"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88D0042, "lv.s        S312, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S312"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88E0042, "lv.s        S322, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S322"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88F0042, "lv.s        S332, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S332"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8900042, "lv.s        S402, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S402"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8910042, "lv.s        S412, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S412"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8920042, "lv.s        S422, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S422"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8930042, "lv.s        S432, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S432"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8940042, "lv.s        S502, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S502"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8950042, "lv.s        S512, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S512"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8960042, "lv.s        S522, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S522"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8970042, "lv.s        S532, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S532"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8980042, "lv.s        S602, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S602"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8990042, "lv.s        S612, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S612"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89A0042, "lv.s        S622, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S622"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89B0042, "lv.s        S632, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S632"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89C0042, "lv.s        S702, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S702"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89D0042, "lv.s        S712, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S712"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89E0042, "lv.s        S722, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S722"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89F0042, "lv.s        S732, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S732"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8800043, "lv.s        S003, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S003"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8810043, "lv.s        S013, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S013"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8820043, "lv.s        S023, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S023"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8830043, "lv.s        S033, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S033"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8840043, "lv.s        S103, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S103"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8850043, "lv.s        S113, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S113"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8860043, "lv.s        S123, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S123"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8870043, "lv.s        S133, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S133"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8880043, "lv.s        S203, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S203"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8890043, "lv.s        S213, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S213"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88A0043, "lv.s        S223, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S223"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88B0043, "lv.s        S233, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S233"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88C0043, "lv.s        S303, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S303"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88D0043, "lv.s        S313, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S313"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88E0043, "lv.s        S323, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S323"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC88F0043, "lv.s        S333, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S333"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8900043, "lv.s        S403, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S403"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8910043, "lv.s        S413, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S413"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8920043, "lv.s        S423, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S423"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8930043, "lv.s        S433, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S433"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8940043, "lv.s        S503, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S503"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8950043, "lv.s        S513, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S513"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8960043, "lv.s        S523, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S523"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8970043, "lv.s        S533, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S533"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8980043, "lv.s        S603, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S603"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC8990043, "lv.s        S613, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S613"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89A0043, "lv.s        S623, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S623"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89B0043, "lv.s        S633, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S633"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89C0043, "lv.s        S703, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S703"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89D0043, "lv.s        S713, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S713"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89E0043, "lv.s        S723, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S723"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xC89F0043, "lv.s        S733, 0x40($a0)", Opcode::r4000allegrex_lv_s, "lv.s", [Some("S733"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8800040, "lv.q        C000, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8810040, "lv.q        C010, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C010"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8820040, "lv.q        C020, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C020"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8830040, "lv.q        C030, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C030"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8840040, "lv.q        C100, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C100"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8850040, "lv.q        C110, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C110"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8860040, "lv.q        C120, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C120"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8870040, "lv.q        C130, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C130"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8880040, "lv.q        C200, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C200"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8890040, "lv.q        C210, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C210"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88A0040, "lv.q        C220, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C220"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88B0040, "lv.q        C230, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C230"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88C0040, "lv.q        C300, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C300"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88D0040, "lv.q        C310, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C310"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88E0040, "lv.q        C320, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C320"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88F0040, "lv.q        C330, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C330"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8900040, "lv.q        C400, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C400"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8910040, "lv.q        C410, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C410"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8920040, "lv.q        C420, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C420"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8930040, "lv.q        C430, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C430"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8940040, "lv.q        C500, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C500"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8950040, "lv.q        C510, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C510"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8960040, "lv.q        C520, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C520"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8970040, "lv.q        C530, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C530"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8980040, "lv.q        C600, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C600"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8990040, "lv.q        C610, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C610"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89A0040, "lv.q        C620, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C620"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89B0040, "lv.q        C630, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C630"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89C0040, "lv.q        C700, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C700"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89D0040, "lv.q        C710, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C710"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89E0040, "lv.q        C720, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C720"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89F0040, "lv.q        C730, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C730"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8800041, "lv.q        R000, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8810041, "lv.q        R001, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R001"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8820041, "lv.q        R002, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R002"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8830041, "lv.q        R003, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R003"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8840041, "lv.q        R100, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R100"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8850041, "lv.q        R101, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R101"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8860041, "lv.q        R102, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R102"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8870041, "lv.q        R103, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R103"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8880041, "lv.q        R200, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R200"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8890041, "lv.q        R201, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R201"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88A0041, "lv.q        R202, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R202"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88B0041, "lv.q        R203, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R203"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88C0041, "lv.q        R300, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R300"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88D0041, "lv.q        R301, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R301"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88E0041, "lv.q        R302, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R302"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88F0041, "lv.q        R303, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R303"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8900041, "lv.q        R400, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R400"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8910041, "lv.q        R401, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R401"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8920041, "lv.q        R402, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R402"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8930041, "lv.q        R403, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R403"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8940041, "lv.q        R500, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R500"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8950041, "lv.q        R501, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R501"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8960041, "lv.q        R502, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R502"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8970041, "lv.q        R503, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R503"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8980041, "lv.q        R600, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R600"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8990041, "lv.q        R601, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R601"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89A0041, "lv.q        R602, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R602"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89B0041, "lv.q        R603, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R603"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89C0041, "lv.q        R700, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R700"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89D0041, "lv.q        R701, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R701"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89E0041, "lv.q        R702, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R702"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89F0041, "lv.q        R703, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("R703"), Some("0x40($a0)"), None, None, None]),


        // According to the manual those should be invalid, but the sn objdump is happy with them, so idk
        /*
        TestEntry::new_r4000allegrex(0xD8800042, "lv.q        C000, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8810042, "lv.q        C010, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C010"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8820042, "lv.q        C020, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C020"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8830042, "lv.q        C030, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C030"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8840042, "lv.q        C100, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C100"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8850042, "lv.q        C110, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C110"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8860042, "lv.q        C120, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C120"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8870042, "lv.q        C130, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C130"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8880042, "lv.q        C200, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C200"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8890042, "lv.q        C210, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C210"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88A0042, "lv.q        C220, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C220"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88B0042, "lv.q        C230, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C230"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88C0042, "lv.q        C300, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C300"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88D0042, "lv.q        C310, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C310"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88E0042, "lv.q        C320, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C320"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD88F0042, "lv.q        C330, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C330"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8900042, "lv.q        C400, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C400"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8910042, "lv.q        C410, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C410"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8920042, "lv.q        C420, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C420"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8930042, "lv.q        C430, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C430"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8940042, "lv.q        C500, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C500"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8950042, "lv.q        C510, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C510"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8960042, "lv.q        C520, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C520"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8970042, "lv.q        C530, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C530"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8980042, "lv.q        C600, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C600"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD8990042, "lv.q        C610, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C610"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89A0042, "lv.q        C620, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C620"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89B0042, "lv.q        C630, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C630"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89C0042, "lv.q        C700, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C700"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89D0042, "lv.q        C710, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C710"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89E0042, "lv.q        C720, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C720"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD89F0042, "lv.q        C730, 0x40($a0)", Opcode::r4000allegrex_lv_q, "lv.q", [Some("C730"), Some("0x40($a0)"), None, None, None]),
        */
        TestEntry {
            instr: Instruction::new(
                0xD8800042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8800042                   /* lv.q        C000, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C000"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8810042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8810042                   /* lv.q        C010, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C010"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8820042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8820042                   /* lv.q        C020, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C020"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8830042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8830042                   /* lv.q        C030, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C030"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8840042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8840042                   /* lv.q        C100, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C100"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8850042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8850042                   /* lv.q        C110, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C110"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8860042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8860042                   /* lv.q        C120, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C120"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8870042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8870042                   /* lv.q        C130, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C130"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8880042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8880042                   /* lv.q        C200, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C200"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8890042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8890042                   /* lv.q        C210, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C210"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD88A0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD88A0042                   /* lv.q        C220, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C220"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD88B0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD88B0042                   /* lv.q        C230, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C230"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD88C0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD88C0042                   /* lv.q        C300, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C300"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD88D0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD88D0042                   /* lv.q        C310, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C310"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD88E0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD88E0042                   /* lv.q        C320, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C320"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD88F0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD88F0042                   /* lv.q        C330, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C330"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8900042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8900042                   /* lv.q        C400, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C400"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8910042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8910042                   /* lv.q        C410, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C410"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8920042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8920042                   /* lv.q        C420, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C420"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8930042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8930042                   /* lv.q        C430, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C430"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8940042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8940042                   /* lv.q        C500, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C500"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8950042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8950042                   /* lv.q        C510, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C510"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8960042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8960042                   /* lv.q        C520, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C520"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8970042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8970042                   /* lv.q        C530, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C530"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8980042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8980042                   /* lv.q        C600, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C600"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD8990042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD8990042                   /* lv.q        C610, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C610"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD89A0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD89A0042                   /* lv.q        C620, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C620"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD89B0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD89B0042                   /* lv.q        C630, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C630"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD89C0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD89C0042                   /* lv.q        C700, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C700"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD89D0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD89D0042                   /* lv.q        C710, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C710"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD89E0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD89E0042                   /* lv.q        C720, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C720"), Some("0x40($a0)"), None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD89F0042,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD89F0042                   /* lv.q        C730, 0x40($a0) / 00000002 <OpcodeCategory: R4000ALLEGREX_NORMAL> */",
            expected_opcode: Opcode::r4000allegrex_lv_q,
            opcode_str: "lv.q",
            operands_str: [Some("C730"), Some("0x40($a0)"), None, None, None],
        },

        TestEntry::new_r4000allegrex(0x48640004, "mfv         $a0, S100", Opcode::r4000allegrex_mfv, "mfv", [Some("$a0"), Some("S100"), None, None, None]),
        // TestEntry::new_r4000allegrex(0x48640080, "mfvc        $a0, $128", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$128"), None, None, None]),
        // TestEntry::new_r4000allegrex(0x48640080, "mfvc        $a0, VFPU_PFXS", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("VFPU_PFXS"), None, None, None]),
        // TestEntry::new_r4000allegrex(0x48640084, "mfvc        $a0, $132", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$132"), None, None, None]),
        // TestEntry::new_r4000allegrex(0x48640084, "mfvc        $a0, VFPU_INF4", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("VFPU_INF4"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48E40004, "mtv         $a0, S100", Opcode::r4000allegrex_mtv, "mtv", [Some("$a0"), Some("S100"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48E40084, "mtvc        $a0, $132", Opcode::r4000allegrex_mtvc, "mtvc", [Some("$a0"), Some("$132"), None, None, None]),
        // TestEntry::new_r4000allegrex(0x48E40084, "mtvc        $a0, VFPU_INF4", Opcode::r4000allegrex_mtvc, "mtvc", [Some("$a0"), Some("VFPU_INF4"), None, None, None]),
        TestEntry::new_r4000allegrex(0xE8800040, "sv.s        S000, 0x40($a0)", Opcode::r4000allegrex_sv_s, "sv.s", [Some("S000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xF8800040, "sv.q        C000, 0x40($a0)", Opcode::r4000allegrex_sv_q, "sv.q", [Some("C000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xF8800042, "sv.q        C000, 0x40($a0), wb", Opcode::r4000allegrex_sv_q, "sv.q", [Some("C000"), Some("0x40($a0), wb"), None, None, None]),
        TestEntry::new_r4000allegrex(0xF4800040, "svl.q       C000, 0x40($a0)", Opcode::r4000allegrex_svl_q, "svl.q", [Some("C000"), Some("0x40($a0)"), None, None, None]),
        TestEntry::new_r4000allegrex(0xF4800042, "svr.q       C000, 0x40($a0)", Opcode::r4000allegrex_svr_q, "svr.q", [Some("C000"), Some("0x40($a0)"), None, None, None]),

    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_scale_quad_x_quad_matrix() {
    const ENTRIES: &[TestEntry] = &[
        /* scale_quad_x_quad_matrix */
        TestEntry::new_r4000allegrex(
            0xF2008080,
            "vmscl.q     M000, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2018080,
            "vmscl.q     M000, M000, S010",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S010"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2028080,
            "vmscl.q     M000, M000, S020",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S020"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2038080,
            "vmscl.q     M000, M000, S030",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S030"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2048080,
            "vmscl.q     M000, M000, S100",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S100"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2058080,
            "vmscl.q     M000, M000, S110",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S110"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2068080,
            "vmscl.q     M000, M000, S120",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S120"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2078080,
            "vmscl.q     M000, M000, S130",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S130"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2088080,
            "vmscl.q     M000, M000, S200",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S200"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2098080,
            "vmscl.q     M000, M000, S210",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S210"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20A8080,
            "vmscl.q     M000, M000, S220",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S220"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20B8080,
            "vmscl.q     M000, M000, S230",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S230"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20C8080,
            "vmscl.q     M000, M000, S300",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S300"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20D8080,
            "vmscl.q     M000, M000, S310",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S310"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20E8080,
            "vmscl.q     M000, M000, S320",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S320"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20F8080,
            "vmscl.q     M000, M000, S330",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S330"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2108080,
            "vmscl.q     M000, M000, S400",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S400"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2118080,
            "vmscl.q     M000, M000, S410",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S410"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2128080,
            "vmscl.q     M000, M000, S420",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S420"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2138080,
            "vmscl.q     M000, M000, S430",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S430"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2148080,
            "vmscl.q     M000, M000, S500",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S500"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2158080,
            "vmscl.q     M000, M000, S510",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S510"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2168080,
            "vmscl.q     M000, M000, S520",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S520"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2178080,
            "vmscl.q     M000, M000, S530",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S530"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2188080,
            "vmscl.q     M000, M000, S600",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S600"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2198080,
            "vmscl.q     M000, M000, S610",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S610"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF21A8080,
            "vmscl.q     M000, M000, S620",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S620"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF21B8080,
            "vmscl.q     M000, M000, S630",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S630"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF21C8080,
            "vmscl.q     M000, M000, S700",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S700"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF21D8080,
            "vmscl.q     M000, M000, S710",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S710"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF21E8080,
            "vmscl.q     M000, M000, S720",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S720"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF21F8080,
            "vmscl.q     M000, M000, S730",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S730"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2208080,
            "vmscl.q     M000, M000, S001",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S001"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2218080,
            "vmscl.q     M000, M000, S011",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S011"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2228080,
            "vmscl.q     M000, M000, S021",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S021"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2238080,
            "vmscl.q     M000, M000, S031",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S031"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2248080,
            "vmscl.q     M000, M000, S101",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S101"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2258080,
            "vmscl.q     M000, M000, S111",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S111"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2268080,
            "vmscl.q     M000, M000, S121",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S121"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2278080,
            "vmscl.q     M000, M000, S131",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S131"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2288080,
            "vmscl.q     M000, M000, S201",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S201"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2298080,
            "vmscl.q     M000, M000, S211",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S211"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF22A8080,
            "vmscl.q     M000, M000, S221",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S221"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF22B8080,
            "vmscl.q     M000, M000, S231",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S231"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF22C8080,
            "vmscl.q     M000, M000, S301",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S301"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF22D8080,
            "vmscl.q     M000, M000, S311",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S311"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF22E8080,
            "vmscl.q     M000, M000, S321",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S321"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF22F8080,
            "vmscl.q     M000, M000, S331",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S331"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2308080,
            "vmscl.q     M000, M000, S401",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S401"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2318080,
            "vmscl.q     M000, M000, S411",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S411"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2328080,
            "vmscl.q     M000, M000, S421",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S421"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2338080,
            "vmscl.q     M000, M000, S431",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S431"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2348080,
            "vmscl.q     M000, M000, S501",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S501"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2358080,
            "vmscl.q     M000, M000, S511",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S511"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2368080,
            "vmscl.q     M000, M000, S521",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S521"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2378080,
            "vmscl.q     M000, M000, S531",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S531"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2388080,
            "vmscl.q     M000, M000, S601",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S601"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2398080,
            "vmscl.q     M000, M000, S611",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S611"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF23A8080,
            "vmscl.q     M000, M000, S621",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S621"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF23B8080,
            "vmscl.q     M000, M000, S631",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S631"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF23C8080,
            "vmscl.q     M000, M000, S701",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S701"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF23D8080,
            "vmscl.q     M000, M000, S711",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S711"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF23E8080,
            "vmscl.q     M000, M000, S721",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S721"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF23F8080,
            "vmscl.q     M000, M000, S731",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S731"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2408080,
            "vmscl.q     M000, M000, S002",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S002"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2418080,
            "vmscl.q     M000, M000, S012",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S012"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2428080,
            "vmscl.q     M000, M000, S022",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S022"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2438080,
            "vmscl.q     M000, M000, S032",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S032"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2448080,
            "vmscl.q     M000, M000, S102",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S102"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2458080,
            "vmscl.q     M000, M000, S112",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S112"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2468080,
            "vmscl.q     M000, M000, S122",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S122"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2478080,
            "vmscl.q     M000, M000, S132",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S132"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2488080,
            "vmscl.q     M000, M000, S202",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S202"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2498080,
            "vmscl.q     M000, M000, S212",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S212"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF24A8080,
            "vmscl.q     M000, M000, S222",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S222"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF24B8080,
            "vmscl.q     M000, M000, S232",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S232"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF24C8080,
            "vmscl.q     M000, M000, S302",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S302"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF24D8080,
            "vmscl.q     M000, M000, S312",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S312"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF24E8080,
            "vmscl.q     M000, M000, S322",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S322"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF24F8080,
            "vmscl.q     M000, M000, S332",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S332"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2508080,
            "vmscl.q     M000, M000, S402",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S402"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2518080,
            "vmscl.q     M000, M000, S412",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S412"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2528080,
            "vmscl.q     M000, M000, S422",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S422"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2538080,
            "vmscl.q     M000, M000, S432",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S432"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2548080,
            "vmscl.q     M000, M000, S502",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S502"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2558080,
            "vmscl.q     M000, M000, S512",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S512"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2568080,
            "vmscl.q     M000, M000, S522",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S522"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2578080,
            "vmscl.q     M000, M000, S532",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S532"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2588080,
            "vmscl.q     M000, M000, S602",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S602"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2598080,
            "vmscl.q     M000, M000, S612",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S612"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF25A8080,
            "vmscl.q     M000, M000, S622",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S622"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF25B8080,
            "vmscl.q     M000, M000, S632",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S632"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF25C8080,
            "vmscl.q     M000, M000, S702",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S702"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF25D8080,
            "vmscl.q     M000, M000, S712",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S712"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF25E8080,
            "vmscl.q     M000, M000, S722",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S722"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF25F8080,
            "vmscl.q     M000, M000, S732",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S732"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2608080,
            "vmscl.q     M000, M000, S003",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S003"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2618080,
            "vmscl.q     M000, M000, S013",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S013"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2628080,
            "vmscl.q     M000, M000, S023",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S023"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2638080,
            "vmscl.q     M000, M000, S033",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S033"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2648080,
            "vmscl.q     M000, M000, S103",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S103"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2658080,
            "vmscl.q     M000, M000, S113",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S113"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2668080,
            "vmscl.q     M000, M000, S123",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S123"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2678080,
            "vmscl.q     M000, M000, S133",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S133"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2688080,
            "vmscl.q     M000, M000, S203",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S203"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2698080,
            "vmscl.q     M000, M000, S213",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S213"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF26A8080,
            "vmscl.q     M000, M000, S223",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S223"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF26B8080,
            "vmscl.q     M000, M000, S233",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S233"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF26C8080,
            "vmscl.q     M000, M000, S303",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S303"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF26D8080,
            "vmscl.q     M000, M000, S313",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S313"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF26E8080,
            "vmscl.q     M000, M000, S323",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S323"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF26F8080,
            "vmscl.q     M000, M000, S333",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S333"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2708080,
            "vmscl.q     M000, M000, S403",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S403"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2718080,
            "vmscl.q     M000, M000, S413",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S413"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2728080,
            "vmscl.q     M000, M000, S423",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S423"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2738080,
            "vmscl.q     M000, M000, S433",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S433"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2748080,
            "vmscl.q     M000, M000, S503",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S503"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2758080,
            "vmscl.q     M000, M000, S513",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S513"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2768080,
            "vmscl.q     M000, M000, S523",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S523"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2778080,
            "vmscl.q     M000, M000, S533",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S533"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2788080,
            "vmscl.q     M000, M000, S603",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S603"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2798080,
            "vmscl.q     M000, M000, S613",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S613"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF27A8080,
            "vmscl.q     M000, M000, S623",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S623"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF27B8080,
            "vmscl.q     M000, M000, S633",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S633"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF27C8080,
            "vmscl.q     M000, M000, S703",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S703"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF27D8080,
            "vmscl.q     M000, M000, S713",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S713"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF27E8080,
            "vmscl.q     M000, M000, S723",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S723"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF27F8080,
            "vmscl.q     M000, M000, S733",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M000"), Some("M000"), Some("S733"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008081,
            "vmscl.q     M010, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M010"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008082,
            "vmscl.q     M020, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M020"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008083,
            "vmscl.q     M030, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M030"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008084,
            "vmscl.q     M100, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M100"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008085,
            "vmscl.q     M110, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M110"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008086,
            "vmscl.q     M120, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M120"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008087,
            "vmscl.q     M130, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M130"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008088,
            "vmscl.q     M200, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M200"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008089,
            "vmscl.q     M210, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M210"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200808A,
            "vmscl.q     M220, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M220"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200808B,
            "vmscl.q     M230, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M230"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200808C,
            "vmscl.q     M300, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M300"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200808D,
            "vmscl.q     M310, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M310"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200808E,
            "vmscl.q     M320, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M320"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200808F,
            "vmscl.q     M330, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M330"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008090,
            "vmscl.q     M400, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M400"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008091,
            "vmscl.q     M410, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M410"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008092,
            "vmscl.q     M420, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M420"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008093,
            "vmscl.q     M430, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M430"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008094,
            "vmscl.q     M500, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M500"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008095,
            "vmscl.q     M510, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M510"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008096,
            "vmscl.q     M520, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M520"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008097,
            "vmscl.q     M530, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M530"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008098,
            "vmscl.q     M600, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M600"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF2008099,
            "vmscl.q     M610, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M610"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200809A,
            "vmscl.q     M620, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M620"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200809B,
            "vmscl.q     M630, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M630"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200809C,
            "vmscl.q     M700, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M700"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200809D,
            "vmscl.q     M710, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M710"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200809E,
            "vmscl.q     M720, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M720"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF200809F,
            "vmscl.q     M730, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M730"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A0,
            "vmscl.q     E000, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E000"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A1,
            "vmscl.q     E001, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E001"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A2,
            "vmscl.q     E002, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E002"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A3,
            "vmscl.q     E003, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E003"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A4,
            "vmscl.q     E100, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E100"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A5,
            "vmscl.q     E101, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E101"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A6,
            "vmscl.q     E102, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E102"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A7,
            "vmscl.q     E103, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E103"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A8,
            "vmscl.q     E200, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E200"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080A9,
            "vmscl.q     E201, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E201"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080AA,
            "vmscl.q     E202, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E202"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080AB,
            "vmscl.q     E203, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E203"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080AC,
            "vmscl.q     E300, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E300"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080AD,
            "vmscl.q     E301, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E301"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080AE,
            "vmscl.q     E302, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E302"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080AF,
            "vmscl.q     E303, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E303"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B0,
            "vmscl.q     E400, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E400"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B1,
            "vmscl.q     E401, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E401"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B2,
            "vmscl.q     E402, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E402"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B3,
            "vmscl.q     E403, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E403"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B4,
            "vmscl.q     E500, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E500"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B5,
            "vmscl.q     E501, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E501"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B6,
            "vmscl.q     E502, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E502"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B7,
            "vmscl.q     E503, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E503"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B8,
            "vmscl.q     E600, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E600"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080B9,
            "vmscl.q     E601, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E601"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080BA,
            "vmscl.q     E602, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E602"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080BB,
            "vmscl.q     E603, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E603"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080BC,
            "vmscl.q     E700, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E700"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080BD,
            "vmscl.q     E701, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E701"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080BE,
            "vmscl.q     E702, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E702"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080BF,
            "vmscl.q     E703, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E703"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C0,
            "vmscl.q     M002, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M002"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C1,
            "vmscl.q     M012, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M012"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C2,
            "vmscl.q     M022, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M022"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C3,
            "vmscl.q     M032, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M032"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C4,
            "vmscl.q     M102, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M102"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C5,
            "vmscl.q     M112, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M112"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C6,
            "vmscl.q     M122, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M122"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C7,
            "vmscl.q     M132, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M132"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C8,
            "vmscl.q     M202, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M202"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080C9,
            "vmscl.q     M212, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M212"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080CA,
            "vmscl.q     M222, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M222"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080CB,
            "vmscl.q     M232, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M232"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080CC,
            "vmscl.q     M302, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M302"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080CD,
            "vmscl.q     M312, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M312"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080CE,
            "vmscl.q     M322, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M322"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080CF,
            "vmscl.q     M332, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M332"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D0,
            "vmscl.q     M402, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M402"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D1,
            "vmscl.q     M412, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M412"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D2,
            "vmscl.q     M422, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M422"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D3,
            "vmscl.q     M432, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M432"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D4,
            "vmscl.q     M502, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M502"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D5,
            "vmscl.q     M512, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M512"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D6,
            "vmscl.q     M522, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M522"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D7,
            "vmscl.q     M532, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M532"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D8,
            "vmscl.q     M602, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M602"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080D9,
            "vmscl.q     M612, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M612"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080DA,
            "vmscl.q     M622, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M622"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080DB,
            "vmscl.q     M632, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M632"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080DC,
            "vmscl.q     M702, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M702"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080DD,
            "vmscl.q     M712, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M712"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080DE,
            "vmscl.q     M722, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M722"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080DF,
            "vmscl.q     M732, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("M732"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E0,
            "vmscl.q     E020, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E020"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E1,
            "vmscl.q     E021, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E021"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E2,
            "vmscl.q     E022, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E022"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E3,
            "vmscl.q     E023, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E023"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E4,
            "vmscl.q     E120, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E120"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E5,
            "vmscl.q     E121, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E121"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E6,
            "vmscl.q     E122, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E122"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E7,
            "vmscl.q     E123, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E123"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E8,
            "vmscl.q     E220, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E220"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080E9,
            "vmscl.q     E221, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E221"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080EA,
            "vmscl.q     E222, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E222"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080EB,
            "vmscl.q     E223, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E223"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080EC,
            "vmscl.q     E320, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E320"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080ED,
            "vmscl.q     E321, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E321"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080EE,
            "vmscl.q     E322, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E322"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080EF,
            "vmscl.q     E323, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E323"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F0,
            "vmscl.q     E420, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E420"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F1,
            "vmscl.q     E421, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E421"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F2,
            "vmscl.q     E422, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E422"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F3,
            "vmscl.q     E423, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E423"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F4,
            "vmscl.q     E520, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E520"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F5,
            "vmscl.q     E521, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E521"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F6,
            "vmscl.q     E522, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E522"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F7,
            "vmscl.q     E523, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E523"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F8,
            "vmscl.q     E620, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E620"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080F9,
            "vmscl.q     E621, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E621"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080FA,
            "vmscl.q     E622, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E622"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080FB,
            "vmscl.q     E623, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E623"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080FC,
            "vmscl.q     E720, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E720"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080FD,
            "vmscl.q     E721, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E721"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080FE,
            "vmscl.q     E722, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E722"), Some("M000"), Some("S000"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF20080FF,
            "vmscl.q     E723, M000, S000",
            Opcode::r4000allegrex_vmscl_q,
            "vmscl.q",
            [Some("E723"), Some("M000"), Some("S000"), None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_abs_all() {
    const ENTRIES: &[TestEntry] = &[
        /* abs_all */
        TestEntry::new_r4000allegrex(
            0xD0010000,
            "vabs.s      S000, S000",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S000"), Some("S000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010101,
            "vabs.s      S010, S010",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S010"), Some("S010"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010202,
            "vabs.s      S020, S020",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S020"), Some("S020"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010303,
            "vabs.s      S030, S030",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S030"), Some("S030"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010404,
            "vabs.s      S100, S100",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S100"), Some("S100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010505,
            "vabs.s      S110, S110",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S110"), Some("S110"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010606,
            "vabs.s      S120, S120",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S120"), Some("S120"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010707,
            "vabs.s      S130, S130",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S130"), Some("S130"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010808,
            "vabs.s      S200, S200",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S200"), Some("S200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010909,
            "vabs.s      S210, S210",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S210"), Some("S210"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010A0A,
            "vabs.s      S220, S220",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S220"), Some("S220"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010B0B,
            "vabs.s      S230, S230",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S230"), Some("S230"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010C0C,
            "vabs.s      S300, S300",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S300"), Some("S300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010D0D,
            "vabs.s      S310, S310",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S310"), Some("S310"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010E0E,
            "vabs.s      S320, S320",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S320"), Some("S320"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010F0F,
            "vabs.s      S330, S330",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S330"), Some("S330"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011010,
            "vabs.s      S400, S400",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S400"), Some("S400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011111,
            "vabs.s      S410, S410",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S410"), Some("S410"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011212,
            "vabs.s      S420, S420",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S420"), Some("S420"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011313,
            "vabs.s      S430, S430",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S430"), Some("S430"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011414,
            "vabs.s      S500, S500",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S500"), Some("S500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011515,
            "vabs.s      S510, S510",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S510"), Some("S510"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011616,
            "vabs.s      S520, S520",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S520"), Some("S520"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011717,
            "vabs.s      S530, S530",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S530"), Some("S530"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011818,
            "vabs.s      S600, S600",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S600"), Some("S600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011919,
            "vabs.s      S610, S610",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S610"), Some("S610"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011A1A,
            "vabs.s      S620, S620",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S620"), Some("S620"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011B1B,
            "vabs.s      S630, S630",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S630"), Some("S630"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011C1C,
            "vabs.s      S700, S700",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S700"), Some("S700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011D1D,
            "vabs.s      S710, S710",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S710"), Some("S710"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011E1E,
            "vabs.s      S720, S720",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S720"), Some("S720"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011F1F,
            "vabs.s      S730, S730",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S730"), Some("S730"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012020,
            "vabs.s      S001, S001",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S001"), Some("S001"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012121,
            "vabs.s      S011, S011",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S011"), Some("S011"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012222,
            "vabs.s      S021, S021",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S021"), Some("S021"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012323,
            "vabs.s      S031, S031",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S031"), Some("S031"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012424,
            "vabs.s      S101, S101",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S101"), Some("S101"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012525,
            "vabs.s      S111, S111",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S111"), Some("S111"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012626,
            "vabs.s      S121, S121",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S121"), Some("S121"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012727,
            "vabs.s      S131, S131",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S131"), Some("S131"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012828,
            "vabs.s      S201, S201",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S201"), Some("S201"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012929,
            "vabs.s      S211, S211",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S211"), Some("S211"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012A2A,
            "vabs.s      S221, S221",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S221"), Some("S221"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012B2B,
            "vabs.s      S231, S231",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S231"), Some("S231"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012C2C,
            "vabs.s      S301, S301",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S301"), Some("S301"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012D2D,
            "vabs.s      S311, S311",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S311"), Some("S311"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012E2E,
            "vabs.s      S321, S321",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S321"), Some("S321"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012F2F,
            "vabs.s      S331, S331",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S331"), Some("S331"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013030,
            "vabs.s      S401, S401",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S401"), Some("S401"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013131,
            "vabs.s      S411, S411",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S411"), Some("S411"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013232,
            "vabs.s      S421, S421",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S421"), Some("S421"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013333,
            "vabs.s      S431, S431",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S431"), Some("S431"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013434,
            "vabs.s      S501, S501",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S501"), Some("S501"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013535,
            "vabs.s      S511, S511",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S511"), Some("S511"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013636,
            "vabs.s      S521, S521",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S521"), Some("S521"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013737,
            "vabs.s      S531, S531",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S531"), Some("S531"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013838,
            "vabs.s      S601, S601",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S601"), Some("S601"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013939,
            "vabs.s      S611, S611",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S611"), Some("S611"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013A3A,
            "vabs.s      S621, S621",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S621"), Some("S621"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013B3B,
            "vabs.s      S631, S631",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S631"), Some("S631"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013C3C,
            "vabs.s      S701, S701",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S701"), Some("S701"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013D3D,
            "vabs.s      S711, S711",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S711"), Some("S711"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013E3E,
            "vabs.s      S721, S721",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S721"), Some("S721"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013F3F,
            "vabs.s      S731, S731",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S731"), Some("S731"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014040,
            "vabs.s      S002, S002",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S002"), Some("S002"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014141,
            "vabs.s      S012, S012",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S012"), Some("S012"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014242,
            "vabs.s      S022, S022",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S022"), Some("S022"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014343,
            "vabs.s      S032, S032",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S032"), Some("S032"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014444,
            "vabs.s      S102, S102",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S102"), Some("S102"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014545,
            "vabs.s      S112, S112",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S112"), Some("S112"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014646,
            "vabs.s      S122, S122",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S122"), Some("S122"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014747,
            "vabs.s      S132, S132",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S132"), Some("S132"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014848,
            "vabs.s      S202, S202",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S202"), Some("S202"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014949,
            "vabs.s      S212, S212",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S212"), Some("S212"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014A4A,
            "vabs.s      S222, S222",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S222"), Some("S222"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014B4B,
            "vabs.s      S232, S232",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S232"), Some("S232"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014C4C,
            "vabs.s      S302, S302",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S302"), Some("S302"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014D4D,
            "vabs.s      S312, S312",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S312"), Some("S312"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014E4E,
            "vabs.s      S322, S322",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S322"), Some("S322"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014F4F,
            "vabs.s      S332, S332",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S332"), Some("S332"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015050,
            "vabs.s      S402, S402",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S402"), Some("S402"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015151,
            "vabs.s      S412, S412",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S412"), Some("S412"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015252,
            "vabs.s      S422, S422",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S422"), Some("S422"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015353,
            "vabs.s      S432, S432",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S432"), Some("S432"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015454,
            "vabs.s      S502, S502",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S502"), Some("S502"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015555,
            "vabs.s      S512, S512",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S512"), Some("S512"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015656,
            "vabs.s      S522, S522",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S522"), Some("S522"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015757,
            "vabs.s      S532, S532",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S532"), Some("S532"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015858,
            "vabs.s      S602, S602",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S602"), Some("S602"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015959,
            "vabs.s      S612, S612",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S612"), Some("S612"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015A5A,
            "vabs.s      S622, S622",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S622"), Some("S622"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015B5B,
            "vabs.s      S632, S632",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S632"), Some("S632"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015C5C,
            "vabs.s      S702, S702",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S702"), Some("S702"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015D5D,
            "vabs.s      S712, S712",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S712"), Some("S712"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015E5E,
            "vabs.s      S722, S722",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S722"), Some("S722"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015F5F,
            "vabs.s      S732, S732",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S732"), Some("S732"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016060,
            "vabs.s      S003, S003",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S003"), Some("S003"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016161,
            "vabs.s      S013, S013",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S013"), Some("S013"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016262,
            "vabs.s      S023, S023",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S023"), Some("S023"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016363,
            "vabs.s      S033, S033",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S033"), Some("S033"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016464,
            "vabs.s      S103, S103",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S103"), Some("S103"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016565,
            "vabs.s      S113, S113",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S113"), Some("S113"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016666,
            "vabs.s      S123, S123",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S123"), Some("S123"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016767,
            "vabs.s      S133, S133",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S133"), Some("S133"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016868,
            "vabs.s      S203, S203",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S203"), Some("S203"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016969,
            "vabs.s      S213, S213",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S213"), Some("S213"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016A6A,
            "vabs.s      S223, S223",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S223"), Some("S223"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016B6B,
            "vabs.s      S233, S233",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S233"), Some("S233"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016C6C,
            "vabs.s      S303, S303",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S303"), Some("S303"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016D6D,
            "vabs.s      S313, S313",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S313"), Some("S313"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016E6E,
            "vabs.s      S323, S323",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S323"), Some("S323"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016F6F,
            "vabs.s      S333, S333",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S333"), Some("S333"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017070,
            "vabs.s      S403, S403",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S403"), Some("S403"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017171,
            "vabs.s      S413, S413",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S413"), Some("S413"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017272,
            "vabs.s      S423, S423",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S423"), Some("S423"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017373,
            "vabs.s      S433, S433",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S433"), Some("S433"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017474,
            "vabs.s      S503, S503",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S503"), Some("S503"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017575,
            "vabs.s      S513, S513",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S513"), Some("S513"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017676,
            "vabs.s      S523, S523",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S523"), Some("S523"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017777,
            "vabs.s      S533, S533",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S533"), Some("S533"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017878,
            "vabs.s      S603, S603",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S603"), Some("S603"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017979,
            "vabs.s      S613, S613",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S613"), Some("S613"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017A7A,
            "vabs.s      S623, S623",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S623"), Some("S623"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017B7B,
            "vabs.s      S633, S633",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S633"), Some("S633"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017C7C,
            "vabs.s      S703, S703",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S703"), Some("S703"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017D7D,
            "vabs.s      S713, S713",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S713"), Some("S713"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017E7E,
            "vabs.s      S723, S723",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S723"), Some("S723"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017F7F,
            "vabs.s      S733, S733",
            Opcode::r4000allegrex_vabs_s,
            "vabs.s",
            [Some("S733"), Some("S733"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010080,
            "vabs.p      C000, C000",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C000"), Some("C000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010181,
            "vabs.p      C010, C010",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C010"), Some("C010"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010282,
            "vabs.p      C020, C020",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C020"), Some("C020"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010383,
            "vabs.p      C030, C030",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C030"), Some("C030"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010484,
            "vabs.p      C100, C100",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C100"), Some("C100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010585,
            "vabs.p      C110, C110",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C110"), Some("C110"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010686,
            "vabs.p      C120, C120",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C120"), Some("C120"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010787,
            "vabs.p      C130, C130",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C130"), Some("C130"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010888,
            "vabs.p      C200, C200",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C200"), Some("C200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010989,
            "vabs.p      C210, C210",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C210"), Some("C210"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010A8A,
            "vabs.p      C220, C220",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C220"), Some("C220"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010B8B,
            "vabs.p      C230, C230",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C230"), Some("C230"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010C8C,
            "vabs.p      C300, C300",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C300"), Some("C300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010D8D,
            "vabs.p      C310, C310",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C310"), Some("C310"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010E8E,
            "vabs.p      C320, C320",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C320"), Some("C320"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0010F8F,
            "vabs.p      C330, C330",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C330"), Some("C330"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011090,
            "vabs.p      C400, C400",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C400"), Some("C400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011191,
            "vabs.p      C410, C410",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C410"), Some("C410"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011292,
            "vabs.p      C420, C420",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C420"), Some("C420"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011393,
            "vabs.p      C430, C430",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C430"), Some("C430"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011494,
            "vabs.p      C500, C500",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C500"), Some("C500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011595,
            "vabs.p      C510, C510",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C510"), Some("C510"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011696,
            "vabs.p      C520, C520",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C520"), Some("C520"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011797,
            "vabs.p      C530, C530",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C530"), Some("C530"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011898,
            "vabs.p      C600, C600",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C600"), Some("C600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011999,
            "vabs.p      C610, C610",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C610"), Some("C610"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011A9A,
            "vabs.p      C620, C620",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C620"), Some("C620"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011B9B,
            "vabs.p      C630, C630",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C630"), Some("C630"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011C9C,
            "vabs.p      C700, C700",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C700"), Some("C700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011D9D,
            "vabs.p      C710, C710",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C710"), Some("C710"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011E9E,
            "vabs.p      C720, C720",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C720"), Some("C720"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0011F9F,
            "vabs.p      C730, C730",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C730"), Some("C730"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00120A0,
            "vabs.p      R000, R000",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R000"), Some("R000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00121A1,
            "vabs.p      R001, R001",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R001"), Some("R001"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00122A2,
            "vabs.p      R002, R002",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R002"), Some("R002"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00123A3,
            "vabs.p      R003, R003",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R003"), Some("R003"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00124A4,
            "vabs.p      R100, R100",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R100"), Some("R100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00125A5,
            "vabs.p      R101, R101",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R101"), Some("R101"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00126A6,
            "vabs.p      R102, R102",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R102"), Some("R102"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00127A7,
            "vabs.p      R103, R103",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R103"), Some("R103"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00128A8,
            "vabs.p      R200, R200",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R200"), Some("R200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00129A9,
            "vabs.p      R201, R201",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R201"), Some("R201"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012AAA,
            "vabs.p      R202, R202",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R202"), Some("R202"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012BAB,
            "vabs.p      R203, R203",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R203"), Some("R203"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012CAC,
            "vabs.p      R300, R300",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R300"), Some("R300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012DAD,
            "vabs.p      R301, R301",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R301"), Some("R301"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012EAE,
            "vabs.p      R302, R302",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R302"), Some("R302"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0012FAF,
            "vabs.p      R303, R303",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R303"), Some("R303"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00130B0,
            "vabs.p      R400, R400",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R400"), Some("R400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00131B1,
            "vabs.p      R401, R401",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R401"), Some("R401"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00132B2,
            "vabs.p      R402, R402",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R402"), Some("R402"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00133B3,
            "vabs.p      R403, R403",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R403"), Some("R403"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00134B4,
            "vabs.p      R500, R500",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R500"), Some("R500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00135B5,
            "vabs.p      R501, R501",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R501"), Some("R501"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00136B6,
            "vabs.p      R502, R502",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R502"), Some("R502"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00137B7,
            "vabs.p      R503, R503",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R503"), Some("R503"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00138B8,
            "vabs.p      R600, R600",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R600"), Some("R600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00139B9,
            "vabs.p      R601, R601",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R601"), Some("R601"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013ABA,
            "vabs.p      R602, R602",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R602"), Some("R602"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013BBB,
            "vabs.p      R603, R603",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R603"), Some("R603"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013CBC,
            "vabs.p      R700, R700",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R700"), Some("R700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013DBD,
            "vabs.p      R701, R701",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R701"), Some("R701"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013EBE,
            "vabs.p      R702, R702",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R702"), Some("R702"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0013FBF,
            "vabs.p      R703, R703",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R703"), Some("R703"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00140C0,
            "vabs.p      C002, C002",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C002"), Some("C002"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00141C1,
            "vabs.p      C012, C012",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C012"), Some("C012"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00142C2,
            "vabs.p      C022, C022",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C022"), Some("C022"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00143C3,
            "vabs.p      C032, C032",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C032"), Some("C032"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00144C4,
            "vabs.p      C102, C102",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C102"), Some("C102"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00145C5,
            "vabs.p      C112, C112",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C112"), Some("C112"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00146C6,
            "vabs.p      C122, C122",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C122"), Some("C122"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00147C7,
            "vabs.p      C132, C132",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C132"), Some("C132"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00148C8,
            "vabs.p      C202, C202",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C202"), Some("C202"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00149C9,
            "vabs.p      C212, C212",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C212"), Some("C212"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014ACA,
            "vabs.p      C222, C222",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C222"), Some("C222"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014BCB,
            "vabs.p      C232, C232",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C232"), Some("C232"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014CCC,
            "vabs.p      C302, C302",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C302"), Some("C302"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014DCD,
            "vabs.p      C312, C312",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C312"), Some("C312"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014ECE,
            "vabs.p      C322, C322",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C322"), Some("C322"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0014FCF,
            "vabs.p      C332, C332",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C332"), Some("C332"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00150D0,
            "vabs.p      C402, C402",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C402"), Some("C402"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00151D1,
            "vabs.p      C412, C412",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C412"), Some("C412"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00152D2,
            "vabs.p      C422, C422",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C422"), Some("C422"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00153D3,
            "vabs.p      C432, C432",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C432"), Some("C432"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00154D4,
            "vabs.p      C502, C502",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C502"), Some("C502"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00155D5,
            "vabs.p      C512, C512",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C512"), Some("C512"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00156D6,
            "vabs.p      C522, C522",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C522"), Some("C522"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00157D7,
            "vabs.p      C532, C532",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C532"), Some("C532"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00158D8,
            "vabs.p      C602, C602",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C602"), Some("C602"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00159D9,
            "vabs.p      C612, C612",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C612"), Some("C612"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015ADA,
            "vabs.p      C622, C622",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C622"), Some("C622"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015BDB,
            "vabs.p      C632, C632",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C632"), Some("C632"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015CDC,
            "vabs.p      C702, C702",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C702"), Some("C702"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015DDD,
            "vabs.p      C712, C712",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C712"), Some("C712"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015EDE,
            "vabs.p      C722, C722",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C722"), Some("C722"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0015FDF,
            "vabs.p      C732, C732",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("C732"), Some("C732"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00160E0,
            "vabs.p      R020, R020",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R020"), Some("R020"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00161E1,
            "vabs.p      R021, R021",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R021"), Some("R021"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00162E2,
            "vabs.p      R022, R022",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R022"), Some("R022"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00163E3,
            "vabs.p      R023, R023",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R023"), Some("R023"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00164E4,
            "vabs.p      R120, R120",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R120"), Some("R120"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00165E5,
            "vabs.p      R121, R121",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R121"), Some("R121"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00166E6,
            "vabs.p      R122, R122",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R122"), Some("R122"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00167E7,
            "vabs.p      R123, R123",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R123"), Some("R123"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00168E8,
            "vabs.p      R220, R220",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R220"), Some("R220"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00169E9,
            "vabs.p      R221, R221",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R221"), Some("R221"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016AEA,
            "vabs.p      R222, R222",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R222"), Some("R222"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016BEB,
            "vabs.p      R223, R223",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R223"), Some("R223"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016CEC,
            "vabs.p      R320, R320",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R320"), Some("R320"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016DED,
            "vabs.p      R321, R321",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R321"), Some("R321"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016EEE,
            "vabs.p      R322, R322",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R322"), Some("R322"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0016FEF,
            "vabs.p      R323, R323",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R323"), Some("R323"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00170F0,
            "vabs.p      R420, R420",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R420"), Some("R420"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00171F1,
            "vabs.p      R421, R421",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R421"), Some("R421"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00172F2,
            "vabs.p      R422, R422",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R422"), Some("R422"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00173F3,
            "vabs.p      R423, R423",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R423"), Some("R423"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00174F4,
            "vabs.p      R520, R520",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R520"), Some("R520"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00175F5,
            "vabs.p      R521, R521",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R521"), Some("R521"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00176F6,
            "vabs.p      R522, R522",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R522"), Some("R522"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00177F7,
            "vabs.p      R523, R523",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R523"), Some("R523"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00178F8,
            "vabs.p      R620, R620",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R620"), Some("R620"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD00179F9,
            "vabs.p      R621, R621",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R621"), Some("R621"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017AFA,
            "vabs.p      R622, R622",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R622"), Some("R622"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017BFB,
            "vabs.p      R623, R623",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R623"), Some("R623"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017CFC,
            "vabs.p      R720, R720",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R720"), Some("R720"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017DFD,
            "vabs.p      R721, R721",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R721"), Some("R721"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017EFE,
            "vabs.p      R722, R722",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R722"), Some("R722"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0017FFF,
            "vabs.p      R723, R723",
            Opcode::r4000allegrex_vabs_p,
            "vabs.p",
            [Some("R723"), Some("R723"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018000,
            "vabs.t      C000, C000",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C000"), Some("C000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018101,
            "vabs.t      C010, C010",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C010"), Some("C010"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018202,
            "vabs.t      C020, C020",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C020"), Some("C020"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018303,
            "vabs.t      C030, C030",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C030"), Some("C030"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018404,
            "vabs.t      C100, C100",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C100"), Some("C100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018505,
            "vabs.t      C110, C110",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C110"), Some("C110"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018606,
            "vabs.t      C120, C120",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C120"), Some("C120"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018707,
            "vabs.t      C130, C130",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C130"), Some("C130"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018808,
            "vabs.t      C200, C200",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C200"), Some("C200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018909,
            "vabs.t      C210, C210",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C210"), Some("C210"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018A0A,
            "vabs.t      C220, C220",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C220"), Some("C220"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018B0B,
            "vabs.t      C230, C230",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C230"), Some("C230"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018C0C,
            "vabs.t      C300, C300",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C300"), Some("C300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018D0D,
            "vabs.t      C310, C310",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C310"), Some("C310"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018E0E,
            "vabs.t      C320, C320",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C320"), Some("C320"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018F0F,
            "vabs.t      C330, C330",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C330"), Some("C330"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019010,
            "vabs.t      C400, C400",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C400"), Some("C400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019111,
            "vabs.t      C410, C410",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C410"), Some("C410"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019212,
            "vabs.t      C420, C420",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C420"), Some("C420"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019313,
            "vabs.t      C430, C430",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C430"), Some("C430"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019414,
            "vabs.t      C500, C500",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C500"), Some("C500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019515,
            "vabs.t      C510, C510",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C510"), Some("C510"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019616,
            "vabs.t      C520, C520",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C520"), Some("C520"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019717,
            "vabs.t      C530, C530",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C530"), Some("C530"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019818,
            "vabs.t      C600, C600",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C600"), Some("C600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019919,
            "vabs.t      C610, C610",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C610"), Some("C610"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019A1A,
            "vabs.t      C620, C620",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C620"), Some("C620"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019B1B,
            "vabs.t      C630, C630",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C630"), Some("C630"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019C1C,
            "vabs.t      C700, C700",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C700"), Some("C700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019D1D,
            "vabs.t      C710, C710",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C710"), Some("C710"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019E1E,
            "vabs.t      C720, C720",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C720"), Some("C720"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019F1F,
            "vabs.t      C730, C730",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C730"), Some("C730"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A020,
            "vabs.t      R000, R000",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R000"), Some("R000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A121,
            "vabs.t      R001, R001",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R001"), Some("R001"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A222,
            "vabs.t      R002, R002",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R002"), Some("R002"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A323,
            "vabs.t      R003, R003",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R003"), Some("R003"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A424,
            "vabs.t      R100, R100",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R100"), Some("R100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A525,
            "vabs.t      R101, R101",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R101"), Some("R101"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A626,
            "vabs.t      R102, R102",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R102"), Some("R102"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A727,
            "vabs.t      R103, R103",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R103"), Some("R103"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A828,
            "vabs.t      R200, R200",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R200"), Some("R200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A929,
            "vabs.t      R201, R201",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R201"), Some("R201"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AA2A,
            "vabs.t      R202, R202",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R202"), Some("R202"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AB2B,
            "vabs.t      R203, R203",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R203"), Some("R203"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AC2C,
            "vabs.t      R300, R300",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R300"), Some("R300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AD2D,
            "vabs.t      R301, R301",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R301"), Some("R301"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AE2E,
            "vabs.t      R302, R302",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R302"), Some("R302"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AF2F,
            "vabs.t      R303, R303",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R303"), Some("R303"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B030,
            "vabs.t      R400, R400",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R400"), Some("R400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B131,
            "vabs.t      R401, R401",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R401"), Some("R401"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B232,
            "vabs.t      R402, R402",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R402"), Some("R402"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B333,
            "vabs.t      R403, R403",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R403"), Some("R403"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B434,
            "vabs.t      R500, R500",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R500"), Some("R500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B535,
            "vabs.t      R501, R501",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R501"), Some("R501"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B636,
            "vabs.t      R502, R502",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R502"), Some("R502"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B737,
            "vabs.t      R503, R503",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R503"), Some("R503"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B838,
            "vabs.t      R600, R600",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R600"), Some("R600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B939,
            "vabs.t      R601, R601",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R601"), Some("R601"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BA3A,
            "vabs.t      R602, R602",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R602"), Some("R602"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BB3B,
            "vabs.t      R603, R603",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R603"), Some("R603"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BC3C,
            "vabs.t      R700, R700",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R700"), Some("R700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BD3D,
            "vabs.t      R701, R701",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R701"), Some("R701"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BE3E,
            "vabs.t      R702, R702",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R702"), Some("R702"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BF3F,
            "vabs.t      R703, R703",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R703"), Some("R703"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C040,
            "vabs.t      C001, C001",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C001"), Some("C001"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C141,
            "vabs.t      C011, C011",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C011"), Some("C011"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C242,
            "vabs.t      C021, C021",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C021"), Some("C021"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C343,
            "vabs.t      C031, C031",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C031"), Some("C031"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C444,
            "vabs.t      C101, C101",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C101"), Some("C101"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C545,
            "vabs.t      C111, C111",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C111"), Some("C111"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C646,
            "vabs.t      C121, C121",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C121"), Some("C121"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C747,
            "vabs.t      C131, C131",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C131"), Some("C131"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C848,
            "vabs.t      C201, C201",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C201"), Some("C201"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C949,
            "vabs.t      C211, C211",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C211"), Some("C211"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CA4A,
            "vabs.t      C221, C221",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C221"), Some("C221"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CB4B,
            "vabs.t      C231, C231",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C231"), Some("C231"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CC4C,
            "vabs.t      C301, C301",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C301"), Some("C301"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CD4D,
            "vabs.t      C311, C311",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C311"), Some("C311"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CE4E,
            "vabs.t      C321, C321",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C321"), Some("C321"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CF4F,
            "vabs.t      C331, C331",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C331"), Some("C331"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D050,
            "vabs.t      C401, C401",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C401"), Some("C401"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D151,
            "vabs.t      C411, C411",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C411"), Some("C411"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D252,
            "vabs.t      C421, C421",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C421"), Some("C421"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D353,
            "vabs.t      C431, C431",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C431"), Some("C431"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D454,
            "vabs.t      C501, C501",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C501"), Some("C501"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D555,
            "vabs.t      C511, C511",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C511"), Some("C511"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D656,
            "vabs.t      C521, C521",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C521"), Some("C521"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D757,
            "vabs.t      C531, C531",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C531"), Some("C531"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D858,
            "vabs.t      C601, C601",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C601"), Some("C601"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D959,
            "vabs.t      C611, C611",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C611"), Some("C611"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DA5A,
            "vabs.t      C621, C621",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C621"), Some("C621"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DB5B,
            "vabs.t      C631, C631",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C631"), Some("C631"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DC5C,
            "vabs.t      C701, C701",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C701"), Some("C701"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DD5D,
            "vabs.t      C711, C711",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C711"), Some("C711"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DE5E,
            "vabs.t      C721, C721",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C721"), Some("C721"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DF5F,
            "vabs.t      C731, C731",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("C731"), Some("C731"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E060,
            "vabs.t      R010, R010",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R010"), Some("R010"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E161,
            "vabs.t      R011, R011",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R011"), Some("R011"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E262,
            "vabs.t      R012, R012",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R012"), Some("R012"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E363,
            "vabs.t      R013, R013",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R013"), Some("R013"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E464,
            "vabs.t      R110, R110",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R110"), Some("R110"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E565,
            "vabs.t      R111, R111",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R111"), Some("R111"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E666,
            "vabs.t      R112, R112",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R112"), Some("R112"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E767,
            "vabs.t      R113, R113",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R113"), Some("R113"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E868,
            "vabs.t      R210, R210",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R210"), Some("R210"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E969,
            "vabs.t      R211, R211",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R211"), Some("R211"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EA6A,
            "vabs.t      R212, R212",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R212"), Some("R212"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EB6B,
            "vabs.t      R213, R213",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R213"), Some("R213"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EC6C,
            "vabs.t      R310, R310",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R310"), Some("R310"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001ED6D,
            "vabs.t      R311, R311",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R311"), Some("R311"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EE6E,
            "vabs.t      R312, R312",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R312"), Some("R312"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EF6F,
            "vabs.t      R313, R313",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R313"), Some("R313"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F070,
            "vabs.t      R410, R410",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R410"), Some("R410"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F171,
            "vabs.t      R411, R411",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R411"), Some("R411"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F272,
            "vabs.t      R412, R412",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R412"), Some("R412"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F373,
            "vabs.t      R413, R413",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R413"), Some("R413"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F474,
            "vabs.t      R510, R510",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R510"), Some("R510"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F575,
            "vabs.t      R511, R511",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R511"), Some("R511"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F676,
            "vabs.t      R512, R512",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R512"), Some("R512"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F777,
            "vabs.t      R513, R513",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R513"), Some("R513"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F878,
            "vabs.t      R610, R610",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R610"), Some("R610"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F979,
            "vabs.t      R611, R611",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R611"), Some("R611"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FA7A,
            "vabs.t      R612, R612",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R612"), Some("R612"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FB7B,
            "vabs.t      R613, R613",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R613"), Some("R613"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FC7C,
            "vabs.t      R710, R710",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R710"), Some("R710"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FD7D,
            "vabs.t      R711, R711",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R711"), Some("R711"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FE7E,
            "vabs.t      R712, R712",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R712"), Some("R712"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FF7F,
            "vabs.t      R713, R713",
            Opcode::r4000allegrex_vabs_t,
            "vabs.t",
            [Some("R713"), Some("R713"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018080,
            "vabs.q      C000, C000",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C000"), Some("C000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018181,
            "vabs.q      C010, C010",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C010"), Some("C010"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018282,
            "vabs.q      C020, C020",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C020"), Some("C020"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018383,
            "vabs.q      C030, C030",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C030"), Some("C030"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018484,
            "vabs.q      C100, C100",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C100"), Some("C100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018585,
            "vabs.q      C110, C110",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C110"), Some("C110"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018686,
            "vabs.q      C120, C120",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C120"), Some("C120"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018787,
            "vabs.q      C130, C130",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C130"), Some("C130"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018888,
            "vabs.q      C200, C200",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C200"), Some("C200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018989,
            "vabs.q      C210, C210",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C210"), Some("C210"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018A8A,
            "vabs.q      C220, C220",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C220"), Some("C220"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018B8B,
            "vabs.q      C230, C230",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C230"), Some("C230"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018C8C,
            "vabs.q      C300, C300",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C300"), Some("C300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018D8D,
            "vabs.q      C310, C310",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C310"), Some("C310"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018E8E,
            "vabs.q      C320, C320",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C320"), Some("C320"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0018F8F,
            "vabs.q      C330, C330",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C330"), Some("C330"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019090,
            "vabs.q      C400, C400",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C400"), Some("C400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019191,
            "vabs.q      C410, C410",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C410"), Some("C410"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019292,
            "vabs.q      C420, C420",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C420"), Some("C420"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019393,
            "vabs.q      C430, C430",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C430"), Some("C430"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019494,
            "vabs.q      C500, C500",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C500"), Some("C500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019595,
            "vabs.q      C510, C510",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C510"), Some("C510"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019696,
            "vabs.q      C520, C520",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C520"), Some("C520"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019797,
            "vabs.q      C530, C530",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C530"), Some("C530"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019898,
            "vabs.q      C600, C600",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C600"), Some("C600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019999,
            "vabs.q      C610, C610",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C610"), Some("C610"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019A9A,
            "vabs.q      C620, C620",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C620"), Some("C620"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019B9B,
            "vabs.q      C630, C630",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C630"), Some("C630"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019C9C,
            "vabs.q      C700, C700",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C700"), Some("C700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019D9D,
            "vabs.q      C710, C710",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C710"), Some("C710"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019E9E,
            "vabs.q      C720, C720",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C720"), Some("C720"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD0019F9F,
            "vabs.q      C730, C730",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C730"), Some("C730"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A0A0,
            "vabs.q      R000, R000",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R000"), Some("R000"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A1A1,
            "vabs.q      R001, R001",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R001"), Some("R001"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A2A2,
            "vabs.q      R002, R002",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R002"), Some("R002"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A3A3,
            "vabs.q      R003, R003",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R003"), Some("R003"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A4A4,
            "vabs.q      R100, R100",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R100"), Some("R100"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A5A5,
            "vabs.q      R101, R101",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R101"), Some("R101"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A6A6,
            "vabs.q      R102, R102",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R102"), Some("R102"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A7A7,
            "vabs.q      R103, R103",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R103"), Some("R103"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A8A8,
            "vabs.q      R200, R200",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R200"), Some("R200"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001A9A9,
            "vabs.q      R201, R201",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R201"), Some("R201"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AAAA,
            "vabs.q      R202, R202",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R202"), Some("R202"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001ABAB,
            "vabs.q      R203, R203",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R203"), Some("R203"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001ACAC,
            "vabs.q      R300, R300",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R300"), Some("R300"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001ADAD,
            "vabs.q      R301, R301",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R301"), Some("R301"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AEAE,
            "vabs.q      R302, R302",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R302"), Some("R302"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001AFAF,
            "vabs.q      R303, R303",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R303"), Some("R303"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B0B0,
            "vabs.q      R400, R400",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R400"), Some("R400"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B1B1,
            "vabs.q      R401, R401",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R401"), Some("R401"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B2B2,
            "vabs.q      R402, R402",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R402"), Some("R402"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B3B3,
            "vabs.q      R403, R403",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R403"), Some("R403"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B4B4,
            "vabs.q      R500, R500",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R500"), Some("R500"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B5B5,
            "vabs.q      R501, R501",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R501"), Some("R501"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B6B6,
            "vabs.q      R502, R502",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R502"), Some("R502"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B7B7,
            "vabs.q      R503, R503",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R503"), Some("R503"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B8B8,
            "vabs.q      R600, R600",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R600"), Some("R600"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001B9B9,
            "vabs.q      R601, R601",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R601"), Some("R601"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BABA,
            "vabs.q      R602, R602",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R602"), Some("R602"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BBBB,
            "vabs.q      R603, R603",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R603"), Some("R603"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BCBC,
            "vabs.q      R700, R700",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R700"), Some("R700"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BDBD,
            "vabs.q      R701, R701",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R701"), Some("R701"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BEBE,
            "vabs.q      R702, R702",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R702"), Some("R702"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001BFBF,
            "vabs.q      R703, R703",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R703"), Some("R703"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C0C0,
            "vabs.q      C002, C002",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C002"), Some("C002"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C1C1,
            "vabs.q      C012, C012",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C012"), Some("C012"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C2C2,
            "vabs.q      C022, C022",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C022"), Some("C022"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C3C3,
            "vabs.q      C032, C032",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C032"), Some("C032"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C4C4,
            "vabs.q      C102, C102",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C102"), Some("C102"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C5C5,
            "vabs.q      C112, C112",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C112"), Some("C112"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C6C6,
            "vabs.q      C122, C122",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C122"), Some("C122"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C7C7,
            "vabs.q      C132, C132",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C132"), Some("C132"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C8C8,
            "vabs.q      C202, C202",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C202"), Some("C202"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001C9C9,
            "vabs.q      C212, C212",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C212"), Some("C212"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CACA,
            "vabs.q      C222, C222",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C222"), Some("C222"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CBCB,
            "vabs.q      C232, C232",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C232"), Some("C232"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CCCC,
            "vabs.q      C302, C302",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C302"), Some("C302"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CDCD,
            "vabs.q      C312, C312",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C312"), Some("C312"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CECE,
            "vabs.q      C322, C322",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C322"), Some("C322"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001CFCF,
            "vabs.q      C332, C332",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C332"), Some("C332"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D0D0,
            "vabs.q      C402, C402",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C402"), Some("C402"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D1D1,
            "vabs.q      C412, C412",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C412"), Some("C412"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D2D2,
            "vabs.q      C422, C422",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C422"), Some("C422"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D3D3,
            "vabs.q      C432, C432",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C432"), Some("C432"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D4D4,
            "vabs.q      C502, C502",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C502"), Some("C502"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D5D5,
            "vabs.q      C512, C512",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C512"), Some("C512"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D6D6,
            "vabs.q      C522, C522",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C522"), Some("C522"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D7D7,
            "vabs.q      C532, C532",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C532"), Some("C532"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D8D8,
            "vabs.q      C602, C602",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C602"), Some("C602"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001D9D9,
            "vabs.q      C612, C612",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C612"), Some("C612"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DADA,
            "vabs.q      C622, C622",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C622"), Some("C622"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DBDB,
            "vabs.q      C632, C632",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C632"), Some("C632"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DCDC,
            "vabs.q      C702, C702",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C702"), Some("C702"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DDDD,
            "vabs.q      C712, C712",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C712"), Some("C712"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DEDE,
            "vabs.q      C722, C722",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C722"), Some("C722"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001DFDF,
            "vabs.q      C732, C732",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("C732"), Some("C732"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E0E0,
            "vabs.q      R020, R020",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R020"), Some("R020"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E1E1,
            "vabs.q      R021, R021",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R021"), Some("R021"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E2E2,
            "vabs.q      R022, R022",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R022"), Some("R022"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E3E3,
            "vabs.q      R023, R023",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R023"), Some("R023"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E4E4,
            "vabs.q      R120, R120",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R120"), Some("R120"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E5E5,
            "vabs.q      R121, R121",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R121"), Some("R121"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E6E6,
            "vabs.q      R122, R122",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R122"), Some("R122"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E7E7,
            "vabs.q      R123, R123",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R123"), Some("R123"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E8E8,
            "vabs.q      R220, R220",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R220"), Some("R220"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001E9E9,
            "vabs.q      R221, R221",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R221"), Some("R221"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EAEA,
            "vabs.q      R222, R222",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R222"), Some("R222"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EBEB,
            "vabs.q      R223, R223",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R223"), Some("R223"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001ECEC,
            "vabs.q      R320, R320",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R320"), Some("R320"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EDED,
            "vabs.q      R321, R321",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R321"), Some("R321"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EEEE,
            "vabs.q      R322, R322",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R322"), Some("R322"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001EFEF,
            "vabs.q      R323, R323",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R323"), Some("R323"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F0F0,
            "vabs.q      R420, R420",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R420"), Some("R420"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F1F1,
            "vabs.q      R421, R421",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R421"), Some("R421"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F2F2,
            "vabs.q      R422, R422",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R422"), Some("R422"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F3F3,
            "vabs.q      R423, R423",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R423"), Some("R423"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F4F4,
            "vabs.q      R520, R520",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R520"), Some("R520"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F5F5,
            "vabs.q      R521, R521",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R521"), Some("R521"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F6F6,
            "vabs.q      R522, R522",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R522"), Some("R522"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F7F7,
            "vabs.q      R523, R523",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R523"), Some("R523"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F8F8,
            "vabs.q      R620, R620",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R620"), Some("R620"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001F9F9,
            "vabs.q      R621, R621",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R621"), Some("R621"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FAFA,
            "vabs.q      R622, R622",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R622"), Some("R622"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FBFB,
            "vabs.q      R623, R623",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R623"), Some("R623"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FCFC,
            "vabs.q      R720, R720",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R720"), Some("R720"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FDFD,
            "vabs.q      R721, R721",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R721"), Some("R721"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FEFE,
            "vabs.q      R722, R722",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R722"), Some("R722"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD001FFFF,
            "vabs.q      R723, R723",
            Opcode::r4000allegrex_vabs_q,
            "vabs.q",
            [Some("R723"), Some("R723"), None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU0() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU0 */
        TestEntry::new_r4000allegrex(0x60001040, "vadd.s      S002, S400, S000", Opcode::r4000allegrex_vadd_s, "vadd.s", [Some("S002"), Some("S400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0x600010C0, "vadd.p      C002, C400, C000", Opcode::r4000allegrex_vadd_p, "vadd.p", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x60009040, "vadd.t      C001, C400, C000", Opcode::r4000allegrex_vadd_t, "vadd.t", [Some("C001"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x600090C0, "vadd.q      C002, C400, C000", Opcode::r4000allegrex_vadd_q, "vadd.q", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x60801040, "vsub.s      S002, S400, S000", Opcode::r4000allegrex_vsub_s, "vsub.s", [Some("S002"), Some("S400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0x608010C0, "vsub.p      C002, C400, C000", Opcode::r4000allegrex_vsub_p, "vsub.p", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x60809040, "vsub.t      C001, C400, C000", Opcode::r4000allegrex_vsub_t, "vsub.t", [Some("C001"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x608090C0, "vsub.q      C002, C400, C000", Opcode::r4000allegrex_vsub_q, "vsub.q", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x61001040, "vsbn.s      S002, S400, S000", Opcode::r4000allegrex_vsbn_s, "vsbn.s", [Some("S002"), Some("S400"), Some("S000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x610010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x610010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x61009040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x61009040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x610090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x610090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x61801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x61801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x618010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x618010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x61809040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x61809040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x618090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x618090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x62001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x62001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x620010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x620010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x62009040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x62009040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x620090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x620090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x62801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x62801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x628010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x628010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x62809040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x62809040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x628090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x628090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x63001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x63001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x630010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x630010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x63009040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x63009040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x630090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x630090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x63801040, "vdiv.s      S002, S400, S000", Opcode::r4000allegrex_vdiv_s, "vdiv.s", [Some("S002"), Some("S400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0x63801084, "vdiv.p      C100, C400, C000", Opcode::r4000allegrex_vdiv_p, "vdiv.p", [Some("C100"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x63809004, "vdiv.t      C100, C400, C000", Opcode::r4000allegrex_vdiv_t, "vdiv.t", [Some("C100"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x63809084, "vdiv.q      C100, C400, C000", Opcode::r4000allegrex_vdiv_q, "vdiv.q", [Some("C100"), Some("C400"), Some("C000"), None, None]),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU1() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU1 */
        TestEntry::new_r4000allegrex(0x64001040, "vmul.s      S002, S400, S000", Opcode::r4000allegrex_vmul_s, "vmul.s", [Some("S002"), Some("S400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0x640010C0, "vmul.p      C002, C400, C000", Opcode::r4000allegrex_vmul_p, "vmul.p", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x64009040, "vmul.t      C001, C400, C000", Opcode::r4000allegrex_vmul_t, "vmul.t", [Some("C001"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x640090C0, "vmul.q      C002, C400, C000", Opcode::r4000allegrex_vmul_q, "vmul.q", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x64801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x64801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x648010C0, "vdot.p      S002, C400, C000", Opcode::r4000allegrex_vdot_p, "vdot.p", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x64809040, "vdot.t      S002, C400, C000", Opcode::r4000allegrex_vdot_t, "vdot.t", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x648090C0, "vdot.q      S002, C400, C000", Opcode::r4000allegrex_vdot_q, "vdot.q", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x65001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x65001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x650010C0, "vscl.p      C002, C400, S000", Opcode::r4000allegrex_vscl_p, "vscl.p", [Some("C002"), Some("C400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0x65009040, "vscl.t      C001, C400, S000", Opcode::r4000allegrex_vscl_t, "vscl.t", [Some("C001"), Some("C400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0x650090C0, "vscl.q      C002, C400, S000", Opcode::r4000allegrex_vscl_q, "vscl.q", [Some("C002"), Some("C400"), Some("S000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x65801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x65801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x658010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x658010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x65809040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x65809040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x658090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x658090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x66001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x66001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x660010C0, "vhdp.p      S002, C400, C000", Opcode::r4000allegrex_vhdp_p, "vhdp.p", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x66009040, "vhdp.t      S002, C400, C000", Opcode::r4000allegrex_vhdp_t, "vhdp.t", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0x660090C0, "vhdp.q      S002, C400, C000", Opcode::r4000allegrex_vhdp_q, "vhdp.q", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x66801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x66801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x668010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x668010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x66809040, "vcrs.t      C001, C400, C000", Opcode::r4000allegrex_vcrs_t, "vcrs.t", [Some("C001"), Some("C400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x668090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x668090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x67001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x67001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x670010C0, "vdet.p      S002, C400, C000", Opcode::r4000allegrex_vdet_p, "vdet.p", [Some("S002"), Some("C400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x67009040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x67009040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x670090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x670090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x67801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x67801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x678010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x678010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x67809040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x67809040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x678090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x678090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU1> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },

    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU3() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU3 */
        TestEntry::new_r4000allegrex(0x6C401001, "vcmp.s      eq, S400, S002", Opcode::r4000allegrex_vcmp_s, "vcmp.s", [Some("eq, S400, S002"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0x6C401081, "vcmp.p      eq, C400, C002", Opcode::r4000allegrex_vcmp_p, "vcmp.p", [Some("eq, C400, C002"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0x6C409001, "vcmp.t      eq, C400, C001", Opcode::r4000allegrex_vcmp_t, "vcmp.t", [Some("eq, C400, C001"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0x6C409081, "vcmp.q      eq, C400, C002", Opcode::r4000allegrex_vcmp_q, "vcmp.q", [Some("eq, C400, C002"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0x6CC01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6CC01001                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x6CC01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6CC01081                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x6CC09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6CC09001                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x6CC09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6CC09081                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x6D401001, "vmin.s      S010, S400, S002", Opcode::r4000allegrex_vmin_s, "vmin.s", [Some("S010"), Some("S400"), Some("S002"), None, None]),
        TestEntry::new_r4000allegrex(0x6D401081, "vmin.p      C010, C400, C002", Opcode::r4000allegrex_vmin_p, "vmin.p", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6D409001, "vmin.t      C010, C400, C001", Opcode::r4000allegrex_vmin_t, "vmin.t", [Some("C010"), Some("C400"), Some("C001"), None, None]),
        TestEntry::new_r4000allegrex(0x6D409081, "vmin.q      C010, C400, C002", Opcode::r4000allegrex_vmin_q, "vmin.q", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6DC01001, "vmax.s      S010, S400, S002", Opcode::r4000allegrex_vmax_s, "vmax.s", [Some("S010"), Some("S400"), Some("S002"), None, None]),
        TestEntry::new_r4000allegrex(0x6DC01081, "vmax.p      C010, C400, C002", Opcode::r4000allegrex_vmax_p, "vmax.p", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6DC09001, "vmax.t      C010, C400, C001", Opcode::r4000allegrex_vmax_t, "vmax.t", [Some("C010"), Some("C400"), Some("C001"), None, None]),
        TestEntry::new_r4000allegrex(0x6DC09081, "vmax.q      C010, C400, C002", Opcode::r4000allegrex_vmax_q, "vmax.q", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0x6E401001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6E401001                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x6E401081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6E401081                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x6E409001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6E409001                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x6E409081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x6E409081                   /* INVALID / 00401001 <OpcodeCategory: R4000ALLEGREX_VFPU3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x6EC01001, "vscmp.s     S010, S400, S002", Opcode::r4000allegrex_vscmp_s, "vscmp.s", [Some("S010"), Some("S400"), Some("S002"), None, None]),
        TestEntry::new_r4000allegrex(0x6EC01081, "vscmp.p     C010, C400, C002", Opcode::r4000allegrex_vscmp_p, "vscmp.p", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6EC09001, "vscmp.t     C010, C400, C001", Opcode::r4000allegrex_vscmp_t, "vscmp.t", [Some("C010"), Some("C400"), Some("C001"), None, None]),
        TestEntry::new_r4000allegrex(0x6EC09081, "vscmp.q     C010, C400, C002", Opcode::r4000allegrex_vscmp_q, "vscmp.q", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6F401001, "vsge.s      S010, S400, S002", Opcode::r4000allegrex_vsge_s, "vsge.s", [Some("S010"), Some("S400"), Some("S002"), None, None]),
        TestEntry::new_r4000allegrex(0x6F401081, "vsge.p      C010, C400, C002", Opcode::r4000allegrex_vsge_p, "vsge.p", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6F409001, "vsge.t      C010, C400, C001", Opcode::r4000allegrex_vsge_t, "vsge.t", [Some("C010"), Some("C400"), Some("C001"), None, None]),
        TestEntry::new_r4000allegrex(0x6F409081, "vsge.q      C010, C400, C002", Opcode::r4000allegrex_vsge_q, "vsge.q", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6FC01001, "vslt.s      S010, S400, S002", Opcode::r4000allegrex_vslt_s, "vslt.s", [Some("S010"), Some("S400"), Some("S002"), None, None]),
        TestEntry::new_r4000allegrex(0x6FC01081, "vslt.p      C010, C400, C002", Opcode::r4000allegrex_vslt_p, "vslt.p", [Some("C010"), Some("C400"), Some("C002"), None, None]),
        TestEntry::new_r4000allegrex(0x6FC09001, "vslt.t      C010, C400, C001", Opcode::r4000allegrex_vslt_t, "vslt.t", [Some("C010"), Some("C400"), Some("C001"), None, None]),
        TestEntry::new_r4000allegrex(0x6FC09081, "vslt.q      C010, C400, C002", Opcode::r4000allegrex_vslt_q, "vslt.q", [Some("C010"), Some("C400"), Some("C002"), None, None]),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vcmp_all_conds() {
    const ENTRIES: &[TestEntry] = &[
        /* vcmp_all_conds */
        TestEntry::new_r4000allegrex(
            0x6C400800,
            "vcmp.s      fl, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("fl, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400880,
            "vcmp.p      fl, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("fl, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408800,
            "vcmp.t      fl, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("fl, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408880,
            "vcmp.q      fl, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("fl, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400801,
            "vcmp.s      eq, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("eq, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400881,
            "vcmp.p      eq, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("eq, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408801,
            "vcmp.t      eq, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("eq, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408881,
            "vcmp.q      eq, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("eq, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400802,
            "vcmp.s      lt, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("lt, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400882,
            "vcmp.p      lt, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("lt, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408802,
            "vcmp.t      lt, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("lt, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408882,
            "vcmp.q      lt, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("lt, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400803,
            "vcmp.s      le, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("le, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400883,
            "vcmp.p      le, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("le, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408803,
            "vcmp.t      le, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("le, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408883,
            "vcmp.q      le, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("le, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400804,
            "vcmp.s      tr, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("tr, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400884,
            "vcmp.p      tr, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("tr, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408804,
            "vcmp.t      tr, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("tr, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408884,
            "vcmp.q      tr, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("tr, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400805,
            "vcmp.s      ne, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ne, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400885,
            "vcmp.p      ne, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ne, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408805,
            "vcmp.t      ne, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ne, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408885,
            "vcmp.q      ne, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ne, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400806,
            "vcmp.s      ge, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ge, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400886,
            "vcmp.p      ge, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ge, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408806,
            "vcmp.t      ge, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ge, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408886,
            "vcmp.q      ge, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ge, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400807,
            "vcmp.s      gt, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("gt, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400887,
            "vcmp.p      gt, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("gt, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408807,
            "vcmp.t      gt, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("gt, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408887,
            "vcmp.q      gt, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("gt, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400808,
            "vcmp.s      ez, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ez, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400888,
            "vcmp.p      ez, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ez, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408808,
            "vcmp.t      ez, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ez, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408888,
            "vcmp.q      ez, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ez, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400809,
            "vcmp.s      en, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("en, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C400889,
            "vcmp.p      en, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("en, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408809,
            "vcmp.t      en, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("en, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C408889,
            "vcmp.q      en, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("en, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40080A,
            "vcmp.s      ei, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ei, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40088A,
            "vcmp.p      ei, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ei, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40880A,
            "vcmp.t      ei, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ei, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40888A,
            "vcmp.q      ei, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ei, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40080B,
            "vcmp.s      es, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("es, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40088B,
            "vcmp.p      es, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("es, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40880B,
            "vcmp.t      es, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("es, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40888B,
            "vcmp.q      es, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("es, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40080C,
            "vcmp.s      nz, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("nz, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40088C,
            "vcmp.p      nz, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("nz, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40880C,
            "vcmp.t      nz, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("nz, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40888C,
            "vcmp.q      nz, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("nz, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40080D,
            "vcmp.s      nn, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("nn, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40088D,
            "vcmp.p      nn, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("nn, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40880D,
            "vcmp.t      nn, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("nn, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40888D,
            "vcmp.q      nn, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("nn, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40080E,
            "vcmp.s      ni, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ni, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40088E,
            "vcmp.p      ni, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ni, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40880E,
            "vcmp.t      ni, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ni, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40888E,
            "vcmp.q      ni, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ni, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40080F,
            "vcmp.s      ns, S200, S002",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ns, S200, S002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40088F,
            "vcmp.p      ns, C200, C002",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ns, C200, C002"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40880F,
            "vcmp.t      ns, C200, C001",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ns, C200, C001"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C40888F,
            "vcmp.q      ns, C200, C002",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ns, C200, C002"), None, None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vcmp_zero() {
    const ENTRIES: &[TestEntry] = &[
        /* vcmp_zero */
        TestEntry::new_r4000allegrex(
            0x6C000000,
            "vcmp.s      fl",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("fl"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000080,
            "vcmp.p      fl",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("fl"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008000,
            "vcmp.t      fl",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("fl"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008080,
            "vcmp.q      fl",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("fl"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000001,
            "vcmp.s      eq, S000, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("eq, S000, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000081,
            "vcmp.p      eq, C000, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("eq, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008001,
            "vcmp.t      eq, C000, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("eq, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008081,
            "vcmp.q      eq, C000, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("eq, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000002,
            "vcmp.s      lt, S000, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("lt, S000, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000082,
            "vcmp.p      lt, C000, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("lt, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008002,
            "vcmp.t      lt, C000, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("lt, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008082,
            "vcmp.q      lt, C000, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("lt, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000003,
            "vcmp.s      le, S000, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("le, S000, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000083,
            "vcmp.p      le, C000, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("le, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008003,
            "vcmp.t      le, C000, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("le, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008083,
            "vcmp.q      le, C000, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("le, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000004,
            "vcmp.s      tr",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("tr"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000084,
            "vcmp.p      tr",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("tr"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008004,
            "vcmp.t      tr",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("tr"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008084,
            "vcmp.q      tr",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("tr"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000005,
            "vcmp.s      ne, S000, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ne, S000, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000085,
            "vcmp.p      ne, C000, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ne, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008005,
            "vcmp.t      ne, C000, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ne, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008085,
            "vcmp.q      ne, C000, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ne, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000006,
            "vcmp.s      ge, S000, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ge, S000, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000086,
            "vcmp.p      ge, C000, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ge, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008006,
            "vcmp.t      ge, C000, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ge, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008086,
            "vcmp.q      ge, C000, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ge, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000007,
            "vcmp.s      gt, S000, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("gt, S000, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000087,
            "vcmp.p      gt, C000, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("gt, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008007,
            "vcmp.t      gt, C000, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("gt, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008087,
            "vcmp.q      gt, C000, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("gt, C000, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000008,
            "vcmp.s      ez, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ez, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000088,
            "vcmp.p      ez, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ez, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008008,
            "vcmp.t      ez, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ez, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008088,
            "vcmp.q      ez, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ez, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000009,
            "vcmp.s      en, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("en, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C000089,
            "vcmp.p      en, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("en, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008009,
            "vcmp.t      en, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("en, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C008089,
            "vcmp.q      en, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("en, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00000A,
            "vcmp.s      ei, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ei, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00008A,
            "vcmp.p      ei, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ei, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00800A,
            "vcmp.t      ei, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ei, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00808A,
            "vcmp.q      ei, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ei, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00000B,
            "vcmp.s      es, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("es, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00008B,
            "vcmp.p      es, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("es, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00800B,
            "vcmp.t      es, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("es, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00808B,
            "vcmp.q      es, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("es, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00000C,
            "vcmp.s      nz, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("nz, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00008C,
            "vcmp.p      nz, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("nz, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00800C,
            "vcmp.t      nz, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("nz, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00808C,
            "vcmp.q      nz, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("nz, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00000D,
            "vcmp.s      nn, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("nn, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00008D,
            "vcmp.p      nn, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("nn, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00800D,
            "vcmp.t      nn, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("nn, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00808D,
            "vcmp.q      nn, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("nn, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00000E,
            "vcmp.s      ni, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ni, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00008E,
            "vcmp.p      ni, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ni, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00800E,
            "vcmp.t      ni, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ni, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00808E,
            "vcmp.q      ni, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ni, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00000F,
            "vcmp.s      ns, S000",
            Opcode::r4000allegrex_vcmp_s,
            "vcmp.s",
            [Some("ns, S000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00008F,
            "vcmp.p      ns, C000",
            Opcode::r4000allegrex_vcmp_p,
            "vcmp.p",
            [Some("ns, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00800F,
            "vcmp.t      ns, C000",
            Opcode::r4000allegrex_vcmp_t,
            "vcmp.t",
            [Some("ns, C000"), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0x6C00808F,
            "vcmp.q      ns, C000",
            Opcode::r4000allegrex_vcmp_q,
            "vcmp.q",
            [Some("ns, C000"), None, None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_COP2() {
    const ENTRIES: &[TestEntry] = &[
        /* COP2 */
        TestEntry::new_r4000allegrex(0x48640080, "mfvc        $a0, $128", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$128"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640081, "mfvc        $a0, $129", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$129"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640082, "mfvc        $a0, $130", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$130"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640083, "mfvc        $a0, $131", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$131"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640084, "mfvc        $a0, $132", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$132"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640085, "mfvc        $a0, $133", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$133"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640086, "mfvc        $a0, $134", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$134"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640087, "mfvc        $a0, $135", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$135"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640088, "mfvc        $a0, $136", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$136"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48640089, "mfvc        $a0, $137", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$137"), None, None, None]),
        TestEntry::new_r4000allegrex(0x4864008A, "mfvc        $a0, $138", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$138"), None, None, None]),
        TestEntry::new_r4000allegrex(0x4864008B, "mfvc        $a0, $139", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$139"), None, None, None]),
        TestEntry::new_r4000allegrex(0x4864008C, "mfvc        $a0, $140", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$140"), None, None, None]),
        TestEntry::new_r4000allegrex(0x4864008D, "mfvc        $a0, $141", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$141"), None, None, None]),
        TestEntry::new_r4000allegrex(0x4864008E, "mfvc        $a0, $142", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$142"), None, None, None]),
        TestEntry::new_r4000allegrex(0x4864008F, "mfvc        $a0, $143", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a0"), Some("$143"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0x48640090,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x48640090                   /* INVALID / 00040000 <OpcodeCategory: R4000ALLEGREX_COP2_MFHC2_P> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x48640091,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x48640091                   /* INVALID / 00040001 <OpcodeCategory: R4000ALLEGREX_COP2_MFHC2_P> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0x48640092,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0x48640092                   /* INVALID / 00040002 <OpcodeCategory: R4000ALLEGREX_COP2_MFHC2_P> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0x486000FF, "vsync2", Opcode::r4000allegrex_vsync2, "vsync2", [Some(""), None, None, None, None]),
        TestEntry::new_r4000allegrex(0x48650000, "mfv         $a1, S000", Opcode::r4000allegrex_mfv, "mfv", [Some("$a1"), Some("S000"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48650081, "mfvc        $a1, $129", Opcode::r4000allegrex_mfvc, "mfvc", [Some("$a1"), Some("$129"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48E50000, "mtv         $a1, S000", Opcode::r4000allegrex_mtv, "mtv", [Some("$a1"), Some("S000"), None, None, None]),
        TestEntry::new_r4000allegrex(0x48E50081, "mtvc        $a1, $129", Opcode::r4000allegrex_mtvc, "mtvc", [Some("$a1"), Some("$129"), None, None, None]),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU4_FMT0() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU4_FMT0 */
        TestEntry::new_r4000allegrex(0xD0001001, "vmov.s      S010, S400", Opcode::r4000allegrex_vmov_s, "vmov.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0001081, "vmov.p      C010, C400", Opcode::r4000allegrex_vmov_p, "vmov.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0009001, "vmov.t      C010, C400", Opcode::r4000allegrex_vmov_t, "vmov.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0009081, "vmov.q      C010, C400", Opcode::r4000allegrex_vmov_q, "vmov.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0011001, "vabs.s      S010, S400", Opcode::r4000allegrex_vabs_s, "vabs.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0011081, "vabs.p      C010, C400", Opcode::r4000allegrex_vabs_p, "vabs.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0019001, "vabs.t      C010, C400", Opcode::r4000allegrex_vabs_t, "vabs.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0019081, "vabs.q      C010, C400", Opcode::r4000allegrex_vabs_q, "vabs.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0021001, "vneg.s      S010, S400", Opcode::r4000allegrex_vneg_s, "vneg.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0021081, "vneg.p      C010, C400", Opcode::r4000allegrex_vneg_p, "vneg.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0029001, "vneg.t      C010, C400", Opcode::r4000allegrex_vneg_t, "vneg.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0029081, "vneg.q      C010, C400", Opcode::r4000allegrex_vneg_q, "vneg.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0031001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0031001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0031081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0031081                   /* vidt.p      C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vidt_p,
            opcode_str: "vidt.p",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0030081, "vidt.p      C010", Opcode::r4000allegrex_vidt_p, "vidt.p", [Some("C010"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0039001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0039001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0039081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0039081                   /* vidt.q      C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vidt_q,
            opcode_str: "vidt.q",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0038081, "vidt.q      C010", Opcode::r4000allegrex_vidt_q, "vidt.q", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0041001, "vsat0.s     S010, S400", Opcode::r4000allegrex_vsat0_s, "vsat0.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0041081, "vsat0.p     C010, C400", Opcode::r4000allegrex_vsat0_p, "vsat0.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0049001, "vsat0.t     C010, C400", Opcode::r4000allegrex_vsat0_t, "vsat0.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0049081, "vsat0.q     C010, C400", Opcode::r4000allegrex_vsat0_q, "vsat0.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0051001, "vsat1.s     S010, S400", Opcode::r4000allegrex_vsat1_s, "vsat1.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0051081, "vsat1.p     C010, C400", Opcode::r4000allegrex_vsat1_p, "vsat1.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0059001, "vsat1.t     C010, C400", Opcode::r4000allegrex_vsat1_t, "vsat1.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0059081, "vsat1.q     C010, C400", Opcode::r4000allegrex_vsat1_q, "vsat1.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0061001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0061001                   /* vzero.s     S010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vzero_s,
            opcode_str: "vzero.s",
            operands_str: [Some("S010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0061081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0061081                   /* vzero.p     C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vzero_p,
            opcode_str: "vzero.p",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0069001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0069001                   /* vzero.t     C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vzero_t,
            opcode_str: "vzero.t",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0069081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0069081                   /* vzero.q     C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vzero_q,
            opcode_str: "vzero.q",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0070001, "vone.s      S010", Opcode::r4000allegrex_vone_s, "vone.s", [Some("S010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0070081, "vone.p      C010", Opcode::r4000allegrex_vone_p, "vone.p", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0078001, "vone.t      C010", Opcode::r4000allegrex_vone_t, "vone.t", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0078081, "vone.q      C010", Opcode::r4000allegrex_vone_q, "vone.q", [Some("C010"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0081001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0081001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0081081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0081081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0089001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0089001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0089081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0089081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0091001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0091001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0091081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0091081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0099001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0099001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0099081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0099081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00A1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00A1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00A1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00A1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00A9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00A9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00A9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00A9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00B1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00B1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00B9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00B9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00C1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00C1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00C9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00C9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00C9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00D1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00D1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00D9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00D9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00E1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00E1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00E9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00E9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00F1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00F1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00F9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD00F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD00F9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0101001, "vrcp.s      S010, S400", Opcode::r4000allegrex_vrcp_s, "vrcp.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0101081, "vrcp.p      C010, C400", Opcode::r4000allegrex_vrcp_p, "vrcp.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0109001, "vrcp.t      C010, C400", Opcode::r4000allegrex_vrcp_t, "vrcp.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0109081, "vrcp.q      C010, C400", Opcode::r4000allegrex_vrcp_q, "vrcp.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0111001, "vrsq.s      S010, S400", Opcode::r4000allegrex_vrsq_s, "vrsq.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0111081, "vrsq.p      C010, C400", Opcode::r4000allegrex_vrsq_p, "vrsq.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0119001, "vrsq.t      C010, C400", Opcode::r4000allegrex_vrsq_t, "vrsq.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0119081, "vrsq.q      C010, C400", Opcode::r4000allegrex_vrsq_q, "vrsq.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0121001, "vsin.s      S010, S400", Opcode::r4000allegrex_vsin_s, "vsin.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0121081, "vsin.p      C010, C400", Opcode::r4000allegrex_vsin_p, "vsin.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0129001, "vsin.t      C010, C400", Opcode::r4000allegrex_vsin_t, "vsin.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0129081, "vsin.q      C010, C400", Opcode::r4000allegrex_vsin_q, "vsin.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0131001, "vcos.s      S010, S400", Opcode::r4000allegrex_vcos_s, "vcos.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0131081, "vcos.p      C010, C400", Opcode::r4000allegrex_vcos_p, "vcos.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0139001, "vcos.t      C010, C400", Opcode::r4000allegrex_vcos_t, "vcos.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0139081, "vcos.q      C010, C400", Opcode::r4000allegrex_vcos_q, "vcos.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0141001, "vexp2.s     S010, S400", Opcode::r4000allegrex_vexp2_s, "vexp2.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0141081, "vexp2.p     C010, C400", Opcode::r4000allegrex_vexp2_p, "vexp2.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0149001, "vexp2.t     C010, C400", Opcode::r4000allegrex_vexp2_t, "vexp2.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0149081, "vexp2.q     C010, C400", Opcode::r4000allegrex_vexp2_q, "vexp2.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0151001, "vlog2.s     S010, S400", Opcode::r4000allegrex_vlog2_s, "vlog2.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0151081, "vlog2.p     C010, C400", Opcode::r4000allegrex_vlog2_p, "vlog2.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0159001, "vlog2.t     C010, C400", Opcode::r4000allegrex_vlog2_t, "vlog2.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0159081, "vlog2.q     C010, C400", Opcode::r4000allegrex_vlog2_q, "vlog2.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0161001, "vsqrt.s     S010, S400", Opcode::r4000allegrex_vsqrt_s, "vsqrt.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0161081, "vsqrt.p     C010, C400", Opcode::r4000allegrex_vsqrt_p, "vsqrt.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0169001, "vsqrt.t     C010, C400", Opcode::r4000allegrex_vsqrt_t, "vsqrt.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0169081, "vsqrt.q     C010, C400", Opcode::r4000allegrex_vsqrt_q, "vsqrt.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0171001, "vasin.s     S010, S400", Opcode::r4000allegrex_vasin_s, "vasin.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0171081, "vasin.p     C010, C400", Opcode::r4000allegrex_vasin_p, "vasin.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0179001, "vasin.t     C010, C400", Opcode::r4000allegrex_vasin_t, "vasin.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0179081, "vasin.q     C010, C400", Opcode::r4000allegrex_vasin_q, "vasin.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0181001, "vnrcp.s     S010, S400", Opcode::r4000allegrex_vnrcp_s, "vnrcp.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0181081, "vnrcp.p     C010, C400", Opcode::r4000allegrex_vnrcp_p, "vnrcp.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0189001, "vnrcp.t     C010, C400", Opcode::r4000allegrex_vnrcp_t, "vnrcp.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0189081, "vnrcp.q     C010, C400", Opcode::r4000allegrex_vnrcp_q, "vnrcp.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0191001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0191001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0191081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0191081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0199001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0199001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0199081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0199081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD01A1001, "vnsin.s     S010, S400", Opcode::r4000allegrex_vnsin_s, "vnsin.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD01A1081, "vnsin.p     C010, C400", Opcode::r4000allegrex_vnsin_p, "vnsin.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD01A9001, "vnsin.t     C010, C400", Opcode::r4000allegrex_vnsin_t, "vnsin.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD01A9081, "vnsin.q     C010, C400", Opcode::r4000allegrex_vnsin_q, "vnsin.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD01B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01B1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01B1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01B9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01B9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD01C1001, "vrexp2.s    S010, S400", Opcode::r4000allegrex_vrexp2_s, "vrexp2.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD01C1081, "vrexp2.p    C010, C400", Opcode::r4000allegrex_vrexp2_p, "vrexp2.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD01C9001, "vrexp2.t    C010, C400", Opcode::r4000allegrex_vrexp2_t, "vrexp2.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD01C9081, "vrexp2.q    C010, C400", Opcode::r4000allegrex_vrexp2_q, "vrexp2.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD01D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01D1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01D1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01D9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01D9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01E1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01E1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01E9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01E9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01F1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01F1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01F9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD01F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD01F9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT3> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0201000, "vrnds.s     S400", Opcode::r4000allegrex_vrnds_s, "vrnds.s", [Some("S400"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0201001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0201001                   /* vrnds.s     S400 / 00000001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrnds_s,
            opcode_str: "vrnds.s",
            operands_str: [Some("S400"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0201081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0201081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0209001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0209001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0209081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0209081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0210001, "vrndi.s     S010", Opcode::r4000allegrex_vrndi_s, "vrndi.s", [Some("S010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0210081, "vrndi.p     C010", Opcode::r4000allegrex_vrndi_p, "vrndi.p", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0218001, "vrndi.t     C010", Opcode::r4000allegrex_vrndi_t, "vrndi.t", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0218081, "vrndi.q     C010", Opcode::r4000allegrex_vrndi_q, "vrndi.q", [Some("C010"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0211001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0211001                   /* vrndi.s     S010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndi_s,
            opcode_str: "vrndi.s",
            operands_str: [Some("S010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0211081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0211081                   /* vrndi.p     C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndi_p,
            opcode_str: "vrndi.p",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0219001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0219001                   /* vrndi.t     C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndi_t,
            opcode_str: "vrndi.t",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0219081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0219081                   /* vrndi.q     C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndi_q,
            opcode_str: "vrndi.q",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0220001, "vrndf1.s    S010", Opcode::r4000allegrex_vrndf1_s, "vrndf1.s", [Some("S010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0220081, "vrndf1.p    C010", Opcode::r4000allegrex_vrndf1_p, "vrndf1.p", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0228001, "vrndf1.t    C010", Opcode::r4000allegrex_vrndf1_t, "vrndf1.t", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0228081, "vrndf1.q    C010", Opcode::r4000allegrex_vrndf1_q, "vrndf1.q", [Some("C010"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0221001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0221001                   /* vrndf1.s    S010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf1_s,
            opcode_str: "vrndf1.s",
            operands_str: [Some("S010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0221081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0221081                   /* vrndf1.p    C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf1_p,
            opcode_str: "vrndf1.p",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0229001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0229001                   /* vrndf1.t    C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf1_t,
            opcode_str: "vrndf1.t",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0229081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0229081                   /* vrndf1.q    C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf1_q,
            opcode_str: "vrndf1.q",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0230001, "vrndf2.s    S010", Opcode::r4000allegrex_vrndf2_s, "vrndf2.s", [Some("S010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0230081, "vrndf2.p    C010", Opcode::r4000allegrex_vrndf2_p, "vrndf2.p", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0238001, "vrndf2.t    C010", Opcode::r4000allegrex_vrndf2_t, "vrndf2.t", [Some("C010"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xD0238081, "vrndf2.q    C010", Opcode::r4000allegrex_vrndf2_q, "vrndf2.q", [Some("C010"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0231001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0231001                   /* vrndf2.s    S010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf2_s,
            opcode_str: "vrndf2.s",
            operands_str: [Some("S010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0231081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0231081                   /* vrndf2.p    C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf2_p,
            opcode_str: "vrndf2.p",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0239001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0239001                   /* vrndf2.t    C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf2_t,
            opcode_str: "vrndf2.t",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0239081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0239081                   /* vrndf2.q    C010 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::r4000allegrex_vrndf2_q,
            opcode_str: "vrndf2.q",
            operands_str: [Some("C010"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0241001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0241001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0241081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0241081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0249001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0249001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0249081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0249081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0251001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0251001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0251081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0251081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0259001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0259001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0259081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0259081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0261001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0261001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0261081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0261081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0269001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0269001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0269081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0269081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0271001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0271001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0271081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0271081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0279001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0279001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0279081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0279081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_RND> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0281001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0281001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0281081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0281081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0289001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0289001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0289081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0289081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0291001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0291001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0291081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0291081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0299001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0299001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0299081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0299081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02A1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02A1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02A1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02A1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02A9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02A9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02A9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02A9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02B1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02B1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02B9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02B9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02C1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02C1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02C9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02C9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02C9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02D1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02D1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02D9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02D9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02E1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02E1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02E9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02E9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02F1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02F1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02F9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD02F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD02F9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0301001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0301001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0301081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0301081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0309001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0309001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0309081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0309081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0311001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0311001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0311081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0311081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0319001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0319001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0319081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0319081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0321001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0321001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0321081, "vf2h.p      S010, C400", Opcode::r4000allegrex_vf2h_p, "vf2h.p", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0329001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0329001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0329081, "vf2h.q      C010, C400", Opcode::r4000allegrex_vf2h_q, "vf2h.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0331001, "vh2f.s      C010, S400", Opcode::r4000allegrex_vh2f_s, "vh2f.s", [Some("C010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0331081, "vh2f.p      C010, C400", Opcode::r4000allegrex_vh2f_p, "vh2f.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0339001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0339001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0339081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0339081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0341001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0341001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0341081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0341081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0349001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0349001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0349081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0349081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0351001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0351001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0351081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0351081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0359001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0359001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0359081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0359081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0361001, "vsbz.s      S010, S400", Opcode::r4000allegrex_vsbz_s, "vsbz.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0361081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0361081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0369001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0369001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0369081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0369081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0371001, "vlgb.s      S010, S400", Opcode::r4000allegrex_vlgb_s, "vlgb.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0371081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0371081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0379001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0379001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0379081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0379081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTFLT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0381001, "vuc2ifs.s   C010, S400", Opcode::r4000allegrex_vuc2ifs_s, "vuc2ifs.s", [Some("C010"), Some("S400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0381081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0381081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0389001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0389001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0389081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0389081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0391001, "vc2i.s      C010, S400", Opcode::r4000allegrex_vc2i_s, "vc2i.s", [Some("C010"), Some("S400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0391081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0391081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0399001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0399001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0399081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0399081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03A1001, "vus2i.s     C010, S400", Opcode::r4000allegrex_vus2i_s, "vus2i.s", [Some("C010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD03A1081, "vus2i.p     C010, C400", Opcode::r4000allegrex_vus2i_p, "vus2i.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03A9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03A9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03A9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03A9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03B1001, "vs2i.s      C010, S400", Opcode::r4000allegrex_vs2i_s, "vs2i.s", [Some("C010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD03B1081, "vs2i.p      C010, C400", Opcode::r4000allegrex_vs2i_p, "vs2i.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03B9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03B9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03C1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03C1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03C9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03C9081, "vi2uc.q     S010, C400", Opcode::r4000allegrex_vi2uc_q, "vi2uc.q", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03D1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03D1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD03D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03D9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03D9081, "vi2c.q      S010, C400", Opcode::r4000allegrex_vi2c_q, "vi2c.q", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03E1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03E1081, "vi2us.p     S010, C400", Opcode::r4000allegrex_vi2us_p, "vi2us.p", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03E9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03E9081, "vi2us.q     C010, C400", Opcode::r4000allegrex_vi2us_q, "vi2us.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03F1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03F1081, "vi2s.p      S010, C400", Opcode::r4000allegrex_vi2s_p, "vi2s.p", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD03F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD03F9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CVTINT> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD03F9081, "vi2s.q      C010, C400", Opcode::r4000allegrex_vi2s_q, "vi2s.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0401001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0401001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0401081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0401081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0409001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0409001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0409081, "vsrt1.q     C010, C400", Opcode::r4000allegrex_vsrt1_q, "vsrt1.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0411001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0411001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0411081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0411081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0419001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0419001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0419081, "vsrt2.q     C010, C400", Opcode::r4000allegrex_vsrt2_q, "vsrt2.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0421001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0421001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0421081, "vbfy1.p     C010, C400", Opcode::r4000allegrex_vbfy1_p, "vbfy1.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0429001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0429001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0429081, "vbfy1.q     C010, C400", Opcode::r4000allegrex_vbfy1_q, "vbfy1.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0431001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0431001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0431081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0431081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0439001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0439001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0439081, "vbfy2.q     C010, C400", Opcode::r4000allegrex_vbfy2_q, "vbfy2.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0441001, "vocp.s      S010, S400", Opcode::r4000allegrex_vocp_s, "vocp.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0441081, "vocp.p      C010, C400", Opcode::r4000allegrex_vocp_p, "vocp.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0449001, "vocp.t      C010, C400", Opcode::r4000allegrex_vocp_t, "vocp.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0449081, "vocp.q      C010, C400", Opcode::r4000allegrex_vocp_q, "vocp.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0451001, "vsocp.s     C010, S400", Opcode::r4000allegrex_vsocp_s, "vsocp.s", [Some("C010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0451081, "vsocp.p     C010, C400", Opcode::r4000allegrex_vsocp_p, "vsocp.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
                instr: Instruction::new(
                0xD0459001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0459001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0459081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0459081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0461001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0461001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry::new_r4000allegrex(0xD0461081, "vfad.p      S010, C400", Opcode::r4000allegrex_vfad_p, "vfad.p", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0469001, "vfad.t      S010, C400", Opcode::r4000allegrex_vfad_t, "vfad.t", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0469081, "vfad.q      S010, C400", Opcode::r4000allegrex_vfad_q, "vfad.q", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
                instr: Instruction::new(
                0xD0471001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0471001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT8> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry::new_r4000allegrex(0xD0471081, "vavg.p      S010, C400", Opcode::r4000allegrex_vavg_p, "vavg.p", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0479001, "vavg.t      S010, C400", Opcode::r4000allegrex_vavg_t, "vavg.t", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0479081, "vavg.q      S010, C400", Opcode::r4000allegrex_vavg_q, "vavg.q", [Some("S010"), Some("C400"), None, None, None]),
        TestEntry {
                instr: Instruction::new(
                0xD0481001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0481001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0481081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0481081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0489001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0489001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry::new_r4000allegrex(0xD0489081, "vsrt3.q     C010, C400", Opcode::r4000allegrex_vsrt3_q, "vsrt3.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
                instr: Instruction::new(
                0xD0491001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0491001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0491081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0491081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
            instr: Instruction::new(
                0xD0499001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0499001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0499081, "vsrt4.q     C010, C400", Opcode::r4000allegrex_vsrt4_q, "vsrt4.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD04A1001, "vsgn.s      S010, S400", Opcode::r4000allegrex_vsgn_s, "vsgn.s", [Some("S010"), Some("S400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD04A1081, "vsgn.p      C010, C400", Opcode::r4000allegrex_vsgn_p, "vsgn.p", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD04A9001, "vsgn.t      C010, C400", Opcode::r4000allegrex_vsgn_t, "vsgn.t", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD04A9081, "vsgn.q      C010, C400", Opcode::r4000allegrex_vsgn_q, "vsgn.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD04B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04B1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04B1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04B9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04B9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04C1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04C1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04C9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04C9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04C9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04D1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04D1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04D9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04D9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04E1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04E1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04E9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04E9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04F1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04F1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04F9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD04F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD04F9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_FMT9> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0508001, "vmfvc       S010, $128", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$128"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508101, "vmfvc       S010, $129", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$129"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508201, "vmfvc       S010, $130", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$130"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508301, "vmfvc       S010, $131", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$131"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508401, "vmfvc       S010, $132", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$132"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508501, "vmfvc       S010, $133", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$133"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508601, "vmfvc       S010, $134", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$134"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508701, "vmfvc       S010, $135", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$135"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508801, "vmfvc       S010, $136", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$136"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508901, "vmfvc       S010, $137", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$137"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508A01, "vmfvc       S010, $138", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$138"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508B01, "vmfvc       S010, $139", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$139"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508C01, "vmfvc       S010, $140", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$140"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508D01, "vmfvc       S010, $141", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$141"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508E01, "vmfvc       S010, $142", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$142"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0508F01, "vmfvc       S010, $143", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$143"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509001, "vmfvc       S010, $144", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$144"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509101, "vmfvc       S010, $145", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$145"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509201, "vmfvc       S010, $146", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$146"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509301, "vmfvc       S010, $147", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$147"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509401, "vmfvc       S010, $148", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$148"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509501, "vmfvc       S010, $149", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$149"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509601, "vmfvc       S010, $150", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$150"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509701, "vmfvc       S010, $151", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$151"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509801, "vmfvc       S010, $152", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$152"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509901, "vmfvc       S010, $153", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$153"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509A01, "vmfvc       S010, $154", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$154"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509B01, "vmfvc       S010, $155", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$155"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509C01, "vmfvc       S010, $156", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$156"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509D01, "vmfvc       S010, $157", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$157"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509E01, "vmfvc       S010, $158", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$158"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0509F01, "vmfvc       S010, $159", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$159"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A001, "vmfvc       S010, $160", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$160"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A101, "vmfvc       S010, $161", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$161"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A201, "vmfvc       S010, $162", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$162"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A301, "vmfvc       S010, $163", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$163"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A401, "vmfvc       S010, $164", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$164"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A501, "vmfvc       S010, $165", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$165"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A601, "vmfvc       S010, $166", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$166"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A701, "vmfvc       S010, $167", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$167"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A801, "vmfvc       S010, $168", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$168"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050A901, "vmfvc       S010, $169", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$169"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050AA01, "vmfvc       S010, $170", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$170"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050AB01, "vmfvc       S010, $171", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$171"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050AC01, "vmfvc       S010, $172", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$172"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050AD01, "vmfvc       S010, $173", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$173"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050AE01, "vmfvc       S010, $174", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$174"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050AF01, "vmfvc       S010, $175", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$175"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B001, "vmfvc       S010, $176", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$176"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B101, "vmfvc       S010, $177", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$177"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B201, "vmfvc       S010, $178", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$178"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B301, "vmfvc       S010, $179", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$179"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B401, "vmfvc       S010, $180", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$180"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B501, "vmfvc       S010, $181", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$181"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B601, "vmfvc       S010, $182", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$182"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B701, "vmfvc       S010, $183", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$183"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B801, "vmfvc       S010, $184", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$184"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050B901, "vmfvc       S010, $185", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$185"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050BA01, "vmfvc       S010, $186", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$186"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050BB01, "vmfvc       S010, $187", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$187"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050BC01, "vmfvc       S010, $188", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$188"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050BD01, "vmfvc       S010, $189", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$189"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050BE01, "vmfvc       S010, $190", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$190"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050BF01, "vmfvc       S010, $191", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$191"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C001, "vmfvc       S010, $192", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$192"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C101, "vmfvc       S010, $193", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$193"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C201, "vmfvc       S010, $194", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$194"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C301, "vmfvc       S010, $195", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$195"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C401, "vmfvc       S010, $196", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$196"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C501, "vmfvc       S010, $197", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$197"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C601, "vmfvc       S010, $198", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$198"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C701, "vmfvc       S010, $199", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$199"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C801, "vmfvc       S010, $200", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$200"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050C901, "vmfvc       S010, $201", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$201"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050CA01, "vmfvc       S010, $202", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$202"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050CB01, "vmfvc       S010, $203", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$203"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050CC01, "vmfvc       S010, $204", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$204"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050CD01, "vmfvc       S010, $205", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$205"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050CE01, "vmfvc       S010, $206", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$206"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050CF01, "vmfvc       S010, $207", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$207"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D001, "vmfvc       S010, $208", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$208"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D101, "vmfvc       S010, $209", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$209"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D201, "vmfvc       S010, $210", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$210"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D301, "vmfvc       S010, $211", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$211"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D401, "vmfvc       S010, $212", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$212"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D501, "vmfvc       S010, $213", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$213"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D601, "vmfvc       S010, $214", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$214"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D701, "vmfvc       S010, $215", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$215"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D801, "vmfvc       S010, $216", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$216"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050D901, "vmfvc       S010, $217", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$217"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050DA01, "vmfvc       S010, $218", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$218"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050DB01, "vmfvc       S010, $219", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$219"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050DC01, "vmfvc       S010, $220", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$220"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050DD01, "vmfvc       S010, $221", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$221"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050DE01, "vmfvc       S010, $222", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$222"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050DF01, "vmfvc       S010, $223", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$223"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E001, "vmfvc       S010, $224", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$224"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E101, "vmfvc       S010, $225", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$225"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E201, "vmfvc       S010, $226", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$226"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E301, "vmfvc       S010, $227", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$227"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E401, "vmfvc       S010, $228", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$228"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E501, "vmfvc       S010, $229", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$229"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E601, "vmfvc       S010, $230", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$230"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E701, "vmfvc       S010, $231", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$231"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E801, "vmfvc       S010, $232", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$232"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050E901, "vmfvc       S010, $233", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$233"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050EA01, "vmfvc       S010, $234", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$234"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050EB01, "vmfvc       S010, $235", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$235"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050EC01, "vmfvc       S010, $236", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$236"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050ED01, "vmfvc       S010, $237", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$237"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050EE01, "vmfvc       S010, $238", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$238"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050EF01, "vmfvc       S010, $239", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$239"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F001, "vmfvc       S010, $240", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$240"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F101, "vmfvc       S010, $241", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$241"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F201, "vmfvc       S010, $242", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$242"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F301, "vmfvc       S010, $243", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$243"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F401, "vmfvc       S010, $244", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$244"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F501, "vmfvc       S010, $245", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$245"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F601, "vmfvc       S010, $246", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$246"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F701, "vmfvc       S010, $247", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$247"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F801, "vmfvc       S010, $248", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$248"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050F901, "vmfvc       S010, $249", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$249"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050FA01, "vmfvc       S010, $250", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$250"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050FB01, "vmfvc       S010, $251", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$251"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050FC01, "vmfvc       S010, $252", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$252"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050FD01, "vmfvc       S010, $253", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$253"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050FE01, "vmfvc       S010, $254", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$254"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD050FF01, "vmfvc       S010, $255", Opcode::r4000allegrex_vmfvc, "vmfvc", [Some("S010"), Some("$255"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510180, "vmtvc       $128, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$128"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510181, "vmtvc       $129, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$129"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510182, "vmtvc       $130, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$130"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510183, "vmtvc       $131, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$131"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510184, "vmtvc       $132, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$132"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510185, "vmtvc       $133, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$133"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510186, "vmtvc       $134, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$134"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510187, "vmtvc       $135, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$135"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510188, "vmtvc       $136, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$136"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510189, "vmtvc       $137, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$137"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051018A, "vmtvc       $138, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$138"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051018B, "vmtvc       $139, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$139"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051018C, "vmtvc       $140, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$140"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051018D, "vmtvc       $141, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$141"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051018E, "vmtvc       $142, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$142"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051018F, "vmtvc       $143, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$143"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510190, "vmtvc       $144, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$144"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510191, "vmtvc       $145, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$145"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510192, "vmtvc       $146, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$146"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510193, "vmtvc       $147, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$147"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510194, "vmtvc       $148, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$148"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510195, "vmtvc       $149, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$149"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510196, "vmtvc       $150, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$150"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510197, "vmtvc       $151, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$151"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510198, "vmtvc       $152, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$152"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0510199, "vmtvc       $153, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$153"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051019A, "vmtvc       $154, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$154"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051019B, "vmtvc       $155, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$155"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051019C, "vmtvc       $156, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$156"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051019D, "vmtvc       $157, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$157"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051019E, "vmtvc       $158, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$158"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD051019F, "vmtvc       $159, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$159"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A0, "vmtvc       $160, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$160"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A1, "vmtvc       $161, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$161"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A2, "vmtvc       $162, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$162"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A3, "vmtvc       $163, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$163"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A4, "vmtvc       $164, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$164"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A5, "vmtvc       $165, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$165"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A6, "vmtvc       $166, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$166"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A7, "vmtvc       $167, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$167"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A8, "vmtvc       $168, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$168"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101A9, "vmtvc       $169, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$169"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101AA, "vmtvc       $170, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$170"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101AB, "vmtvc       $171, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$171"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101AC, "vmtvc       $172, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$172"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101AD, "vmtvc       $173, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$173"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101AE, "vmtvc       $174, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$174"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101AF, "vmtvc       $175, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$175"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B0, "vmtvc       $176, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$176"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B1, "vmtvc       $177, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$177"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B2, "vmtvc       $178, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$178"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B3, "vmtvc       $179, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$179"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B4, "vmtvc       $180, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$180"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B5, "vmtvc       $181, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$181"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B6, "vmtvc       $182, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$182"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B7, "vmtvc       $183, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$183"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B8, "vmtvc       $184, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$184"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101B9, "vmtvc       $185, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$185"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101BA, "vmtvc       $186, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$186"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101BB, "vmtvc       $187, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$187"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101BC, "vmtvc       $188, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$188"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101BD, "vmtvc       $189, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$189"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101BE, "vmtvc       $190, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$190"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101BF, "vmtvc       $191, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$191"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C0, "vmtvc       $192, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$192"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C1, "vmtvc       $193, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$193"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C2, "vmtvc       $194, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$194"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C3, "vmtvc       $195, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$195"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C4, "vmtvc       $196, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$196"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C5, "vmtvc       $197, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$197"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C6, "vmtvc       $198, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$198"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C7, "vmtvc       $199, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$199"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C8, "vmtvc       $200, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$200"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101C9, "vmtvc       $201, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$201"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101CA, "vmtvc       $202, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$202"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101CB, "vmtvc       $203, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$203"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101CC, "vmtvc       $204, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$204"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101CD, "vmtvc       $205, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$205"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101CE, "vmtvc       $206, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$206"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101CF, "vmtvc       $207, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$207"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D0, "vmtvc       $208, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$208"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D1, "vmtvc       $209, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$209"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D2, "vmtvc       $210, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$210"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D3, "vmtvc       $211, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$211"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D4, "vmtvc       $212, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$212"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D5, "vmtvc       $213, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$213"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D6, "vmtvc       $214, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$214"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D7, "vmtvc       $215, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$215"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D8, "vmtvc       $216, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$216"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101D9, "vmtvc       $217, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$217"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101DA, "vmtvc       $218, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$218"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101DB, "vmtvc       $219, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$219"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101DC, "vmtvc       $220, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$220"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101DD, "vmtvc       $221, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$221"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101DE, "vmtvc       $222, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$222"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101DF, "vmtvc       $223, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$223"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E0, "vmtvc       $224, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$224"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E1, "vmtvc       $225, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$225"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E2, "vmtvc       $226, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$226"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E3, "vmtvc       $227, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$227"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E4, "vmtvc       $228, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$228"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E5, "vmtvc       $229, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$229"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E6, "vmtvc       $230, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$230"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E7, "vmtvc       $231, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$231"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E8, "vmtvc       $232, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$232"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101E9, "vmtvc       $233, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$233"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101EA, "vmtvc       $234, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$234"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101EB, "vmtvc       $235, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$235"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101EC, "vmtvc       $236, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$236"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101ED, "vmtvc       $237, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$237"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101EE, "vmtvc       $238, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$238"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101EF, "vmtvc       $239, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$239"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F0, "vmtvc       $240, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$240"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F1, "vmtvc       $241, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$241"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F2, "vmtvc       $242, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$242"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F3, "vmtvc       $243, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$243"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F4, "vmtvc       $244, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$244"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F5, "vmtvc       $245, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$245"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F6, "vmtvc       $246, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$246"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F7, "vmtvc       $247, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$247"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F8, "vmtvc       $248, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$248"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101F9, "vmtvc       $249, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$249"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101FA, "vmtvc       $250, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$250"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101FB, "vmtvc       $251, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$251"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101FC, "vmtvc       $252, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$252"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101FD, "vmtvc       $253, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$253"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101FE, "vmtvc       $254, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$254"), Some("S010"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD05101FF, "vmtvc       $255, S010", Opcode::r4000allegrex_vmtvc, "vmtvc", [Some("$255"), Some("S010"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0521001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0521001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0521081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0521081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0529001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0529001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0529081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0529081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0531001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0531001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0531081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0531081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0539001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0539001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0539081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0539081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0541001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0541001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0541081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0541081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0549001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0549001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0549081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0549081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0551001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0551001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0551081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0551081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0559001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0559001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0559081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0559081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0561001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0561001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0561081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0561081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0569001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0569001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0569081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0569081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0571001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0571001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0571081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0571081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0579001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0579001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0579081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0579081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_CONTROL> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0581001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0581001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0581081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0581081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0589001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0589001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0589081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0589081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0591001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0591001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0591081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0591081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0599001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0599001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0599081, "vt4444.q    C010, C400", Opcode::r4000allegrex_vt4444_q, "vt4444.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD05A1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05A1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05A1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05A1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05A9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05A9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD05A9081, "vt5551.q    C010, C400", Opcode::r4000allegrex_vt5551_q, "vt5551.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD05B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05B1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05B1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05B9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD05B9081, "vt5650.q    C010, C400", Opcode::r4000allegrex_vt5650_q, "vt5650.q", [Some("C010"), Some("C400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD05C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05C1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05C1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05C9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05C9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05C9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05D1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05D1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05D9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05D9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05E1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05E1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05E9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05E9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05F1001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05F1081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05F9001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD05F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD05F9081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0_COLOR> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xD0600001, "vcst.s      S010, INVALID_0", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_0"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0600081, "vcst.p      C010, INVALID_0", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_0"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0608001, "vcst.t      C010, INVALID_0", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_0"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0608081, "vcst.q      C010, INVALID_0", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_0"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0610001, "vcst.s      S010, VFPU_HUGE", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_HUGE"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0610081, "vcst.p      C010, VFPU_HUGE", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_HUGE"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0618001, "vcst.t      C010, VFPU_HUGE", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_HUGE"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0618081, "vcst.q      C010, VFPU_HUGE", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_HUGE"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0620001, "vcst.s      S010, VFPU_SQRT2", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_SQRT2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0620081, "vcst.p      C010, VFPU_SQRT2", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_SQRT2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0628001, "vcst.t      C010, VFPU_SQRT2", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_SQRT2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0628081, "vcst.q      C010, VFPU_SQRT2", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_SQRT2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0630001, "vcst.s      S010, VFPU_SQRT1_2", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_SQRT1_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0630081, "vcst.p      C010, VFPU_SQRT1_2", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_SQRT1_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0638001, "vcst.t      C010, VFPU_SQRT1_2", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_SQRT1_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0638081, "vcst.q      C010, VFPU_SQRT1_2", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_SQRT1_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0640001, "vcst.s      S010, VFPU_2_SQRTPI", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_2_SQRTPI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0640081, "vcst.p      C010, VFPU_2_SQRTPI", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_2_SQRTPI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0648001, "vcst.t      C010, VFPU_2_SQRTPI", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_2_SQRTPI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0648081, "vcst.q      C010, VFPU_2_SQRTPI", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_2_SQRTPI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0650001, "vcst.s      S010, VFPU_2_PI", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_2_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0650081, "vcst.p      C010, VFPU_2_PI", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_2_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0658001, "vcst.t      C010, VFPU_2_PI", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_2_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0658081, "vcst.q      C010, VFPU_2_PI", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_2_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0660001, "vcst.s      S010, VFPU_1_PI", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_1_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0660081, "vcst.p      C010, VFPU_1_PI", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_1_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0668001, "vcst.t      C010, VFPU_1_PI", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_1_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0668081, "vcst.q      C010, VFPU_1_PI", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_1_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0670001, "vcst.s      S010, VFPU_PI_4", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_PI_4"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0670081, "vcst.p      C010, VFPU_PI_4", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_PI_4"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0678001, "vcst.t      C010, VFPU_PI_4", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_PI_4"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0678081, "vcst.q      C010, VFPU_PI_4", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_PI_4"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0680001, "vcst.s      S010, VFPU_PI_2", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_PI_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0680081, "vcst.p      C010, VFPU_PI_2", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_PI_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0688001, "vcst.t      C010, VFPU_PI_2", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_PI_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0688081, "vcst.q      C010, VFPU_PI_2", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_PI_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0690001, "vcst.s      S010, VFPU_PI", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0690081, "vcst.p      C010, VFPU_PI", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0698001, "vcst.t      C010, VFPU_PI", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0698081, "vcst.q      C010, VFPU_PI", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06A0001, "vcst.s      S010, VFPU_E", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06A0081, "vcst.p      C010, VFPU_E", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06A8001, "vcst.t      C010, VFPU_E", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06A8081, "vcst.q      C010, VFPU_E", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06B0001, "vcst.s      S010, VFPU_LOG2E", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_LOG2E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06B0081, "vcst.p      C010, VFPU_LOG2E", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_LOG2E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06B8001, "vcst.t      C010, VFPU_LOG2E", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_LOG2E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06B8081, "vcst.q      C010, VFPU_LOG2E", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_LOG2E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06C0001, "vcst.s      S010, VFPU_LOG10E", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_LOG10E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06C0081, "vcst.p      C010, VFPU_LOG10E", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_LOG10E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06C8001, "vcst.t      C010, VFPU_LOG10E", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_LOG10E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06C8081, "vcst.q      C010, VFPU_LOG10E", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_LOG10E"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06D0001, "vcst.s      S010, VFPU_LN2", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_LN2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06D0081, "vcst.p      C010, VFPU_LN2", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_LN2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06D8001, "vcst.t      C010, VFPU_LN2", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_LN2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06D8081, "vcst.q      C010, VFPU_LN2", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_LN2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06E0001, "vcst.s      S010, VFPU_LN10", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_LN10"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06E0081, "vcst.p      C010, VFPU_LN10", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_LN10"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06E8001, "vcst.t      C010, VFPU_LN10", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_LN10"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06E8081, "vcst.q      C010, VFPU_LN10", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_LN10"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06F0001, "vcst.s      S010, VFPU_2PI", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_2PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06F0081, "vcst.p      C010, VFPU_2PI", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_2PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06F8001, "vcst.t      C010, VFPU_2PI", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_2PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD06F8081, "vcst.q      C010, VFPU_2PI", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_2PI"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0700001, "vcst.s      S010, VFPU_PI_6", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_PI_6"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0700081, "vcst.p      C010, VFPU_PI_6", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_PI_6"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0708001, "vcst.t      C010, VFPU_PI_6", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_PI_6"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0708081, "vcst.q      C010, VFPU_PI_6", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_PI_6"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0710001, "vcst.s      S010, VFPU_LOG10TWO", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_LOG10TWO"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0710081, "vcst.p      C010, VFPU_LOG10TWO", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_LOG10TWO"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0718001, "vcst.t      C010, VFPU_LOG10TWO", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_LOG10TWO"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0718081, "vcst.q      C010, VFPU_LOG10TWO", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_LOG10TWO"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0720001, "vcst.s      S010, VFPU_LOG2TEN", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_LOG2TEN"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0720081, "vcst.p      C010, VFPU_LOG2TEN", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_LOG2TEN"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0728001, "vcst.t      C010, VFPU_LOG2TEN", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_LOG2TEN"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0728081, "vcst.q      C010, VFPU_LOG2TEN", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_LOG2TEN"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0730001, "vcst.s      S010, VFPU_SQRT3_2", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("VFPU_SQRT3_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0730081, "vcst.p      C010, VFPU_SQRT3_2", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("VFPU_SQRT3_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0738001, "vcst.t      C010, VFPU_SQRT3_2", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("VFPU_SQRT3_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0738081, "vcst.q      C010, VFPU_SQRT3_2", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("VFPU_SQRT3_2"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0740001, "vcst.s      S010, INVALID_20", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_20"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0740081, "vcst.p      C010, INVALID_20", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_20"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0748001, "vcst.t      C010, INVALID_20", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_20"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0748081, "vcst.q      C010, INVALID_20", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_20"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0750001, "vcst.s      S010, INVALID_21", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_21"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0750081, "vcst.p      C010, INVALID_21", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_21"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0758001, "vcst.t      C010, INVALID_21", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_21"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0758081, "vcst.q      C010, INVALID_21", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_21"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0760001, "vcst.s      S010, INVALID_22", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_22"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0760081, "vcst.p      C010, INVALID_22", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_22"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0768001, "vcst.t      C010, INVALID_22", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_22"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0768081, "vcst.q      C010, INVALID_22", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_22"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0770001, "vcst.s      S010, INVALID_23", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_23"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0770081, "vcst.p      C010, INVALID_23", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_23"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0778001, "vcst.t      C010, INVALID_23", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_23"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0778081, "vcst.q      C010, INVALID_23", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_23"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0780001, "vcst.s      S010, INVALID_24", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_24"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0780081, "vcst.p      C010, INVALID_24", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_24"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0788001, "vcst.t      C010, INVALID_24", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_24"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0788081, "vcst.q      C010, INVALID_24", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_24"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0790001, "vcst.s      S010, INVALID_25", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_25"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0790081, "vcst.p      C010, INVALID_25", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_25"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0798001, "vcst.t      C010, INVALID_25", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_25"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD0798081, "vcst.q      C010, INVALID_25", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_25"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07A0001, "vcst.s      S010, INVALID_26", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_26"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07A0081, "vcst.p      C010, INVALID_26", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_26"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07A8001, "vcst.t      C010, INVALID_26", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_26"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07A8081, "vcst.q      C010, INVALID_26", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_26"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07B0001, "vcst.s      S010, INVALID_27", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_27"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07B0081, "vcst.p      C010, INVALID_27", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_27"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07B8001, "vcst.t      C010, INVALID_27", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_27"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07B8081, "vcst.q      C010, INVALID_27", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_27"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07C0001, "vcst.s      S010, INVALID_28", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_28"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07C0081, "vcst.p      C010, INVALID_28", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_28"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07C8001, "vcst.t      C010, INVALID_28", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_28"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07C8081, "vcst.q      C010, INVALID_28", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_28"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07D0001, "vcst.s      S010, INVALID_29", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_29"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07D0081, "vcst.p      C010, INVALID_29", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_29"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07D8001, "vcst.t      C010, INVALID_29", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_29"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07D8081, "vcst.q      C010, INVALID_29", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_29"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07E0001, "vcst.s      S010, INVALID_30", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_30"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07E0081, "vcst.p      C010, INVALID_30", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_30"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07E8001, "vcst.t      C010, INVALID_30", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_30"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07E8081, "vcst.q      C010, INVALID_30", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_30"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07F0001, "vcst.s      S010, INVALID_31", Opcode::r4000allegrex_vcst_s, "vcst.s", [Some("S010"), Some("INVALID_31"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07F0081, "vcst.p      C010, INVALID_31", Opcode::r4000allegrex_vcst_p, "vcst.p", [Some("C010"), Some("INVALID_31"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07F8001, "vcst.t      C010, INVALID_31", Opcode::r4000allegrex_vcst_t, "vcst.t", [Some("C010"), Some("INVALID_31"), None, None, None]),
        TestEntry::new_r4000allegrex(0xD07F8081, "vcst.q      C010, INVALID_31", Opcode::r4000allegrex_vcst_q, "vcst.q", [Some("C010"), Some("INVALID_31"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD0801001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0801001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0801081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0801081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0809001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0809001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0809081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0809081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0811001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0811001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0811081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0811081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0819001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0819001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0819081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0819081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0821001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0821001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0821081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0821081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0829001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0829001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0829081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0829081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0831001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0831001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0831081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0831081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0839001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0839001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0839081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0839081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0841001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0841001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0841081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0841081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0849001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0849001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0849081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0849081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0851001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0851001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0851081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0851081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0859001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0859001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0859081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0859081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0861001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0861001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0861081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0861081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0869001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0869001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0869081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0869081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0871001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0871001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0871081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0871081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0879001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0879001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0879081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0879081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0881001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0881001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0881081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0881081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0889001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0889001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0889081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0889081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0891001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0891001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0891081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0891081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0899001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0899001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0899081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0899081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08A1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08A1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08A1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08A1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08A9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08A9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08A9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08A9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08B1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08B1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08B9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08B9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08C1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08C1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08C9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08C9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08C9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08D1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08D1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08D9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08D9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08E1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08E1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08E9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08E9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08F1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08F1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08F9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD08F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD08F9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0901001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0901001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0901081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0901081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0909001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0909001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0909081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0909081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0911001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0911001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0911081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0911081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0919001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0919001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0919081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0919081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0921001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0921001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0921081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0921081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0929001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0929001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0929081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0929081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0931001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0931001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0931081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0931081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0939001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0939001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0939081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0939081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0941001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0941001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0941081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0941081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0949001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0949001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0949081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0949081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0951001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0951001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0951081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0951081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0959001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0959001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0959081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0959081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0961001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0961001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0961081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0961081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0969001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0969001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0969081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0969081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0971001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0971001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0971081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0971081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0979001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0979001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0979081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0979081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0981001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0981001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0981081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0981081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0989001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0989001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0989081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0989081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0991001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0991001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0991081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0991081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0999001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0999001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0999081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0999081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09A1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09A1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09A1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09A1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09A9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09A9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09A9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09A9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09B1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09B1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09B1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09B1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09B9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09B9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09B9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09B9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09C1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09C1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09C1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09C1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09C9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09C9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09C9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09C9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09D1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09D1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09D1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09D1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09D9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09D9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09D9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09D9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09E1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09E1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09E1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09E1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09E9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09E9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09E9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09E9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09F1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09F1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09F1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09F1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09F9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09F9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD09F9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD09F9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0A99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0A99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AD1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AD1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AD9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AD9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AF1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AF1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AF9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0AF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0AF9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0B99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0B99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BD1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BD1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BD9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BD9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BF1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BF1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BF9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0BF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0BF9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0C99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0C99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CD1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CD1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CD9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CD9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CF1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CF1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CF9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0CF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0CF9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0D99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0D99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DD1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DD1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DD9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DD9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DF1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DF1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DF9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0DF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0DF9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0E99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0E99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0ED1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0ED1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0ED1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0ED1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0ED9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0ED9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0ED9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0ED9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EF1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EF1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EF9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0EF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0EF9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0F99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0F99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FD1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD0FD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FD1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
                instr: Instruction::new(
                0xD0FD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FD9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FD9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FF1001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FF1081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FF9001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
            },
        TestEntry {
                instr: Instruction::new(
                0xD0FF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD0FF9081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU4_FMT2() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU4_FMT2 */
        TestEntry::new_r4000allegrex(0xD2001001, "vf2in.s     S010, S400, 0", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2001081, "vf2in.p     C010, C400, 0", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2009001, "vf2in.t     C010, C400, 0", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2009081, "vf2in.q     C010, C400, 0", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2011001, "vf2in.s     S010, S400, 1", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2011081, "vf2in.p     C010, C400, 1", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2019001, "vf2in.t     C010, C400, 1", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2019081, "vf2in.q     C010, C400, 1", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2021001, "vf2in.s     S010, S400, 2", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2021081, "vf2in.p     C010, C400, 2", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2029001, "vf2in.t     C010, C400, 2", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2029081, "vf2in.q     C010, C400, 2", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2031001, "vf2in.s     S010, S400, 3", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2031081, "vf2in.p     C010, C400, 3", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2039001, "vf2in.t     C010, C400, 3", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2039081, "vf2in.q     C010, C400, 3", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2041001, "vf2in.s     S010, S400, 4", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2041081, "vf2in.p     C010, C400, 4", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2049001, "vf2in.t     C010, C400, 4", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2049081, "vf2in.q     C010, C400, 4", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2051001, "vf2in.s     S010, S400, 5", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2051081, "vf2in.p     C010, C400, 5", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2059001, "vf2in.t     C010, C400, 5", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2059081, "vf2in.q     C010, C400, 5", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2061001, "vf2in.s     S010, S400, 6", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2061081, "vf2in.p     C010, C400, 6", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2069001, "vf2in.t     C010, C400, 6", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2069081, "vf2in.q     C010, C400, 6", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2071001, "vf2in.s     S010, S400, 7", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2071081, "vf2in.p     C010, C400, 7", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2079001, "vf2in.t     C010, C400, 7", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2079081, "vf2in.q     C010, C400, 7", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2081001, "vf2in.s     S010, S400, 8", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2081081, "vf2in.p     C010, C400, 8", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2089001, "vf2in.t     C010, C400, 8", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2089081, "vf2in.q     C010, C400, 8", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2091001, "vf2in.s     S010, S400, 9", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2091081, "vf2in.p     C010, C400, 9", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2099001, "vf2in.t     C010, C400, 9", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2099081, "vf2in.q     C010, C400, 9", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD20A1001, "vf2in.s     S010, S400, 10", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD20A1081, "vf2in.p     C010, C400, 10", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD20A9001, "vf2in.t     C010, C400, 10", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD20A9081, "vf2in.q     C010, C400, 10", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD20B1001, "vf2in.s     S010, S400, 11", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD20B1081, "vf2in.p     C010, C400, 11", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD20B9001, "vf2in.t     C010, C400, 11", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD20B9081, "vf2in.q     C010, C400, 11", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD20C1001, "vf2in.s     S010, S400, 12", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD20C1081, "vf2in.p     C010, C400, 12", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD20C9001, "vf2in.t     C010, C400, 12", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD20C9081, "vf2in.q     C010, C400, 12", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD20D1001, "vf2in.s     S010, S400, 13", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD20D1081, "vf2in.p     C010, C400, 13", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD20D9001, "vf2in.t     C010, C400, 13", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD20D9081, "vf2in.q     C010, C400, 13", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD20E1001, "vf2in.s     S010, S400, 14", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD20E1081, "vf2in.p     C010, C400, 14", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD20E9001, "vf2in.t     C010, C400, 14", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD20E9081, "vf2in.q     C010, C400, 14", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD20F1001, "vf2in.s     S010, S400, 15", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD20F1081, "vf2in.p     C010, C400, 15", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD20F9001, "vf2in.t     C010, C400, 15", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD20F9081, "vf2in.q     C010, C400, 15", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD2101001, "vf2in.s     S010, S400, 16", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2101081, "vf2in.p     C010, C400, 16", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2109001, "vf2in.t     C010, C400, 16", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2109081, "vf2in.q     C010, C400, 16", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2111001, "vf2in.s     S010, S400, 17", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2111081, "vf2in.p     C010, C400, 17", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2119001, "vf2in.t     C010, C400, 17", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2119081, "vf2in.q     C010, C400, 17", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2121001, "vf2in.s     S010, S400, 18", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2121081, "vf2in.p     C010, C400, 18", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2129001, "vf2in.t     C010, C400, 18", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2129081, "vf2in.q     C010, C400, 18", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2131001, "vf2in.s     S010, S400, 19", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2131081, "vf2in.p     C010, C400, 19", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2139001, "vf2in.t     C010, C400, 19", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2139081, "vf2in.q     C010, C400, 19", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2141001, "vf2in.s     S010, S400, 20", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2141081, "vf2in.p     C010, C400, 20", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2149001, "vf2in.t     C010, C400, 20", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2149081, "vf2in.q     C010, C400, 20", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2151001, "vf2in.s     S010, S400, 21", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2151081, "vf2in.p     C010, C400, 21", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2159001, "vf2in.t     C010, C400, 21", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2159081, "vf2in.q     C010, C400, 21", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2161001, "vf2in.s     S010, S400, 22", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2161081, "vf2in.p     C010, C400, 22", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2169001, "vf2in.t     C010, C400, 22", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2169081, "vf2in.q     C010, C400, 22", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2171001, "vf2in.s     S010, S400, 23", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2171081, "vf2in.p     C010, C400, 23", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2179001, "vf2in.t     C010, C400, 23", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2179081, "vf2in.q     C010, C400, 23", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2181001, "vf2in.s     S010, S400, 24", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2181081, "vf2in.p     C010, C400, 24", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2189001, "vf2in.t     C010, C400, 24", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2189081, "vf2in.q     C010, C400, 24", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2191001, "vf2in.s     S010, S400, 25", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2191081, "vf2in.p     C010, C400, 25", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2199001, "vf2in.t     C010, C400, 25", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2199081, "vf2in.q     C010, C400, 25", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD21A1001, "vf2in.s     S010, S400, 26", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD21A1081, "vf2in.p     C010, C400, 26", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD21A9001, "vf2in.t     C010, C400, 26", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD21A9081, "vf2in.q     C010, C400, 26", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD21B1001, "vf2in.s     S010, S400, 27", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD21B1081, "vf2in.p     C010, C400, 27", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD21B9001, "vf2in.t     C010, C400, 27", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD21B9081, "vf2in.q     C010, C400, 27", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD21C1001, "vf2in.s     S010, S400, 28", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD21C1081, "vf2in.p     C010, C400, 28", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD21C9001, "vf2in.t     C010, C400, 28", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD21C9081, "vf2in.q     C010, C400, 28", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD21D1001, "vf2in.s     S010, S400, 29", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD21D1081, "vf2in.p     C010, C400, 29", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD21D9001, "vf2in.t     C010, C400, 29", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD21D9081, "vf2in.q     C010, C400, 29", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD21E1001, "vf2in.s     S010, S400, 30", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD21E1081, "vf2in.p     C010, C400, 30", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD21E9001, "vf2in.t     C010, C400, 30", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD21E9081, "vf2in.q     C010, C400, 30", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD21F1001, "vf2in.s     S010, S400, 31", Opcode::r4000allegrex_vf2in_s, "vf2in.s", [Some("S010"), Some("S400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD21F1081, "vf2in.p     C010, C400, 31", Opcode::r4000allegrex_vf2in_p, "vf2in.p", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD21F9001, "vf2in.t     C010, C400, 31", Opcode::r4000allegrex_vf2in_t, "vf2in.t", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD21F9081, "vf2in.q     C010, C400, 31", Opcode::r4000allegrex_vf2in_q, "vf2in.q", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD2201001, "vf2iz.s     S010, S400, 0", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2201081, "vf2iz.p     C010, C400, 0", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2209001, "vf2iz.t     C010, C400, 0", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2209081, "vf2iz.q     C010, C400, 0", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2211001, "vf2iz.s     S010, S400, 1", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2211081, "vf2iz.p     C010, C400, 1", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2219001, "vf2iz.t     C010, C400, 1", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2219081, "vf2iz.q     C010, C400, 1", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2221001, "vf2iz.s     S010, S400, 2", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2221081, "vf2iz.p     C010, C400, 2", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2229001, "vf2iz.t     C010, C400, 2", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2229081, "vf2iz.q     C010, C400, 2", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2231001, "vf2iz.s     S010, S400, 3", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2231081, "vf2iz.p     C010, C400, 3", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2239001, "vf2iz.t     C010, C400, 3", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2239081, "vf2iz.q     C010, C400, 3", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2241001, "vf2iz.s     S010, S400, 4", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2241081, "vf2iz.p     C010, C400, 4", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2249001, "vf2iz.t     C010, C400, 4", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2249081, "vf2iz.q     C010, C400, 4", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2251001, "vf2iz.s     S010, S400, 5", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2251081, "vf2iz.p     C010, C400, 5", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2259001, "vf2iz.t     C010, C400, 5", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2259081, "vf2iz.q     C010, C400, 5", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2261001, "vf2iz.s     S010, S400, 6", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2261081, "vf2iz.p     C010, C400, 6", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2269001, "vf2iz.t     C010, C400, 6", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2269081, "vf2iz.q     C010, C400, 6", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2271001, "vf2iz.s     S010, S400, 7", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2271081, "vf2iz.p     C010, C400, 7", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2279001, "vf2iz.t     C010, C400, 7", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2279081, "vf2iz.q     C010, C400, 7", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2281001, "vf2iz.s     S010, S400, 8", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2281081, "vf2iz.p     C010, C400, 8", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2289001, "vf2iz.t     C010, C400, 8", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2289081, "vf2iz.q     C010, C400, 8", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2291001, "vf2iz.s     S010, S400, 9", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2291081, "vf2iz.p     C010, C400, 9", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2299001, "vf2iz.t     C010, C400, 9", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2299081, "vf2iz.q     C010, C400, 9", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD22A1001, "vf2iz.s     S010, S400, 10", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD22A1081, "vf2iz.p     C010, C400, 10", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD22A9001, "vf2iz.t     C010, C400, 10", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD22A9081, "vf2iz.q     C010, C400, 10", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD22B1001, "vf2iz.s     S010, S400, 11", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD22B1081, "vf2iz.p     C010, C400, 11", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD22B9001, "vf2iz.t     C010, C400, 11", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD22B9081, "vf2iz.q     C010, C400, 11", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD22C1001, "vf2iz.s     S010, S400, 12", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD22C1081, "vf2iz.p     C010, C400, 12", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD22C9001, "vf2iz.t     C010, C400, 12", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD22C9081, "vf2iz.q     C010, C400, 12", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD22D1001, "vf2iz.s     S010, S400, 13", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD22D1081, "vf2iz.p     C010, C400, 13", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD22D9001, "vf2iz.t     C010, C400, 13", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD22D9081, "vf2iz.q     C010, C400, 13", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD22E1001, "vf2iz.s     S010, S400, 14", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD22E1081, "vf2iz.p     C010, C400, 14", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD22E9001, "vf2iz.t     C010, C400, 14", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD22E9081, "vf2iz.q     C010, C400, 14", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD22F1001, "vf2iz.s     S010, S400, 15", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD22F1081, "vf2iz.p     C010, C400, 15", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD22F9001, "vf2iz.t     C010, C400, 15", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD22F9081, "vf2iz.q     C010, C400, 15", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD2301001, "vf2iz.s     S010, S400, 16", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2301081, "vf2iz.p     C010, C400, 16", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2309001, "vf2iz.t     C010, C400, 16", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2309081, "vf2iz.q     C010, C400, 16", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2311001, "vf2iz.s     S010, S400, 17", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2311081, "vf2iz.p     C010, C400, 17", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2319001, "vf2iz.t     C010, C400, 17", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2319081, "vf2iz.q     C010, C400, 17", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2321001, "vf2iz.s     S010, S400, 18", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2321081, "vf2iz.p     C010, C400, 18", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2329001, "vf2iz.t     C010, C400, 18", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2329081, "vf2iz.q     C010, C400, 18", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2331001, "vf2iz.s     S010, S400, 19", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2331081, "vf2iz.p     C010, C400, 19", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2339001, "vf2iz.t     C010, C400, 19", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2339081, "vf2iz.q     C010, C400, 19", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2341001, "vf2iz.s     S010, S400, 20", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2341081, "vf2iz.p     C010, C400, 20", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2349001, "vf2iz.t     C010, C400, 20", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2349081, "vf2iz.q     C010, C400, 20", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2351001, "vf2iz.s     S010, S400, 21", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2351081, "vf2iz.p     C010, C400, 21", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2359001, "vf2iz.t     C010, C400, 21", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2359081, "vf2iz.q     C010, C400, 21", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2361001, "vf2iz.s     S010, S400, 22", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2361081, "vf2iz.p     C010, C400, 22", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2369001, "vf2iz.t     C010, C400, 22", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2369081, "vf2iz.q     C010, C400, 22", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2371001, "vf2iz.s     S010, S400, 23", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2371081, "vf2iz.p     C010, C400, 23", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2379001, "vf2iz.t     C010, C400, 23", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2379081, "vf2iz.q     C010, C400, 23", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2381001, "vf2iz.s     S010, S400, 24", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2381081, "vf2iz.p     C010, C400, 24", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2389001, "vf2iz.t     C010, C400, 24", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2389081, "vf2iz.q     C010, C400, 24", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2391001, "vf2iz.s     S010, S400, 25", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2391081, "vf2iz.p     C010, C400, 25", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2399001, "vf2iz.t     C010, C400, 25", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2399081, "vf2iz.q     C010, C400, 25", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD23A1001, "vf2iz.s     S010, S400, 26", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD23A1081, "vf2iz.p     C010, C400, 26", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD23A9001, "vf2iz.t     C010, C400, 26", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD23A9081, "vf2iz.q     C010, C400, 26", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD23B1001, "vf2iz.s     S010, S400, 27", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD23B1081, "vf2iz.p     C010, C400, 27", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD23B9001, "vf2iz.t     C010, C400, 27", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD23B9081, "vf2iz.q     C010, C400, 27", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD23C1001, "vf2iz.s     S010, S400, 28", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD23C1081, "vf2iz.p     C010, C400, 28", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD23C9001, "vf2iz.t     C010, C400, 28", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD23C9081, "vf2iz.q     C010, C400, 28", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD23D1001, "vf2iz.s     S010, S400, 29", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD23D1081, "vf2iz.p     C010, C400, 29", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD23D9001, "vf2iz.t     C010, C400, 29", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD23D9081, "vf2iz.q     C010, C400, 29", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD23E1001, "vf2iz.s     S010, S400, 30", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD23E1081, "vf2iz.p     C010, C400, 30", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD23E9001, "vf2iz.t     C010, C400, 30", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD23E9081, "vf2iz.q     C010, C400, 30", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD23F1001, "vf2iz.s     S010, S400, 31", Opcode::r4000allegrex_vf2iz_s, "vf2iz.s", [Some("S010"), Some("S400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD23F1081, "vf2iz.p     C010, C400, 31", Opcode::r4000allegrex_vf2iz_p, "vf2iz.p", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD23F9001, "vf2iz.t     C010, C400, 31", Opcode::r4000allegrex_vf2iz_t, "vf2iz.t", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD23F9081, "vf2iz.q     C010, C400, 31", Opcode::r4000allegrex_vf2iz_q, "vf2iz.q", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD2401001, "vf2iu.s     S010, S400, 0", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2401081, "vf2iu.p     C010, C400, 0", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2409001, "vf2iu.t     C010, C400, 0", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2409081, "vf2iu.q     C010, C400, 0", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2411001, "vf2iu.s     S010, S400, 1", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2411081, "vf2iu.p     C010, C400, 1", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2419001, "vf2iu.t     C010, C400, 1", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2419081, "vf2iu.q     C010, C400, 1", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2421001, "vf2iu.s     S010, S400, 2", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2421081, "vf2iu.p     C010, C400, 2", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2429001, "vf2iu.t     C010, C400, 2", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2429081, "vf2iu.q     C010, C400, 2", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2431001, "vf2iu.s     S010, S400, 3", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2431081, "vf2iu.p     C010, C400, 3", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2439001, "vf2iu.t     C010, C400, 3", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2439081, "vf2iu.q     C010, C400, 3", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2441001, "vf2iu.s     S010, S400, 4", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2441081, "vf2iu.p     C010, C400, 4", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2449001, "vf2iu.t     C010, C400, 4", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2449081, "vf2iu.q     C010, C400, 4", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2451001, "vf2iu.s     S010, S400, 5", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2451081, "vf2iu.p     C010, C400, 5", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2459001, "vf2iu.t     C010, C400, 5", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2459081, "vf2iu.q     C010, C400, 5", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2461001, "vf2iu.s     S010, S400, 6", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2461081, "vf2iu.p     C010, C400, 6", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2469001, "vf2iu.t     C010, C400, 6", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2469081, "vf2iu.q     C010, C400, 6", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2471001, "vf2iu.s     S010, S400, 7", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2471081, "vf2iu.p     C010, C400, 7", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2479001, "vf2iu.t     C010, C400, 7", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2479081, "vf2iu.q     C010, C400, 7", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2481001, "vf2iu.s     S010, S400, 8", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2481081, "vf2iu.p     C010, C400, 8", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2489001, "vf2iu.t     C010, C400, 8", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2489081, "vf2iu.q     C010, C400, 8", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2491001, "vf2iu.s     S010, S400, 9", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2491081, "vf2iu.p     C010, C400, 9", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2499001, "vf2iu.t     C010, C400, 9", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2499081, "vf2iu.q     C010, C400, 9", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD24A1001, "vf2iu.s     S010, S400, 10", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD24A1081, "vf2iu.p     C010, C400, 10", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD24A9001, "vf2iu.t     C010, C400, 10", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD24A9081, "vf2iu.q     C010, C400, 10", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD24B1001, "vf2iu.s     S010, S400, 11", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD24B1081, "vf2iu.p     C010, C400, 11", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD24B9001, "vf2iu.t     C010, C400, 11", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD24B9081, "vf2iu.q     C010, C400, 11", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD24C1001, "vf2iu.s     S010, S400, 12", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD24C1081, "vf2iu.p     C010, C400, 12", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD24C9001, "vf2iu.t     C010, C400, 12", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD24C9081, "vf2iu.q     C010, C400, 12", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD24D1001, "vf2iu.s     S010, S400, 13", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD24D1081, "vf2iu.p     C010, C400, 13", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD24D9001, "vf2iu.t     C010, C400, 13", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD24D9081, "vf2iu.q     C010, C400, 13", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD24E1001, "vf2iu.s     S010, S400, 14", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD24E1081, "vf2iu.p     C010, C400, 14", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD24E9001, "vf2iu.t     C010, C400, 14", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD24E9081, "vf2iu.q     C010, C400, 14", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD24F1001, "vf2iu.s     S010, S400, 15", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD24F1081, "vf2iu.p     C010, C400, 15", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD24F9001, "vf2iu.t     C010, C400, 15", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD24F9081, "vf2iu.q     C010, C400, 15", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD2501001, "vf2iu.s     S010, S400, 16", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2501081, "vf2iu.p     C010, C400, 16", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2509001, "vf2iu.t     C010, C400, 16", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2509081, "vf2iu.q     C010, C400, 16", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2511001, "vf2iu.s     S010, S400, 17", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2511081, "vf2iu.p     C010, C400, 17", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2519001, "vf2iu.t     C010, C400, 17", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2519081, "vf2iu.q     C010, C400, 17", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2521001, "vf2iu.s     S010, S400, 18", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2521081, "vf2iu.p     C010, C400, 18", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2529001, "vf2iu.t     C010, C400, 18", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2529081, "vf2iu.q     C010, C400, 18", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2531001, "vf2iu.s     S010, S400, 19", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2531081, "vf2iu.p     C010, C400, 19", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2539001, "vf2iu.t     C010, C400, 19", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2539081, "vf2iu.q     C010, C400, 19", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2541001, "vf2iu.s     S010, S400, 20", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2541081, "vf2iu.p     C010, C400, 20", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2549001, "vf2iu.t     C010, C400, 20", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2549081, "vf2iu.q     C010, C400, 20", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2551001, "vf2iu.s     S010, S400, 21", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2551081, "vf2iu.p     C010, C400, 21", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2559001, "vf2iu.t     C010, C400, 21", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2559081, "vf2iu.q     C010, C400, 21", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2561001, "vf2iu.s     S010, S400, 22", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2561081, "vf2iu.p     C010, C400, 22", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2569001, "vf2iu.t     C010, C400, 22", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2569081, "vf2iu.q     C010, C400, 22", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2571001, "vf2iu.s     S010, S400, 23", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2571081, "vf2iu.p     C010, C400, 23", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2579001, "vf2iu.t     C010, C400, 23", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2579081, "vf2iu.q     C010, C400, 23", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2581001, "vf2iu.s     S010, S400, 24", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2581081, "vf2iu.p     C010, C400, 24", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2589001, "vf2iu.t     C010, C400, 24", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2589081, "vf2iu.q     C010, C400, 24", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2591001, "vf2iu.s     S010, S400, 25", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2591081, "vf2iu.p     C010, C400, 25", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2599001, "vf2iu.t     C010, C400, 25", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2599081, "vf2iu.q     C010, C400, 25", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD25A1001, "vf2iu.s     S010, S400, 26", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD25A1081, "vf2iu.p     C010, C400, 26", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD25A9001, "vf2iu.t     C010, C400, 26", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD25A9081, "vf2iu.q     C010, C400, 26", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD25B1001, "vf2iu.s     S010, S400, 27", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD25B1081, "vf2iu.p     C010, C400, 27", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD25B9001, "vf2iu.t     C010, C400, 27", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD25B9081, "vf2iu.q     C010, C400, 27", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD25C1001, "vf2iu.s     S010, S400, 28", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD25C1081, "vf2iu.p     C010, C400, 28", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD25C9001, "vf2iu.t     C010, C400, 28", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD25C9081, "vf2iu.q     C010, C400, 28", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD25D1001, "vf2iu.s     S010, S400, 29", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD25D1081, "vf2iu.p     C010, C400, 29", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD25D9001, "vf2iu.t     C010, C400, 29", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD25D9081, "vf2iu.q     C010, C400, 29", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD25E1001, "vf2iu.s     S010, S400, 30", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD25E1081, "vf2iu.p     C010, C400, 30", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD25E9001, "vf2iu.t     C010, C400, 30", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD25E9081, "vf2iu.q     C010, C400, 30", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD25F1001, "vf2iu.s     S010, S400, 31", Opcode::r4000allegrex_vf2iu_s, "vf2iu.s", [Some("S010"), Some("S400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD25F1081, "vf2iu.p     C010, C400, 31", Opcode::r4000allegrex_vf2iu_p, "vf2iu.p", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD25F9001, "vf2iu.t     C010, C400, 31", Opcode::r4000allegrex_vf2iu_t, "vf2iu.t", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD25F9081, "vf2iu.q     C010, C400, 31", Opcode::r4000allegrex_vf2iu_q, "vf2iu.q", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD2601001, "vf2id.s     S010, S400, 0", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2601081, "vf2id.p     C010, C400, 0", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2609001, "vf2id.t     C010, C400, 0", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2609081, "vf2id.q     C010, C400, 0", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2611001, "vf2id.s     S010, S400, 1", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2611081, "vf2id.p     C010, C400, 1", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2619001, "vf2id.t     C010, C400, 1", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2619081, "vf2id.q     C010, C400, 1", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2621001, "vf2id.s     S010, S400, 2", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2621081, "vf2id.p     C010, C400, 2", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2629001, "vf2id.t     C010, C400, 2", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2629081, "vf2id.q     C010, C400, 2", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2631001, "vf2id.s     S010, S400, 3", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2631081, "vf2id.p     C010, C400, 3", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2639001, "vf2id.t     C010, C400, 3", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2639081, "vf2id.q     C010, C400, 3", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2641001, "vf2id.s     S010, S400, 4", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2641081, "vf2id.p     C010, C400, 4", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2649001, "vf2id.t     C010, C400, 4", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2649081, "vf2id.q     C010, C400, 4", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2651001, "vf2id.s     S010, S400, 5", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2651081, "vf2id.p     C010, C400, 5", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2659001, "vf2id.t     C010, C400, 5", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2659081, "vf2id.q     C010, C400, 5", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2661001, "vf2id.s     S010, S400, 6", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2661081, "vf2id.p     C010, C400, 6", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2669001, "vf2id.t     C010, C400, 6", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2669081, "vf2id.q     C010, C400, 6", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2671001, "vf2id.s     S010, S400, 7", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2671081, "vf2id.p     C010, C400, 7", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2679001, "vf2id.t     C010, C400, 7", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2679081, "vf2id.q     C010, C400, 7", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2681001, "vf2id.s     S010, S400, 8", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2681081, "vf2id.p     C010, C400, 8", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2689001, "vf2id.t     C010, C400, 8", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2689081, "vf2id.q     C010, C400, 8", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2691001, "vf2id.s     S010, S400, 9", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2691081, "vf2id.p     C010, C400, 9", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2699001, "vf2id.t     C010, C400, 9", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2699081, "vf2id.q     C010, C400, 9", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD26A1001, "vf2id.s     S010, S400, 10", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD26A1081, "vf2id.p     C010, C400, 10", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD26A9001, "vf2id.t     C010, C400, 10", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD26A9081, "vf2id.q     C010, C400, 10", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD26B1001, "vf2id.s     S010, S400, 11", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD26B1081, "vf2id.p     C010, C400, 11", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD26B9001, "vf2id.t     C010, C400, 11", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD26B9081, "vf2id.q     C010, C400, 11", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD26C1001, "vf2id.s     S010, S400, 12", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD26C1081, "vf2id.p     C010, C400, 12", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD26C9001, "vf2id.t     C010, C400, 12", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD26C9081, "vf2id.q     C010, C400, 12", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD26D1001, "vf2id.s     S010, S400, 13", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD26D1081, "vf2id.p     C010, C400, 13", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD26D9001, "vf2id.t     C010, C400, 13", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD26D9081, "vf2id.q     C010, C400, 13", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD26E1001, "vf2id.s     S010, S400, 14", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD26E1081, "vf2id.p     C010, C400, 14", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD26E9001, "vf2id.t     C010, C400, 14", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD26E9081, "vf2id.q     C010, C400, 14", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD26F1001, "vf2id.s     S010, S400, 15", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD26F1081, "vf2id.p     C010, C400, 15", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD26F9001, "vf2id.t     C010, C400, 15", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD26F9081, "vf2id.q     C010, C400, 15", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD2701001, "vf2id.s     S010, S400, 16", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2701081, "vf2id.p     C010, C400, 16", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2709001, "vf2id.t     C010, C400, 16", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2709081, "vf2id.q     C010, C400, 16", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2711001, "vf2id.s     S010, S400, 17", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2711081, "vf2id.p     C010, C400, 17", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2719001, "vf2id.t     C010, C400, 17", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2719081, "vf2id.q     C010, C400, 17", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2721001, "vf2id.s     S010, S400, 18", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2721081, "vf2id.p     C010, C400, 18", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2729001, "vf2id.t     C010, C400, 18", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2729081, "vf2id.q     C010, C400, 18", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2731001, "vf2id.s     S010, S400, 19", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2731081, "vf2id.p     C010, C400, 19", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2739001, "vf2id.t     C010, C400, 19", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2739081, "vf2id.q     C010, C400, 19", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2741001, "vf2id.s     S010, S400, 20", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2741081, "vf2id.p     C010, C400, 20", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2749001, "vf2id.t     C010, C400, 20", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2749081, "vf2id.q     C010, C400, 20", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2751001, "vf2id.s     S010, S400, 21", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2751081, "vf2id.p     C010, C400, 21", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2759001, "vf2id.t     C010, C400, 21", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2759081, "vf2id.q     C010, C400, 21", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2761001, "vf2id.s     S010, S400, 22", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2761081, "vf2id.p     C010, C400, 22", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2769001, "vf2id.t     C010, C400, 22", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2769081, "vf2id.q     C010, C400, 22", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2771001, "vf2id.s     S010, S400, 23", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2771081, "vf2id.p     C010, C400, 23", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2779001, "vf2id.t     C010, C400, 23", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2779081, "vf2id.q     C010, C400, 23", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2781001, "vf2id.s     S010, S400, 24", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2781081, "vf2id.p     C010, C400, 24", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2789001, "vf2id.t     C010, C400, 24", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2789081, "vf2id.q     C010, C400, 24", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2791001, "vf2id.s     S010, S400, 25", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2791081, "vf2id.p     C010, C400, 25", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2799001, "vf2id.t     C010, C400, 25", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2799081, "vf2id.q     C010, C400, 25", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD27A1001, "vf2id.s     S010, S400, 26", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD27A1081, "vf2id.p     C010, C400, 26", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD27A9001, "vf2id.t     C010, C400, 26", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD27A9081, "vf2id.q     C010, C400, 26", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD27B1001, "vf2id.s     S010, S400, 27", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD27B1081, "vf2id.p     C010, C400, 27", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD27B9001, "vf2id.t     C010, C400, 27", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD27B9081, "vf2id.q     C010, C400, 27", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD27C1001, "vf2id.s     S010, S400, 28", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD27C1081, "vf2id.p     C010, C400, 28", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD27C9001, "vf2id.t     C010, C400, 28", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD27C9081, "vf2id.q     C010, C400, 28", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD27D1001, "vf2id.s     S010, S400, 29", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD27D1081, "vf2id.p     C010, C400, 29", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD27D9001, "vf2id.t     C010, C400, 29", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD27D9081, "vf2id.q     C010, C400, 29", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD27E1001, "vf2id.s     S010, S400, 30", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD27E1081, "vf2id.p     C010, C400, 30", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD27E9001, "vf2id.t     C010, C400, 30", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD27E9081, "vf2id.q     C010, C400, 30", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD27F1001, "vf2id.s     S010, S400, 31", Opcode::r4000allegrex_vf2id_s, "vf2id.s", [Some("S010"), Some("S400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD27F1081, "vf2id.p     C010, C400, 31", Opcode::r4000allegrex_vf2id_p, "vf2id.p", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD27F9001, "vf2id.t     C010, C400, 31", Opcode::r4000allegrex_vf2id_t, "vf2id.t", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD27F9081, "vf2id.q     C010, C400, 31", Opcode::r4000allegrex_vf2id_q, "vf2id.q", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD2801001, "vi2f.s      S010, S400, 0", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2801081, "vi2f.p      C010, C400, 0", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2809001, "vi2f.t      C010, C400, 0", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2809081, "vi2f.q      C010, C400, 0", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2811001, "vi2f.s      S010, S400, 1", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2811081, "vi2f.p      C010, C400, 1", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2819001, "vi2f.t      C010, C400, 1", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2819081, "vi2f.q      C010, C400, 1", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2821001, "vi2f.s      S010, S400, 2", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2821081, "vi2f.p      C010, C400, 2", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2829001, "vi2f.t      C010, C400, 2", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2829081, "vi2f.q      C010, C400, 2", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2831001, "vi2f.s      S010, S400, 3", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2831081, "vi2f.p      C010, C400, 3", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2839001, "vi2f.t      C010, C400, 3", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2839081, "vi2f.q      C010, C400, 3", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2841001, "vi2f.s      S010, S400, 4", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2841081, "vi2f.p      C010, C400, 4", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2849001, "vi2f.t      C010, C400, 4", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2849081, "vi2f.q      C010, C400, 4", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2851001, "vi2f.s      S010, S400, 5", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2851081, "vi2f.p      C010, C400, 5", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2859001, "vi2f.t      C010, C400, 5", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2859081, "vi2f.q      C010, C400, 5", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2861001, "vi2f.s      S010, S400, 6", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2861081, "vi2f.p      C010, C400, 6", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2869001, "vi2f.t      C010, C400, 6", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2869081, "vi2f.q      C010, C400, 6", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2871001, "vi2f.s      S010, S400, 7", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2871081, "vi2f.p      C010, C400, 7", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2879001, "vi2f.t      C010, C400, 7", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2879081, "vi2f.q      C010, C400, 7", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("7"), None, None]),
        TestEntry::new_r4000allegrex(0xD2881001, "vi2f.s      S010, S400, 8", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2881081, "vi2f.p      C010, C400, 8", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2889001, "vi2f.t      C010, C400, 8", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2889081, "vi2f.q      C010, C400, 8", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("8"), None, None]),
        TestEntry::new_r4000allegrex(0xD2891001, "vi2f.s      S010, S400, 9", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2891081, "vi2f.p      C010, C400, 9", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2899001, "vi2f.t      C010, C400, 9", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD2899081, "vi2f.q      C010, C400, 9", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("9"), None, None]),
        TestEntry::new_r4000allegrex(0xD28A1001, "vi2f.s      S010, S400, 10", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD28A1081, "vi2f.p      C010, C400, 10", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD28A9001, "vi2f.t      C010, C400, 10", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD28A9081, "vi2f.q      C010, C400, 10", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("10"), None, None]),
        TestEntry::new_r4000allegrex(0xD28B1001, "vi2f.s      S010, S400, 11", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD28B1081, "vi2f.p      C010, C400, 11", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD28B9001, "vi2f.t      C010, C400, 11", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD28B9081, "vi2f.q      C010, C400, 11", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("11"), None, None]),
        TestEntry::new_r4000allegrex(0xD28C1001, "vi2f.s      S010, S400, 12", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD28C1081, "vi2f.p      C010, C400, 12", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD28C9001, "vi2f.t      C010, C400, 12", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD28C9081, "vi2f.q      C010, C400, 12", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("12"), None, None]),
        TestEntry::new_r4000allegrex(0xD28D1001, "vi2f.s      S010, S400, 13", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD28D1081, "vi2f.p      C010, C400, 13", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD28D9001, "vi2f.t      C010, C400, 13", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD28D9081, "vi2f.q      C010, C400, 13", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("13"), None, None]),
        TestEntry::new_r4000allegrex(0xD28E1001, "vi2f.s      S010, S400, 14", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD28E1081, "vi2f.p      C010, C400, 14", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD28E9001, "vi2f.t      C010, C400, 14", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD28E9081, "vi2f.q      C010, C400, 14", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("14"), None, None]),
        TestEntry::new_r4000allegrex(0xD28F1001, "vi2f.s      S010, S400, 15", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD28F1081, "vi2f.p      C010, C400, 15", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD28F9001, "vi2f.t      C010, C400, 15", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD28F9081, "vi2f.q      C010, C400, 15", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("15"), None, None]),
        TestEntry::new_r4000allegrex(0xD2901001, "vi2f.s      S010, S400, 16", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2901081, "vi2f.p      C010, C400, 16", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2909001, "vi2f.t      C010, C400, 16", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2909081, "vi2f.q      C010, C400, 16", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("16"), None, None]),
        TestEntry::new_r4000allegrex(0xD2911001, "vi2f.s      S010, S400, 17", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2911081, "vi2f.p      C010, C400, 17", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2919001, "vi2f.t      C010, C400, 17", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2919081, "vi2f.q      C010, C400, 17", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("17"), None, None]),
        TestEntry::new_r4000allegrex(0xD2921001, "vi2f.s      S010, S400, 18", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2921081, "vi2f.p      C010, C400, 18", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2929001, "vi2f.t      C010, C400, 18", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2929081, "vi2f.q      C010, C400, 18", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("18"), None, None]),
        TestEntry::new_r4000allegrex(0xD2931001, "vi2f.s      S010, S400, 19", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2931081, "vi2f.p      C010, C400, 19", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2939001, "vi2f.t      C010, C400, 19", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2939081, "vi2f.q      C010, C400, 19", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("19"), None, None]),
        TestEntry::new_r4000allegrex(0xD2941001, "vi2f.s      S010, S400, 20", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2941081, "vi2f.p      C010, C400, 20", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2949001, "vi2f.t      C010, C400, 20", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2949081, "vi2f.q      C010, C400, 20", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("20"), None, None]),
        TestEntry::new_r4000allegrex(0xD2951001, "vi2f.s      S010, S400, 21", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2951081, "vi2f.p      C010, C400, 21", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2959001, "vi2f.t      C010, C400, 21", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2959081, "vi2f.q      C010, C400, 21", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("21"), None, None]),
        TestEntry::new_r4000allegrex(0xD2961001, "vi2f.s      S010, S400, 22", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2961081, "vi2f.p      C010, C400, 22", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2969001, "vi2f.t      C010, C400, 22", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2969081, "vi2f.q      C010, C400, 22", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("22"), None, None]),
        TestEntry::new_r4000allegrex(0xD2971001, "vi2f.s      S010, S400, 23", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2971081, "vi2f.p      C010, C400, 23", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2979001, "vi2f.t      C010, C400, 23", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2979081, "vi2f.q      C010, C400, 23", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("23"), None, None]),
        TestEntry::new_r4000allegrex(0xD2981001, "vi2f.s      S010, S400, 24", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2981081, "vi2f.p      C010, C400, 24", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2989001, "vi2f.t      C010, C400, 24", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2989081, "vi2f.q      C010, C400, 24", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("24"), None, None]),
        TestEntry::new_r4000allegrex(0xD2991001, "vi2f.s      S010, S400, 25", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2991081, "vi2f.p      C010, C400, 25", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2999001, "vi2f.t      C010, C400, 25", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD2999081, "vi2f.q      C010, C400, 25", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("25"), None, None]),
        TestEntry::new_r4000allegrex(0xD29A1001, "vi2f.s      S010, S400, 26", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD29A1081, "vi2f.p      C010, C400, 26", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD29A9001, "vi2f.t      C010, C400, 26", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD29A9081, "vi2f.q      C010, C400, 26", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("26"), None, None]),
        TestEntry::new_r4000allegrex(0xD29B1001, "vi2f.s      S010, S400, 27", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD29B1081, "vi2f.p      C010, C400, 27", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD29B9001, "vi2f.t      C010, C400, 27", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD29B9081, "vi2f.q      C010, C400, 27", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("27"), None, None]),
        TestEntry::new_r4000allegrex(0xD29C1001, "vi2f.s      S010, S400, 28", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD29C1081, "vi2f.p      C010, C400, 28", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD29C9001, "vi2f.t      C010, C400, 28", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD29C9081, "vi2f.q      C010, C400, 28", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("28"), None, None]),
        TestEntry::new_r4000allegrex(0xD29D1001, "vi2f.s      S010, S400, 29", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD29D1081, "vi2f.p      C010, C400, 29", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD29D9001, "vi2f.t      C010, C400, 29", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD29D9081, "vi2f.q      C010, C400, 29", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("29"), None, None]),
        TestEntry::new_r4000allegrex(0xD29E1001, "vi2f.s      S010, S400, 30", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD29E1081, "vi2f.p      C010, C400, 30", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD29E9001, "vi2f.t      C010, C400, 30", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD29E9081, "vi2f.q      C010, C400, 30", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("30"), None, None]),
        TestEntry::new_r4000allegrex(0xD29F1001, "vi2f.s      S010, S400, 31", Opcode::r4000allegrex_vi2f_s, "vi2f.s", [Some("S010"), Some("S400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD29F1081, "vi2f.p      C010, C400, 31", Opcode::r4000allegrex_vi2f_p, "vi2f.p", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD29F9001, "vi2f.t      C010, C400, 31", Opcode::r4000allegrex_vi2f_t, "vi2f.t", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD29F9081, "vi2f.q      C010, C400, 31", Opcode::r4000allegrex_vi2f_q, "vi2f.q", [Some("C010"), Some("C400"), Some("31"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A01001, "vcmovt.s    S010, S400, 0", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A01081, "vcmovt.p    C010, C400, 0", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A09001, "vcmovt.t    C010, C400, 0", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A09081, "vcmovt.q    C010, C400, 0", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A11001, "vcmovt.s    S010, S400, 1", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A11081, "vcmovt.p    C010, C400, 1", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A19001, "vcmovt.t    C010, C400, 1", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A19081, "vcmovt.q    C010, C400, 1", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A21001, "vcmovt.s    S010, S400, 2", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A21081, "vcmovt.p    C010, C400, 2", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A29001, "vcmovt.t    C010, C400, 2", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A29081, "vcmovt.q    C010, C400, 2", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A31001, "vcmovt.s    S010, S400, 3", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A31081, "vcmovt.p    C010, C400, 3", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A39001, "vcmovt.t    C010, C400, 3", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A39081, "vcmovt.q    C010, C400, 3", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A41001, "vcmovt.s    S010, S400, 4", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A41081, "vcmovt.p    C010, C400, 4", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A49001, "vcmovt.t    C010, C400, 4", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A49081, "vcmovt.q    C010, C400, 4", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A51001, "vcmovt.s    S010, S400, 5", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A51081, "vcmovt.p    C010, C400, 5", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A59001, "vcmovt.t    C010, C400, 5", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A59081, "vcmovt.q    C010, C400, 5", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A61001, "vcmovt.s    S010, S400, 6", Opcode::r4000allegrex_vcmovt_s, "vcmovt.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A61081, "vcmovt.p    C010, C400, 6", Opcode::r4000allegrex_vcmovt_p, "vcmovt.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A69001, "vcmovt.t    C010, C400, 6", Opcode::r4000allegrex_vcmovt_t, "vcmovt.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A69081, "vcmovt.q    C010, C400, 6", Opcode::r4000allegrex_vcmovt_q, "vcmovt.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A81001, "vcmovf.s    S010, S400, 0", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A81081, "vcmovf.p    C010, C400, 0", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A89001, "vcmovf.t    C010, C400, 0", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A89081, "vcmovf.q    C010, C400, 0", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("0"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A91001, "vcmovf.s    S010, S400, 1", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A91081, "vcmovf.p    C010, C400, 1", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A99001, "vcmovf.t    C010, C400, 1", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2A99081, "vcmovf.q    C010, C400, 1", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("1"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AA1001, "vcmovf.s    S010, S400, 2", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AA1081, "vcmovf.p    C010, C400, 2", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AA9001, "vcmovf.t    C010, C400, 2", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AA9081, "vcmovf.q    C010, C400, 2", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("2"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AB1001, "vcmovf.s    S010, S400, 3", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AB1081, "vcmovf.p    C010, C400, 3", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AB9001, "vcmovf.t    C010, C400, 3", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AB9081, "vcmovf.q    C010, C400, 3", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("3"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AC1001, "vcmovf.s    S010, S400, 4", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AC1081, "vcmovf.p    C010, C400, 4", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AC9001, "vcmovf.t    C010, C400, 4", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AC9081, "vcmovf.q    C010, C400, 4", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("4"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AD1001, "vcmovf.s    S010, S400, 5", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AD1081, "vcmovf.p    C010, C400, 5", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AD9001, "vcmovf.t    C010, C400, 5", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AD9081, "vcmovf.q    C010, C400, 5", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("5"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AE1001, "vcmovf.s    S010, S400, 6", Opcode::r4000allegrex_vcmovf_s, "vcmovf.s", [Some("S010"), Some("S400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AE1081, "vcmovf.p    C010, C400, 6", Opcode::r4000allegrex_vcmovf_p, "vcmovf.p", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AE9001, "vcmovf.t    C010, C400, 6", Opcode::r4000allegrex_vcmovf_t, "vcmovf.t", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry::new_r4000allegrex(0xD2AE9081, "vcmovf.q    C010, C400, 6", Opcode::r4000allegrex_vcmovf_q, "vcmovf.q", [Some("C010"), Some("C400"), Some("6"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0xD2B01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B81001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B81081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B89001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B89081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B91001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B91081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B99001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2B99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2B99081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BA1001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BA1081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BA9001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BA9081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BB1001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BB1081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BB9001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BB9081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BC1001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BC1081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BC9001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BC9081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BD1001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BD1081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BD9001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BD9081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BE1001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BE1081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BE9001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2BE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2BE9081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2_CNDMOVE> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C81001                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C81081                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C89001                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C89081                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C91001                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C91081                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C99001                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2C99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2C99081                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CA1001                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CA1081                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CA9001                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CA9081                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CB1001                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CB1081                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CB9001                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CB9081                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CC1001                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CC1081                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CC9001                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CC9081                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CD1001                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CD1081                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CD9001                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CD9081                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CE1001                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CE1081                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CE9001                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CE9081                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CF1001                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CF1081                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CF9001                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2CF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2CF9081                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D01001                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D01081                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D09001                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D09081                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D11001                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D11081                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D19001                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D19081                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D21001                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D21081                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D29001                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D29081                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D31001                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D31081                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D39001                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D39081                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D41001                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D41081                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D49001                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D49081                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D51001                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D51081                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D59001                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D59081                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D61001                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D61081                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D69001                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D69081                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D71001                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D71081                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D79001                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D79081                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D81001                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D81081                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D89001                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D89081                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D91001                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D91081                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D99001                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2D99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2D99081                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DA1001                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DA1081                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DA9001                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DA9081                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DB1001                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DB1081                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DB9001                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DB9081                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DC1001                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DC1081                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DC9001                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DC9081                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DD1001                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DD1081                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DD9001                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DD9081                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DE1001                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DE1081                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DE9001                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DE9081                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DF1001                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DF1081                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DF9001                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2DF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2DF9081                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E01001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E01081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E09001                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E09081                   /* INVALID / 00001001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E11001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E11081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E19001                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E19081                   /* INVALID / 00011001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E21001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E21081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E29001                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E29081                   /* INVALID / 00021001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E31001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E31081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E39001                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E39081                   /* INVALID / 00031001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E41001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E41081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E49001                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E49081                   /* INVALID / 00041001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E51001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E51081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E59001                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E59081                   /* INVALID / 00051001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E61001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E61081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E69001                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E69081                   /* INVALID / 00061001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E71001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E71081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E79001                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E79081                   /* INVALID / 00071001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E81001                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E81081                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E89001                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E89081                   /* INVALID / 00081001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E91001                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E91081                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E99001                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2E99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2E99081                   /* INVALID / 00091001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EA1001                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EA1081                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EA9001                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EA9081                   /* INVALID / 000A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EB1001                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EB1081                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EB9001                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EB9081                   /* INVALID / 000B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EC1001                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EC1081                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EC9001                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EC9081                   /* INVALID / 000C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2ED1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2ED1001                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2ED1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2ED1081                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2ED9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2ED9001                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2ED9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2ED9081                   /* INVALID / 000D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EE1001                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EE1081                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EE9001                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EE9081                   /* INVALID / 000E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EF1001                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EF1081                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EF9001                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2EF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2EF9081                   /* INVALID / 000F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F01001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F01001                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F01081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F01081                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F09001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F09001                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F09081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F09081                   /* INVALID / 00101001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F11001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F11001                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F11081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F11081                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F19001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F19001                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F19081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F19081                   /* INVALID / 00111001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F21001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F21001                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F21081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F21081                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F29001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F29001                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F29081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F29081                   /* INVALID / 00121001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F31001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F31001                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F31081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F31081                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F39001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F39001                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F39081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F39081                   /* INVALID / 00131001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F41001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F41001                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F41081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F41081                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F49001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F49001                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F49081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F49081                   /* INVALID / 00141001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F51001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F51001                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F51081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F51081                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F59001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F59001                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F59081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F59081                   /* INVALID / 00151001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F61001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F61001                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F61081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F61081                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F69001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F69001                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F69081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F69081                   /* INVALID / 00161001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F71001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F71001                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F71081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F71081                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F79001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F79001                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F79081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F79081                   /* INVALID / 00171001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F81001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F81001                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F81081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F81081                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F89001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F89001                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F89081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F89081                   /* INVALID / 00181001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F91001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F91001                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F91081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F91081                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F99001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F99001                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2F99081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2F99081                   /* INVALID / 00191001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FA1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FA1001                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FA1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FA1081                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FA9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FA9001                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FA9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FA9081                   /* INVALID / 001A1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FB1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FB1001                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FB1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FB1081                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FB9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FB9001                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FB9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FB9081                   /* INVALID / 001B1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FC1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FC1001                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FC1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FC1081                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FC9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FC9001                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FC9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FC9081                   /* INVALID / 001C1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FD1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FD1001                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FD1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FD1081                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FD9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FD9001                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FD9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FD9081                   /* INVALID / 001D1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FE1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FE1001                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FE1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FE1081                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FE9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FE9001                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FE9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FE9081                   /* INVALID / 001E1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FF1001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FF1001                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FF1081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FF1081                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FF9001,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FF9001                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xD2FF9081,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xD2FF9081                   /* INVALID / 001F1001 <OpcodeCategory: R4000ALLEGREX_VFPU4_FMT2> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU4_FMT3() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU4_FMT3 */
        TestEntry::new_r4000allegrex(
            0xD3001001,
            "vwbn.s      S010, S400, 0",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("0"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3011001,
            "vwbn.s      S010, S400, 1",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("1"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3021001,
            "vwbn.s      S010, S400, 2",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("2"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3031001,
            "vwbn.s      S010, S400, 3",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("3"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3041001,
            "vwbn.s      S010, S400, 4",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("4"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3051001,
            "vwbn.s      S010, S400, 5",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("5"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3061001,
            "vwbn.s      S010, S400, 6",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("6"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3071001,
            "vwbn.s      S010, S400, 7",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("7"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3081001,
            "vwbn.s      S010, S400, 8",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("8"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3091001,
            "vwbn.s      S010, S400, 9",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("9"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD30A1001,
            "vwbn.s      S010, S400, 10",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("10"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD30B1001,
            "vwbn.s      S010, S400, 11",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("11"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD30C1001,
            "vwbn.s      S010, S400, 12",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("12"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD30D1001,
            "vwbn.s      S010, S400, 13",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("13"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD30E1001,
            "vwbn.s      S010, S400, 14",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("14"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD30F1001,
            "vwbn.s      S010, S400, 15",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("15"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3101001,
            "vwbn.s      S010, S400, 16",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("16"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3111001,
            "vwbn.s      S010, S400, 17",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("17"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3121001,
            "vwbn.s      S010, S400, 18",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("18"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3131001,
            "vwbn.s      S010, S400, 19",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("19"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3141001,
            "vwbn.s      S010, S400, 20",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("20"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3151001,
            "vwbn.s      S010, S400, 21",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("21"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3161001,
            "vwbn.s      S010, S400, 22",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("22"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3171001,
            "vwbn.s      S010, S400, 23",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("23"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3181001,
            "vwbn.s      S010, S400, 24",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("24"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3191001,
            "vwbn.s      S010, S400, 25",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("25"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD31A1001,
            "vwbn.s      S010, S400, 26",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("26"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD31B1001,
            "vwbn.s      S010, S400, 27",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("27"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD31C1001,
            "vwbn.s      S010, S400, 28",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("28"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD31D1001,
            "vwbn.s      S010, S400, 29",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("29"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD31E1001,
            "vwbn.s      S010, S400, 30",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("30"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD31F1001,
            "vwbn.s      S010, S400, 31",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("31"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3201001,
            "vwbn.s      S010, S400, 32",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("32"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3211001,
            "vwbn.s      S010, S400, 33",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("33"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3221001,
            "vwbn.s      S010, S400, 34",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("34"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3231001,
            "vwbn.s      S010, S400, 35",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("35"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3241001,
            "vwbn.s      S010, S400, 36",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("36"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3251001,
            "vwbn.s      S010, S400, 37",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("37"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3261001,
            "vwbn.s      S010, S400, 38",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("38"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3271001,
            "vwbn.s      S010, S400, 39",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("39"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3281001,
            "vwbn.s      S010, S400, 40",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("40"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3291001,
            "vwbn.s      S010, S400, 41",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("41"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD32A1001,
            "vwbn.s      S010, S400, 42",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("42"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD32B1001,
            "vwbn.s      S010, S400, 43",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("43"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD32C1001,
            "vwbn.s      S010, S400, 44",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("44"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD32D1001,
            "vwbn.s      S010, S400, 45",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("45"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD32E1001,
            "vwbn.s      S010, S400, 46",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("46"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD32F1001,
            "vwbn.s      S010, S400, 47",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("47"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3301001,
            "vwbn.s      S010, S400, 48",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("48"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3311001,
            "vwbn.s      S010, S400, 49",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("49"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3321001,
            "vwbn.s      S010, S400, 50",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("50"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3331001,
            "vwbn.s      S010, S400, 51",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("51"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3341001,
            "vwbn.s      S010, S400, 52",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("52"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3351001,
            "vwbn.s      S010, S400, 53",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("53"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3361001,
            "vwbn.s      S010, S400, 54",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("54"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3371001,
            "vwbn.s      S010, S400, 55",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("55"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3381001,
            "vwbn.s      S010, S400, 56",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("56"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3391001,
            "vwbn.s      S010, S400, 57",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("57"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD33A1001,
            "vwbn.s      S010, S400, 58",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("58"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD33B1001,
            "vwbn.s      S010, S400, 59",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("59"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD33C1001,
            "vwbn.s      S010, S400, 60",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("60"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD33D1001,
            "vwbn.s      S010, S400, 61",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("61"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD33E1001,
            "vwbn.s      S010, S400, 62",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("62"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD33F1001,
            "vwbn.s      S010, S400, 63",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("63"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3401001,
            "vwbn.s      S010, S400, 64",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("64"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3411001,
            "vwbn.s      S010, S400, 65",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("65"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3421001,
            "vwbn.s      S010, S400, 66",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("66"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3431001,
            "vwbn.s      S010, S400, 67",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("67"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3441001,
            "vwbn.s      S010, S400, 68",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("68"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3451001,
            "vwbn.s      S010, S400, 69",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("69"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3461001,
            "vwbn.s      S010, S400, 70",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("70"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3471001,
            "vwbn.s      S010, S400, 71",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("71"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3481001,
            "vwbn.s      S010, S400, 72",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("72"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3491001,
            "vwbn.s      S010, S400, 73",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("73"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD34A1001,
            "vwbn.s      S010, S400, 74",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("74"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD34B1001,
            "vwbn.s      S010, S400, 75",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("75"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD34C1001,
            "vwbn.s      S010, S400, 76",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("76"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD34D1001,
            "vwbn.s      S010, S400, 77",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("77"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD34E1001,
            "vwbn.s      S010, S400, 78",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("78"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD34F1001,
            "vwbn.s      S010, S400, 79",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("79"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3501001,
            "vwbn.s      S010, S400, 80",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("80"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3511001,
            "vwbn.s      S010, S400, 81",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("81"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3521001,
            "vwbn.s      S010, S400, 82",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("82"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3531001,
            "vwbn.s      S010, S400, 83",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("83"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3541001,
            "vwbn.s      S010, S400, 84",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("84"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3551001,
            "vwbn.s      S010, S400, 85",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("85"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3561001,
            "vwbn.s      S010, S400, 86",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("86"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3571001,
            "vwbn.s      S010, S400, 87",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("87"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3581001,
            "vwbn.s      S010, S400, 88",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("88"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3591001,
            "vwbn.s      S010, S400, 89",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("89"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD35A1001,
            "vwbn.s      S010, S400, 90",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("90"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD35B1001,
            "vwbn.s      S010, S400, 91",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("91"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD35C1001,
            "vwbn.s      S010, S400, 92",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("92"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD35D1001,
            "vwbn.s      S010, S400, 93",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("93"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD35E1001,
            "vwbn.s      S010, S400, 94",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("94"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD35F1001,
            "vwbn.s      S010, S400, 95",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("95"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3601001,
            "vwbn.s      S010, S400, 96",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("96"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3611001,
            "vwbn.s      S010, S400, 97",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("97"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3621001,
            "vwbn.s      S010, S400, 98",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("98"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3631001,
            "vwbn.s      S010, S400, 99",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("99"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3641001,
            "vwbn.s      S010, S400, 100",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("100"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3651001,
            "vwbn.s      S010, S400, 101",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("101"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3661001,
            "vwbn.s      S010, S400, 102",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("102"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3671001,
            "vwbn.s      S010, S400, 103",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("103"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3681001,
            "vwbn.s      S010, S400, 104",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("104"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3691001,
            "vwbn.s      S010, S400, 105",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("105"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD36A1001,
            "vwbn.s      S010, S400, 106",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("106"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD36B1001,
            "vwbn.s      S010, S400, 107",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("107"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD36C1001,
            "vwbn.s      S010, S400, 108",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("108"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD36D1001,
            "vwbn.s      S010, S400, 109",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("109"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD36E1001,
            "vwbn.s      S010, S400, 110",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("110"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD36F1001,
            "vwbn.s      S010, S400, 111",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("111"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3701001,
            "vwbn.s      S010, S400, 112",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("112"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3711001,
            "vwbn.s      S010, S400, 113",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("113"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3721001,
            "vwbn.s      S010, S400, 114",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("114"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3731001,
            "vwbn.s      S010, S400, 115",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("115"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3741001,
            "vwbn.s      S010, S400, 116",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("116"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3751001,
            "vwbn.s      S010, S400, 117",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("117"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3761001,
            "vwbn.s      S010, S400, 118",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("118"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3771001,
            "vwbn.s      S010, S400, 119",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("119"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3781001,
            "vwbn.s      S010, S400, 120",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("120"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3791001,
            "vwbn.s      S010, S400, 121",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("121"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD37A1001,
            "vwbn.s      S010, S400, 122",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("122"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD37B1001,
            "vwbn.s      S010, S400, 123",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("123"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD37C1001,
            "vwbn.s      S010, S400, 124",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("124"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD37D1001,
            "vwbn.s      S010, S400, 125",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("125"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD37E1001,
            "vwbn.s      S010, S400, 126",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("126"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD37F1001,
            "vwbn.s      S010, S400, 127",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("127"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3801001,
            "vwbn.s      S010, S400, 128",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("128"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3811001,
            "vwbn.s      S010, S400, 129",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("129"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3821001,
            "vwbn.s      S010, S400, 130",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("130"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3831001,
            "vwbn.s      S010, S400, 131",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("131"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3841001,
            "vwbn.s      S010, S400, 132",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("132"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3851001,
            "vwbn.s      S010, S400, 133",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("133"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3861001,
            "vwbn.s      S010, S400, 134",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("134"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3871001,
            "vwbn.s      S010, S400, 135",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("135"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3881001,
            "vwbn.s      S010, S400, 136",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("136"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3891001,
            "vwbn.s      S010, S400, 137",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("137"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD38A1001,
            "vwbn.s      S010, S400, 138",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("138"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD38B1001,
            "vwbn.s      S010, S400, 139",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("139"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD38C1001,
            "vwbn.s      S010, S400, 140",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("140"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD38D1001,
            "vwbn.s      S010, S400, 141",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("141"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD38E1001,
            "vwbn.s      S010, S400, 142",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("142"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD38F1001,
            "vwbn.s      S010, S400, 143",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("143"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3901001,
            "vwbn.s      S010, S400, 144",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("144"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3911001,
            "vwbn.s      S010, S400, 145",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("145"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3921001,
            "vwbn.s      S010, S400, 146",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("146"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3931001,
            "vwbn.s      S010, S400, 147",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("147"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3941001,
            "vwbn.s      S010, S400, 148",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("148"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3951001,
            "vwbn.s      S010, S400, 149",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("149"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3961001,
            "vwbn.s      S010, S400, 150",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("150"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3971001,
            "vwbn.s      S010, S400, 151",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("151"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3981001,
            "vwbn.s      S010, S400, 152",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("152"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3991001,
            "vwbn.s      S010, S400, 153",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("153"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD39A1001,
            "vwbn.s      S010, S400, 154",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("154"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD39B1001,
            "vwbn.s      S010, S400, 155",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("155"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD39C1001,
            "vwbn.s      S010, S400, 156",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("156"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD39D1001,
            "vwbn.s      S010, S400, 157",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("157"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD39E1001,
            "vwbn.s      S010, S400, 158",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("158"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD39F1001,
            "vwbn.s      S010, S400, 159",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("159"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A01001,
            "vwbn.s      S010, S400, 160",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("160"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A11001,
            "vwbn.s      S010, S400, 161",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("161"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A21001,
            "vwbn.s      S010, S400, 162",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("162"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A31001,
            "vwbn.s      S010, S400, 163",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("163"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A41001,
            "vwbn.s      S010, S400, 164",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("164"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A51001,
            "vwbn.s      S010, S400, 165",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("165"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A61001,
            "vwbn.s      S010, S400, 166",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("166"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A71001,
            "vwbn.s      S010, S400, 167",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("167"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A81001,
            "vwbn.s      S010, S400, 168",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("168"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3A91001,
            "vwbn.s      S010, S400, 169",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("169"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3AA1001,
            "vwbn.s      S010, S400, 170",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("170"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3AB1001,
            "vwbn.s      S010, S400, 171",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("171"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3AC1001,
            "vwbn.s      S010, S400, 172",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("172"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3AD1001,
            "vwbn.s      S010, S400, 173",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("173"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3AE1001,
            "vwbn.s      S010, S400, 174",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("174"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3AF1001,
            "vwbn.s      S010, S400, 175",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("175"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B01001,
            "vwbn.s      S010, S400, 176",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("176"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B11001,
            "vwbn.s      S010, S400, 177",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("177"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B21001,
            "vwbn.s      S010, S400, 178",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("178"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B31001,
            "vwbn.s      S010, S400, 179",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("179"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B41001,
            "vwbn.s      S010, S400, 180",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("180"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B51001,
            "vwbn.s      S010, S400, 181",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("181"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B61001,
            "vwbn.s      S010, S400, 182",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("182"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B71001,
            "vwbn.s      S010, S400, 183",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("183"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B81001,
            "vwbn.s      S010, S400, 184",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("184"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3B91001,
            "vwbn.s      S010, S400, 185",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("185"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3BA1001,
            "vwbn.s      S010, S400, 186",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("186"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3BB1001,
            "vwbn.s      S010, S400, 187",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("187"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3BC1001,
            "vwbn.s      S010, S400, 188",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("188"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3BD1001,
            "vwbn.s      S010, S400, 189",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("189"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3BE1001,
            "vwbn.s      S010, S400, 190",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("190"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3BF1001,
            "vwbn.s      S010, S400, 191",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("191"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C01001,
            "vwbn.s      S010, S400, 192",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("192"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C11001,
            "vwbn.s      S010, S400, 193",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("193"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C21001,
            "vwbn.s      S010, S400, 194",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("194"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C31001,
            "vwbn.s      S010, S400, 195",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("195"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C41001,
            "vwbn.s      S010, S400, 196",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("196"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C51001,
            "vwbn.s      S010, S400, 197",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("197"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C61001,
            "vwbn.s      S010, S400, 198",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("198"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C71001,
            "vwbn.s      S010, S400, 199",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("199"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C81001,
            "vwbn.s      S010, S400, 200",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("200"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3C91001,
            "vwbn.s      S010, S400, 201",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("201"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3CA1001,
            "vwbn.s      S010, S400, 202",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("202"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3CB1001,
            "vwbn.s      S010, S400, 203",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("203"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3CC1001,
            "vwbn.s      S010, S400, 204",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("204"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3CD1001,
            "vwbn.s      S010, S400, 205",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("205"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3CE1001,
            "vwbn.s      S010, S400, 206",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("206"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3CF1001,
            "vwbn.s      S010, S400, 207",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("207"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D01001,
            "vwbn.s      S010, S400, 208",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("208"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D11001,
            "vwbn.s      S010, S400, 209",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("209"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D21001,
            "vwbn.s      S010, S400, 210",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("210"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D31001,
            "vwbn.s      S010, S400, 211",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("211"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D41001,
            "vwbn.s      S010, S400, 212",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("212"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D51001,
            "vwbn.s      S010, S400, 213",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("213"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D61001,
            "vwbn.s      S010, S400, 214",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("214"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D71001,
            "vwbn.s      S010, S400, 215",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("215"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D81001,
            "vwbn.s      S010, S400, 216",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("216"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3D91001,
            "vwbn.s      S010, S400, 217",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("217"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3DA1001,
            "vwbn.s      S010, S400, 218",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("218"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3DB1001,
            "vwbn.s      S010, S400, 219",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("219"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3DC1001,
            "vwbn.s      S010, S400, 220",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("220"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3DD1001,
            "vwbn.s      S010, S400, 221",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("221"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3DE1001,
            "vwbn.s      S010, S400, 222",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("222"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3DF1001,
            "vwbn.s      S010, S400, 223",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("223"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E01001,
            "vwbn.s      S010, S400, 224",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("224"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E11001,
            "vwbn.s      S010, S400, 225",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("225"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E21001,
            "vwbn.s      S010, S400, 226",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("226"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E31001,
            "vwbn.s      S010, S400, 227",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("227"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E41001,
            "vwbn.s      S010, S400, 228",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("228"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E51001,
            "vwbn.s      S010, S400, 229",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("229"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E61001,
            "vwbn.s      S010, S400, 230",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("230"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E71001,
            "vwbn.s      S010, S400, 231",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("231"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E81001,
            "vwbn.s      S010, S400, 232",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("232"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3E91001,
            "vwbn.s      S010, S400, 233",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("233"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3EA1001,
            "vwbn.s      S010, S400, 234",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("234"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3EB1001,
            "vwbn.s      S010, S400, 235",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("235"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3EC1001,
            "vwbn.s      S010, S400, 236",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("236"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3ED1001,
            "vwbn.s      S010, S400, 237",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("237"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3EE1001,
            "vwbn.s      S010, S400, 238",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("238"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3EF1001,
            "vwbn.s      S010, S400, 239",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("239"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F01001,
            "vwbn.s      S010, S400, 240",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("240"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F11001,
            "vwbn.s      S010, S400, 241",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("241"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F21001,
            "vwbn.s      S010, S400, 242",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("242"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F31001,
            "vwbn.s      S010, S400, 243",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("243"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F41001,
            "vwbn.s      S010, S400, 244",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("244"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F51001,
            "vwbn.s      S010, S400, 245",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("245"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F61001,
            "vwbn.s      S010, S400, 246",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("246"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F71001,
            "vwbn.s      S010, S400, 247",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("247"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F81001,
            "vwbn.s      S010, S400, 248",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("248"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3F91001,
            "vwbn.s      S010, S400, 249",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("249"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3FA1001,
            "vwbn.s      S010, S400, 250",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("250"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3FB1001,
            "vwbn.s      S010, S400, 251",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("251"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3FC1001,
            "vwbn.s      S010, S400, 252",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("252"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3FD1001,
            "vwbn.s      S010, S400, 253",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("253"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3FE1001,
            "vwbn.s      S010, S400, 254",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("254"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xD3FF1001,
            "vwbn.s      S010, S400, 255",
            Opcode::r4000allegrex_vwbn_s,
            "vwbn.s",
            [Some("S010"), Some("S400"), Some("255"), None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU5() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU5 */
        TestEntry::new_r4000allegrex(
            0xDC000000,
            "vpfxs       X, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000000,
            "vpfxt       X, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000000,
            "vpfxd       , , , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF400000,
            "viim.s      S002, 0",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("0"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF400001,
            "viim.s      S002, 1",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("1"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF400010,
            "viim.s      S002, 16",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("16"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF400100,
            "viim.s      S002, 256",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("256"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF401000,
            "viim.s      S002, 4096",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("4096"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF408000,
            "viim.s      S002, -32768",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("-32768"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDF40F000,
            "viim.s      S002, -4096",
            Opcode::r4000allegrex_viim_s,
            "viim.s",
            [Some("S002"), Some("-4096"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC00000,
            "vfim.s      S002, 0",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("0"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC08000,
            "vfim.s      S002, -0",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("-0"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC01000,
            "vfim.s      S002, 0.00048828125",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("0.00048828125"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC0F000,
            "vfim.s      S002, -8192",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("-8192"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC05450,
            "vfim.s      S002, 69",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("69"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC0D450,
            "vfim.s      S002, -69",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("-69"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC03C00,
            "vfim.s      S002, 1",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("1"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC0BC00,
            "vfim.s      S002, -1",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("-1"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC07C00,
            "vfim.s      S002, inf",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("inf"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC0FC00,
            "vfim.s      S002, -inf",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("-inf"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC07BFF,
            "vfim.s      S002, 65504",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("65504"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC0FBFF,
            "vfim.s      S002, -65504",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("-65504"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC07EAA,
            "vfim.s      S002, NaN",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("NaN"), None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xDFC0FEAA,
            "vfim.s      S002, NaN",
            Opcode::r4000allegrex_vfim_s,
            "vfim.s",
            [Some("S002"), Some("NaN"), None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU6() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU6 */
        TestEntry {
            instr: Instruction::new(
                0xF0001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF00010C0, "vmmul.p     M002, E400, M000", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M002"), Some("E400"), Some("M000"), None, None]),
        TestEntry::new_r4000allegrex(0xF0009004, "vmmul.t     M100, E400, M000", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M100"), Some("E400"), Some("M000"), None, None]),
        TestEntry::new_r4000allegrex(0xF00090C0, "vmmul.q     M002, E400, M000", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M002"), Some("E400"), Some("M000"), None, None]),
        TestEntry::new_r4000allegrex(0xF0800440, "vhtfm2.p    C002, M100, C000", Opcode::r4000allegrex_vhtfm2_p, "vhtfm2.p", [Some("C002"), Some("M100"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF08010C0, "vtfm2.p     C002, M400, C000", Opcode::r4000allegrex_vtfm2_p, "vtfm2.p", [Some("C002"), Some("M400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF0801084, "vtfm2.p     C100, M400, C000", Opcode::r4000allegrex_vtfm2_p, "vtfm2.p", [Some("C100"), Some("M400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF1008004, "vtfm3.t     C100, M000, C000", Opcode::r4000allegrex_vtfm3_t, "vtfm3.t", [Some("C100"), Some("M000"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF1001084, "vhtfm3.t    C100, M400, C000", Opcode::r4000allegrex_vhtfm3_t, "vhtfm3.t", [Some("C100"), Some("M400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF1009004, "vtfm3.t     C100, M400, C000", Opcode::r4000allegrex_vtfm3_t, "vtfm3.t", [Some("C100"), Some("M400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF1808088, "vtfm4.q     C200, M000, C000", Opcode::r4000allegrex_vtfm4_q, "vtfm4.q", [Some("C200"), Some("M000"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF1809040, "vhtfm4.q    C002, M400, C000", Opcode::r4000allegrex_vhtfm4_q, "vhtfm4.q", [Some("C002"), Some("M400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF18090C0, "vtfm4.q     C002, M400, C000", Opcode::r4000allegrex_vtfm4_q, "vtfm4.q", [Some("C002"), Some("M400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0xF2001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF2001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF20010C0, "vmscl.p     M002, M400, S000", Opcode::r4000allegrex_vmscl_p, "vmscl.p", [Some("M002"), Some("M400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0xF2009040, "vmscl.t     M001, M400, S000", Opcode::r4000allegrex_vmscl_t, "vmscl.t", [Some("M001"), Some("M400"), Some("S000"), None, None]),
        TestEntry::new_r4000allegrex(0xF20090C0, "vmscl.q     M002, M400, S000", Opcode::r4000allegrex_vmscl_q, "vmscl.q", [Some("M002"), Some("M400"), Some("S000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0xF2801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF2801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF28010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF28010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF2809004, "vcrsp.t     C100, C400, C000", Opcode::r4000allegrex_vcrsp_t, "vcrsp.t", [Some("C100"), Some("C400"), Some("C000"), None, None]),
        TestEntry::new_r4000allegrex(0xF28090C0, "vqmul.q     C002, C400, C000", Opcode::r4000allegrex_vqmul_q, "vqmul.q", [Some("C002"), Some("C400"), Some("C000"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0xF3001040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3001040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF30010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF30010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3009040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3009040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF30090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF30090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3801040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3801040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF38010C0, "vmmov.p     M002, M400", Opcode::r4000allegrex_vmmov_p, "vmmov.p", [Some("M002"), Some("M400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xF3809040, "vmmov.t     M001, M400", Opcode::r4000allegrex_vmmov_t, "vmmov.t", [Some("M001"), Some("M400"), None, None, None]),
        TestEntry::new_r4000allegrex(0xF38090C0, "vmmov.q     M002, M400", Opcode::r4000allegrex_vmmov_q, "vmmov.q", [Some("M002"), Some("M400"), None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xF3811040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3811040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38110C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38110C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3819040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3819040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38190C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38190C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3821040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3821040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38210C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38210C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3829040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3829040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38290C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38290C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3831040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3831040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF38300C0, "vmidt.p     M002", Opcode::r4000allegrex_vmidt_p, "vmidt.p", [Some("M002"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xF3838040, "vmidt.t     M001", Opcode::r4000allegrex_vmidt_t, "vmidt.t", [Some("M001"), None, None, None, None]),
        TestEntry::new_r4000allegrex(0xF38380C0, "vmidt.q     M002", Opcode::r4000allegrex_vmidt_q, "vmidt.q", [Some("M002"), None, None, None, None]),
        TestEntry {
            instr: Instruction::new(
                0xF3841040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3841040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38410C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38410C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3849040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3849040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38490C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38490C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3851040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3851040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38510C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38510C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3859040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3859040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38590C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38590C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3861040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3861040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38610C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38610C0                   /* vmzero.p    M002 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vmzero_p,
            opcode_str: "vmzero.p",
            operands_str: [Some("M002"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3869040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3869040                   /* vmzero.t    M001 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vmzero_t,
            opcode_str: "vmzero.t",
            operands_str: [Some("M001"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38690C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38690C0                   /* vmzero.q    M002 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vmzero_q,
            opcode_str: "vmzero.q",
            operands_str: [Some("M002"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3871040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3871040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38710C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38710C0                   /* vmone.p     M002 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vmone_p,
            opcode_str: "vmone.p",
            operands_str: [Some("M002"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3879040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3879040                   /* vmone.t     M001 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vmone_t,
            opcode_str: "vmone.t",
            operands_str: [Some("M001"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF38790C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF38790C0                   /* vmone.q     M002 / 00001000 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7_FMT0> */",
            expected_opcode: Opcode::r4000allegrex_vmone_q,
            opcode_str: "vmone.q",
            operands_str: [Some("M002"), None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3A01040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3A01040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF3A010C0, "vrot.p      C002, S400, [C,S]", Opcode::r4000allegrex_vrot_p, "vrot.p", [Some("C002"), Some("S400"), Some("[C,S]"), None, None]),
        TestEntry::new_r4000allegrex(0xF3A09040, "vrot.t      C001, S400, [C,S,S]", Opcode::r4000allegrex_vrot_t, "vrot.t", [Some("C001"), Some("S400"), Some("[C,S,S]"), None, None]),
        TestEntry::new_r4000allegrex(0xF3A090C0, "vrot.q      C002, S400, [C,S,S,S]", Opcode::r4000allegrex_vrot_q, "vrot.q", [Some("C002"), Some("S400"), Some("[C,S,S,S]"), None, None]),
        TestEntry {
            instr: Instruction::new(
                0xF3C01040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3C01040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3C010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3C010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3C09040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3C09040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3C090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3C090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3E01040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3E01040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3E010C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3E010C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3E09040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3E09040                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF3E090C0,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF3E090C0                   /* INVALID / 00001040 <OpcodeCategory: R4000ALLEGREX_VFPU6_FMT7> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[allow(non_snake_case)]
#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_VFPU7() {
    const ENTRIES: &[TestEntry] = &[
        /* VFPU7 */
        TestEntry::new_r4000allegrex(
            0xFFFF040D,
            "vflush",
            Opcode::r4000allegrex_vflush,
            "vflush",
            [Some(""), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xFFFF0000,
            "vnop",
            Opcode::r4000allegrex_vnop,
            "vnop",
            [Some(""), None, None, None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xFFFF0320,
            "vsync",
            Opcode::r4000allegrex_vsync,
            "vsync",
            [Some(""), None, None, None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vmmul() {
    const ENTRIES: &[TestEntry] = &[
        /* vmmul */
        TestEntry {
            instr: Instruction::new(
                0xF0000040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000040                   /* INVALID / 00000040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000140                   /* INVALID / 00000140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000240                   /* INVALID / 00000240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000340                   /* INVALID / 00000340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000440                   /* INVALID / 00000440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000540                   /* INVALID / 00000540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000640                   /* INVALID / 00000640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000740                   /* INVALID / 00000740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000840                   /* INVALID / 00000840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000940                   /* INVALID / 00000940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000A40                   /* INVALID / 00000A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000B40                   /* INVALID / 00000B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000C40                   /* INVALID / 00000C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000D40                   /* INVALID / 00000D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000E40                   /* INVALID / 00000E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0000F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0000F40                   /* INVALID / 00000F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001140                   /* INVALID / 00001140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001240                   /* INVALID / 00001240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001340                   /* INVALID / 00001340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001440                   /* INVALID / 00001440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001540                   /* INVALID / 00001540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001640                   /* INVALID / 00001640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001740                   /* INVALID / 00001740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001840                   /* INVALID / 00001840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001940                   /* INVALID / 00001940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001A40                   /* INVALID / 00001A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001B40                   /* INVALID / 00001B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001C40                   /* INVALID / 00001C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001D40                   /* INVALID / 00001D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001E40                   /* INVALID / 00001E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0001F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0001F40                   /* INVALID / 00001F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002040                   /* INVALID / 00002040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002140                   /* INVALID / 00002140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002240                   /* INVALID / 00002240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002340                   /* INVALID / 00002340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002440                   /* INVALID / 00002440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002540                   /* INVALID / 00002540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002640                   /* INVALID / 00002640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002740                   /* INVALID / 00002740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002840                   /* INVALID / 00002840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002940                   /* INVALID / 00002940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002A40                   /* INVALID / 00002A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002B40                   /* INVALID / 00002B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002C40                   /* INVALID / 00002C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002D40                   /* INVALID / 00002D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002E40                   /* INVALID / 00002E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0002F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0002F40                   /* INVALID / 00002F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003040                   /* INVALID / 00003040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003140                   /* INVALID / 00003140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003240                   /* INVALID / 00003240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003340                   /* INVALID / 00003340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003440                   /* INVALID / 00003440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003540                   /* INVALID / 00003540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003640                   /* INVALID / 00003640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003740                   /* INVALID / 00003740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003840                   /* INVALID / 00003840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003940                   /* INVALID / 00003940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003A40                   /* INVALID / 00003A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003B40                   /* INVALID / 00003B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003C40                   /* INVALID / 00003C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003D40                   /* INVALID / 00003D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003E40                   /* INVALID / 00003E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0003F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0003F40                   /* INVALID / 00003F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004040                   /* INVALID / 00004040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004140                   /* INVALID / 00004140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004240                   /* INVALID / 00004240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004340                   /* INVALID / 00004340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004440                   /* INVALID / 00004440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004540                   /* INVALID / 00004540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004640                   /* INVALID / 00004640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004740                   /* INVALID / 00004740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004840                   /* INVALID / 00004840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004940                   /* INVALID / 00004940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004A40                   /* INVALID / 00004A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004B40                   /* INVALID / 00004B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004C40                   /* INVALID / 00004C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004D40                   /* INVALID / 00004D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004E40                   /* INVALID / 00004E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0004F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0004F40                   /* INVALID / 00004F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005040                   /* INVALID / 00005040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005140                   /* INVALID / 00005140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005240                   /* INVALID / 00005240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005340                   /* INVALID / 00005340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005440                   /* INVALID / 00005440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005540                   /* INVALID / 00005540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005640                   /* INVALID / 00005640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005740                   /* INVALID / 00005740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005840                   /* INVALID / 00005840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005940                   /* INVALID / 00005940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005A40                   /* INVALID / 00005A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005B40                   /* INVALID / 00005B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005C40                   /* INVALID / 00005C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005D40                   /* INVALID / 00005D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005E40                   /* INVALID / 00005E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0005F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0005F40                   /* INVALID / 00005F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006040                   /* INVALID / 00006040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006140                   /* INVALID / 00006140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006240                   /* INVALID / 00006240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006340                   /* INVALID / 00006340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006440                   /* INVALID / 00006440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006540                   /* INVALID / 00006540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006640                   /* INVALID / 00006640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006740                   /* INVALID / 00006740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006840                   /* INVALID / 00006840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006940                   /* INVALID / 00006940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006A40                   /* INVALID / 00006A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006B40                   /* INVALID / 00006B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006C40                   /* INVALID / 00006C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006D40                   /* INVALID / 00006D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006E40                   /* INVALID / 00006E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0006F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0006F40                   /* INVALID / 00006F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007040,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007040                   /* INVALID / 00007040 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007140,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007140                   /* INVALID / 00007140 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007240,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007240                   /* INVALID / 00007240 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007340,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007340                   /* INVALID / 00007340 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007440,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007440                   /* INVALID / 00007440 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007540,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007540                   /* INVALID / 00007540 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007640,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007640                   /* INVALID / 00007640 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007740,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007740                   /* INVALID / 00007740 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007840,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007840                   /* INVALID / 00007840 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007940,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007940                   /* INVALID / 00007940 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007A40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007A40                   /* INVALID / 00007A40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007B40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007B40                   /* INVALID / 00007B40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007C40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007C40                   /* INVALID / 00007C40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007D40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007D40                   /* INVALID / 00007D40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007E40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007E40                   /* INVALID / 00007E40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry {
            instr: Instruction::new(
                0xF0007F40,
                Vram::new(0x80000000),
                InstructionFlags::new_extension(IsaExtension::R4000ALLEGREX).with_abi(Abi::EABI64),
            ),
            imm_override: None,
            display_flags: InstructionDisplayFlags::default(),
            valid: false,
            expected: ".word       0xF0007F40                   /* INVALID / 00007F40 <OpcodeCategory: R4000ALLEGREX_VFPU6> */",
            expected_opcode: Opcode::ALL_INVALID,
            opcode_str: "INVALID",
            operands_str: [None, None, None, None, None],
        },
        TestEntry::new_r4000allegrex(0xF004009F, "vmmul.p     M730, E000, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E000"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004019F, "vmmul.p     M730, E001, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E001"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004029F, "vmmul.p     M730, E002, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E002"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004039F, "vmmul.p     M730, E003, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E003"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004049F, "vmmul.p     M730, E100, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E100"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004059F, "vmmul.p     M730, E101, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E101"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004069F, "vmmul.p     M730, E102, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E102"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004079F, "vmmul.p     M730, E103, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E103"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004089F, "vmmul.p     M730, E200, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E200"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004099F, "vmmul.p     M730, E201, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E201"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0040A9F, "vmmul.p     M730, E202, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E202"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0040B9F, "vmmul.p     M730, E203, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E203"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0040C9F, "vmmul.p     M730, E300, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E300"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0040D9F, "vmmul.p     M730, E301, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E301"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0040E9F, "vmmul.p     M730, E302, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E302"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0040F9F, "vmmul.p     M730, E303, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E303"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004109F, "vmmul.p     M730, E400, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E400"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004119F, "vmmul.p     M730, E401, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E401"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004129F, "vmmul.p     M730, E402, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E402"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004139F, "vmmul.p     M730, E403, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E403"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004149F, "vmmul.p     M730, E500, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E500"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004159F, "vmmul.p     M730, E501, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E501"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004169F, "vmmul.p     M730, E502, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E502"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004179F, "vmmul.p     M730, E503, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E503"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004189F, "vmmul.p     M730, E600, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E600"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004199F, "vmmul.p     M730, E601, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E601"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0041A9F, "vmmul.p     M730, E602, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E602"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0041B9F, "vmmul.p     M730, E603, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E603"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0041C88, "vmmul.p     M200, E700, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E700"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0041D88, "vmmul.p     M200, E701, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E701"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0041E88, "vmmul.p     M200, E702, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E702"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0041F88, "vmmul.p     M200, E703, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E703"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004209F, "vmmul.p     M730, M000, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M000"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004219F, "vmmul.p     M730, M010, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M010"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004229F, "vmmul.p     M730, M020, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M020"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004239F, "vmmul.p     M730, M030, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M030"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004249F, "vmmul.p     M730, M100, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M100"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004259F, "vmmul.p     M730, M110, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M110"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004269F, "vmmul.p     M730, M120, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M120"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004279F, "vmmul.p     M730, M130, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M130"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004289F, "vmmul.p     M730, M200, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M200"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004299F, "vmmul.p     M730, M210, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M210"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0042A9F, "vmmul.p     M730, M220, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M220"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0042B9F, "vmmul.p     M730, M230, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M230"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0042C9F, "vmmul.p     M730, M300, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M300"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0042D9F, "vmmul.p     M730, M310, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M310"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0042E9F, "vmmul.p     M730, M320, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M320"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0042F9F, "vmmul.p     M730, M330, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M330"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004309F, "vmmul.p     M730, M400, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M400"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004319F, "vmmul.p     M730, M410, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M410"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004329F, "vmmul.p     M730, M420, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M420"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004339F, "vmmul.p     M730, M430, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M430"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004349F, "vmmul.p     M730, M500, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M500"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004359F, "vmmul.p     M730, M510, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M510"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004369F, "vmmul.p     M730, M520, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M520"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004379F, "vmmul.p     M730, M530, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M530"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004389F, "vmmul.p     M730, M600, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M600"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004399F, "vmmul.p     M730, M610, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M610"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0043A9F, "vmmul.p     M730, M620, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M620"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0043B9F, "vmmul.p     M730, M630, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M630"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0043C88, "vmmul.p     M200, M700, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M700"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0043D88, "vmmul.p     M200, M710, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M710"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0043E88, "vmmul.p     M200, M720, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M720"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0043F88, "vmmul.p     M200, M730, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M730"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004409F, "vmmul.p     M730, E020, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E020"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004419F, "vmmul.p     M730, E021, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E021"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004429F, "vmmul.p     M730, E022, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E022"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004439F, "vmmul.p     M730, E023, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E023"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004449F, "vmmul.p     M730, E120, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E120"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004459F, "vmmul.p     M730, E121, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E121"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004469F, "vmmul.p     M730, E122, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E122"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004479F, "vmmul.p     M730, E123, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E123"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004489F, "vmmul.p     M730, E220, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E220"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004499F, "vmmul.p     M730, E221, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E221"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0044A9F, "vmmul.p     M730, E222, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E222"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0044B9F, "vmmul.p     M730, E223, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E223"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0044C9F, "vmmul.p     M730, E320, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E320"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0044D9F, "vmmul.p     M730, E321, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E321"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0044E9F, "vmmul.p     M730, E322, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E322"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0044F9F, "vmmul.p     M730, E323, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E323"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004509F, "vmmul.p     M730, E420, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E420"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004519F, "vmmul.p     M730, E421, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E421"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004529F, "vmmul.p     M730, E422, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E422"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004539F, "vmmul.p     M730, E423, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E423"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004549F, "vmmul.p     M730, E520, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E520"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004559F, "vmmul.p     M730, E521, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E521"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004569F, "vmmul.p     M730, E522, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E522"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004579F, "vmmul.p     M730, E523, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E523"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004589F, "vmmul.p     M730, E620, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E620"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004599F, "vmmul.p     M730, E621, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E621"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0045A9F, "vmmul.p     M730, E622, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E622"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0045B9F, "vmmul.p     M730, E623, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("E623"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0045C88, "vmmul.p     M200, E720, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E720"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0045D88, "vmmul.p     M200, E721, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E721"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0045E88, "vmmul.p     M200, E722, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E722"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0045F88, "vmmul.p     M200, E723, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("E723"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004609F, "vmmul.p     M730, M002, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M002"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004619F, "vmmul.p     M730, M012, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M012"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004629F, "vmmul.p     M730, M022, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M022"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004639F, "vmmul.p     M730, M032, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M032"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004649F, "vmmul.p     M730, M102, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M102"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004659F, "vmmul.p     M730, M112, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M112"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004669F, "vmmul.p     M730, M122, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M122"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004679F, "vmmul.p     M730, M132, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M132"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004689F, "vmmul.p     M730, M202, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M202"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004699F, "vmmul.p     M730, M212, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M212"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0046A9F, "vmmul.p     M730, M222, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M222"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0046B9F, "vmmul.p     M730, M232, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M232"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0046C9F, "vmmul.p     M730, M302, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M302"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0046D9F, "vmmul.p     M730, M312, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M312"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0046E9F, "vmmul.p     M730, M322, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M322"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0046F9F, "vmmul.p     M730, M332, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M332"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004709F, "vmmul.p     M730, M402, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M402"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004719F, "vmmul.p     M730, M412, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M412"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004729F, "vmmul.p     M730, M422, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M422"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004739F, "vmmul.p     M730, M432, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M432"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004749F, "vmmul.p     M730, M502, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M502"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004759F, "vmmul.p     M730, M512, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M512"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004769F, "vmmul.p     M730, M522, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M522"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004779F, "vmmul.p     M730, M532, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M532"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004789F, "vmmul.p     M730, M602, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M602"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004799F, "vmmul.p     M730, M612, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M612"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0047A9F, "vmmul.p     M730, M622, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M622"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0047B9F, "vmmul.p     M730, M632, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M730"), Some("M632"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0047C88, "vmmul.p     M200, M702, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M702"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0047D88, "vmmul.p     M200, M712, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M712"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0047E88, "vmmul.p     M200, M722, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M722"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0047F88, "vmmul.p     M200, M732, M100", Opcode::r4000allegrex_vmmul_p, "vmmul.p", [Some("M200"), Some("M732"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004801F, "vmmul.t     M730, E000, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E000"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004811F, "vmmul.t     M730, E001, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E001"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004821F, "vmmul.t     M730, E002, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E002"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004831F, "vmmul.t     M730, E003, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E003"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004841F, "vmmul.t     M730, E100, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E100"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004851F, "vmmul.t     M730, E101, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E101"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004861F, "vmmul.t     M730, E102, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E102"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004871F, "vmmul.t     M730, E103, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E103"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004881F, "vmmul.t     M730, E200, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E200"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004891F, "vmmul.t     M730, E201, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E201"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048A1F, "vmmul.t     M730, E202, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E202"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048B1F, "vmmul.t     M730, E203, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E203"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048C1F, "vmmul.t     M730, E300, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E300"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048D1F, "vmmul.t     M730, E301, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E301"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048E1F, "vmmul.t     M730, E302, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E302"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048F1F, "vmmul.t     M730, E303, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E303"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004901F, "vmmul.t     M730, E400, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E400"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004911F, "vmmul.t     M730, E401, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E401"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004921F, "vmmul.t     M730, E402, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E402"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004931F, "vmmul.t     M730, E403, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E403"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004941F, "vmmul.t     M730, E500, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E500"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004951F, "vmmul.t     M730, E501, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E501"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004961F, "vmmul.t     M730, E502, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E502"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004971F, "vmmul.t     M730, E503, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E503"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004981F, "vmmul.t     M730, E600, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E600"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004991F, "vmmul.t     M730, E601, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E601"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049A1F, "vmmul.t     M730, E602, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E602"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049B1F, "vmmul.t     M730, E603, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E603"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049C08, "vmmul.t     M200, E700, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E700"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049D08, "vmmul.t     M200, E701, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E701"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049E08, "vmmul.t     M200, E702, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E702"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049F08, "vmmul.t     M200, E703, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E703"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A01F, "vmmul.t     M730, M000, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M000"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A11F, "vmmul.t     M730, M010, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M010"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A21F, "vmmul.t     M730, M020, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M020"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A31F, "vmmul.t     M730, M030, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M030"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A41F, "vmmul.t     M730, M100, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M100"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A51F, "vmmul.t     M730, M110, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M110"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A61F, "vmmul.t     M730, M120, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M120"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A71F, "vmmul.t     M730, M130, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M130"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A81F, "vmmul.t     M730, M200, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M200"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A91F, "vmmul.t     M730, M210, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M210"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AA1F, "vmmul.t     M730, M220, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M220"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AB1F, "vmmul.t     M730, M230, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M230"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AC1F, "vmmul.t     M730, M300, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M300"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AD1F, "vmmul.t     M730, M310, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M310"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AE1F, "vmmul.t     M730, M320, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M320"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AF1F, "vmmul.t     M730, M330, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M330"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B01F, "vmmul.t     M730, M400, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M400"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B11F, "vmmul.t     M730, M410, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M410"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B21F, "vmmul.t     M730, M420, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M420"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B31F, "vmmul.t     M730, M430, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M430"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B41F, "vmmul.t     M730, M500, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M500"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B51F, "vmmul.t     M730, M510, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M510"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B61F, "vmmul.t     M730, M520, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M520"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B71F, "vmmul.t     M730, M530, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M530"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B81F, "vmmul.t     M730, M600, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M600"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B91F, "vmmul.t     M730, M610, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M610"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BA1F, "vmmul.t     M730, M620, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M620"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BB1F, "vmmul.t     M730, M630, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M630"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BC08, "vmmul.t     M200, M700, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M700"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BD08, "vmmul.t     M200, M710, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M710"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BE08, "vmmul.t     M200, M720, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M720"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BF08, "vmmul.t     M200, M730, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M730"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C01F, "vmmul.t     M730, E010, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E010"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C11F, "vmmul.t     M730, E011, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E011"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C21F, "vmmul.t     M730, E012, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E012"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C31F, "vmmul.t     M730, E013, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E013"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C41F, "vmmul.t     M730, E110, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E110"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C51F, "vmmul.t     M730, E111, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E111"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C61F, "vmmul.t     M730, E112, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E112"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C71F, "vmmul.t     M730, E113, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E113"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C81F, "vmmul.t     M730, E210, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E210"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C91F, "vmmul.t     M730, E211, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E211"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CA1F, "vmmul.t     M730, E212, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E212"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CB1F, "vmmul.t     M730, E213, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E213"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CC1F, "vmmul.t     M730, E310, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E310"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CD1F, "vmmul.t     M730, E311, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E311"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CE1F, "vmmul.t     M730, E312, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E312"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CF1F, "vmmul.t     M730, E313, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E313"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D01F, "vmmul.t     M730, E410, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E410"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D11F, "vmmul.t     M730, E411, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E411"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D21F, "vmmul.t     M730, E412, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E412"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D31F, "vmmul.t     M730, E413, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E413"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D41F, "vmmul.t     M730, E510, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E510"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D51F, "vmmul.t     M730, E511, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E511"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D61F, "vmmul.t     M730, E512, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E512"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D71F, "vmmul.t     M730, E513, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E513"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D81F, "vmmul.t     M730, E610, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E610"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D91F, "vmmul.t     M730, E611, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E611"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DA1F, "vmmul.t     M730, E612, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E612"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DB1F, "vmmul.t     M730, E613, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("E613"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DC08, "vmmul.t     M200, E710, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E710"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DD08, "vmmul.t     M200, E711, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E711"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DE08, "vmmul.t     M200, E712, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E712"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DF08, "vmmul.t     M200, E713, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("E713"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E01F, "vmmul.t     M730, M001, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M001"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E11F, "vmmul.t     M730, M011, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M011"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E21F, "vmmul.t     M730, M021, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M021"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E31F, "vmmul.t     M730, M031, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M031"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E41F, "vmmul.t     M730, M101, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M101"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E51F, "vmmul.t     M730, M111, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M111"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E61F, "vmmul.t     M730, M121, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M121"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E71F, "vmmul.t     M730, M131, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M131"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E81F, "vmmul.t     M730, M201, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M201"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E91F, "vmmul.t     M730, M211, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M211"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EA1F, "vmmul.t     M730, M221, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M221"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EB1F, "vmmul.t     M730, M231, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M231"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EC1F, "vmmul.t     M730, M301, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M301"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004ED1F, "vmmul.t     M730, M311, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M311"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EE1F, "vmmul.t     M730, M321, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M321"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EF1F, "vmmul.t     M730, M331, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M331"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F01F, "vmmul.t     M730, M401, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M401"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F11F, "vmmul.t     M730, M411, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M411"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F21F, "vmmul.t     M730, M421, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M421"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F31F, "vmmul.t     M730, M431, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M431"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F41F, "vmmul.t     M730, M501, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M501"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F51F, "vmmul.t     M730, M511, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M511"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F61F, "vmmul.t     M730, M521, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M521"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F71F, "vmmul.t     M730, M531, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M531"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F81F, "vmmul.t     M730, M601, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M601"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F91F, "vmmul.t     M730, M611, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M611"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FA1F, "vmmul.t     M730, M621, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M621"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FB1F, "vmmul.t     M730, M631, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M730"), Some("M631"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FC08, "vmmul.t     M200, M701, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M701"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FD08, "vmmul.t     M200, M711, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M711"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FE08, "vmmul.t     M200, M721, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M721"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FF08, "vmmul.t     M200, M731, M100", Opcode::r4000allegrex_vmmul_t, "vmmul.t", [Some("M200"), Some("M731"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004809F, "vmmul.q     M730, E000, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E000"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004819F, "vmmul.q     M730, E001, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E001"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004829F, "vmmul.q     M730, E002, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E002"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004839F, "vmmul.q     M730, E003, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E003"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004849F, "vmmul.q     M730, E100, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E100"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004859F, "vmmul.q     M730, E101, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E101"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004869F, "vmmul.q     M730, E102, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E102"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004879F, "vmmul.q     M730, E103, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E103"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004889F, "vmmul.q     M730, E200, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E200"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004899F, "vmmul.q     M730, E201, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E201"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048A9F, "vmmul.q     M730, E202, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E202"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048B9F, "vmmul.q     M730, E203, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E203"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048C9F, "vmmul.q     M730, E300, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E300"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048D9F, "vmmul.q     M730, E301, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E301"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048E9F, "vmmul.q     M730, E302, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E302"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0048F9F, "vmmul.q     M730, E303, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E303"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004909F, "vmmul.q     M730, E400, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E400"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004919F, "vmmul.q     M730, E401, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E401"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004929F, "vmmul.q     M730, E402, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E402"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004939F, "vmmul.q     M730, E403, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E403"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004949F, "vmmul.q     M730, E500, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E500"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004959F, "vmmul.q     M730, E501, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E501"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004969F, "vmmul.q     M730, E502, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E502"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004979F, "vmmul.q     M730, E503, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E503"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004989F, "vmmul.q     M730, E600, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E600"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004999F, "vmmul.q     M730, E601, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E601"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049A9F, "vmmul.q     M730, E602, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E602"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049B9F, "vmmul.q     M730, E603, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E603"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049C9F, "vmmul.q     M730, E700, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E700"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049D9F, "vmmul.q     M730, E701, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E701"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049E9F, "vmmul.q     M730, E702, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E702"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF0049F9F, "vmmul.q     M730, E703, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E703"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A09F, "vmmul.q     M730, M000, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M000"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A19F, "vmmul.q     M730, M010, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M010"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A29F, "vmmul.q     M730, M020, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M020"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A39F, "vmmul.q     M730, M030, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M030"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A49F, "vmmul.q     M730, M100, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M100"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A59F, "vmmul.q     M730, M110, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M110"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A69F, "vmmul.q     M730, M120, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M120"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A79F, "vmmul.q     M730, M130, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M130"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A89F, "vmmul.q     M730, M200, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M200"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004A99F, "vmmul.q     M730, M210, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M210"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AA9F, "vmmul.q     M730, M220, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M220"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AB9F, "vmmul.q     M730, M230, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M230"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AC9F, "vmmul.q     M730, M300, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M300"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AD9F, "vmmul.q     M730, M310, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M310"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AE9F, "vmmul.q     M730, M320, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M320"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004AF9F, "vmmul.q     M730, M330, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M330"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B09F, "vmmul.q     M730, M400, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M400"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B19F, "vmmul.q     M730, M410, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M410"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B29F, "vmmul.q     M730, M420, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M420"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B39F, "vmmul.q     M730, M430, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M430"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B49F, "vmmul.q     M730, M500, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M500"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B59F, "vmmul.q     M730, M510, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M510"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B69F, "vmmul.q     M730, M520, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M520"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B79F, "vmmul.q     M730, M530, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M530"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B89F, "vmmul.q     M730, M600, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M600"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004B99F, "vmmul.q     M730, M610, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M610"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BA9F, "vmmul.q     M730, M620, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M620"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BB9F, "vmmul.q     M730, M630, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M630"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BC9F, "vmmul.q     M730, M700, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M700"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BD9F, "vmmul.q     M730, M710, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M710"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BE9F, "vmmul.q     M730, M720, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M720"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004BF9F, "vmmul.q     M730, M730, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M730"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C09F, "vmmul.q     M730, E020, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E020"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C19F, "vmmul.q     M730, E021, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E021"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C29F, "vmmul.q     M730, E022, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E022"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C39F, "vmmul.q     M730, E023, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E023"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C49F, "vmmul.q     M730, E120, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E120"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C59F, "vmmul.q     M730, E121, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E121"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C69F, "vmmul.q     M730, E122, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E122"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C79F, "vmmul.q     M730, E123, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E123"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C89F, "vmmul.q     M730, E220, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E220"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004C99F, "vmmul.q     M730, E221, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E221"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CA9F, "vmmul.q     M730, E222, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E222"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CB9F, "vmmul.q     M730, E223, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E223"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CC9F, "vmmul.q     M730, E320, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E320"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CD9F, "vmmul.q     M730, E321, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E321"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CE9F, "vmmul.q     M730, E322, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E322"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004CF9F, "vmmul.q     M730, E323, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E323"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D09F, "vmmul.q     M730, E420, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E420"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D19F, "vmmul.q     M730, E421, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E421"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D29F, "vmmul.q     M730, E422, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E422"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D39F, "vmmul.q     M730, E423, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E423"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D49F, "vmmul.q     M730, E520, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E520"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D59F, "vmmul.q     M730, E521, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E521"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D69F, "vmmul.q     M730, E522, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E522"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D79F, "vmmul.q     M730, E523, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E523"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D89F, "vmmul.q     M730, E620, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E620"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004D99F, "vmmul.q     M730, E621, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E621"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DA9F, "vmmul.q     M730, E622, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E622"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DB9F, "vmmul.q     M730, E623, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E623"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DC9F, "vmmul.q     M730, E720, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E720"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DD9F, "vmmul.q     M730, E721, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E721"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DE9F, "vmmul.q     M730, E722, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E722"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004DF9F, "vmmul.q     M730, E723, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("E723"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E09F, "vmmul.q     M730, M002, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M002"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E19F, "vmmul.q     M730, M012, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M012"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E29F, "vmmul.q     M730, M022, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M022"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E39F, "vmmul.q     M730, M032, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M032"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E49F, "vmmul.q     M730, M102, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M102"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E59F, "vmmul.q     M730, M112, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M112"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E69F, "vmmul.q     M730, M122, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M122"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E79F, "vmmul.q     M730, M132, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M132"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E89F, "vmmul.q     M730, M202, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M202"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004E99F, "vmmul.q     M730, M212, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M212"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EA9F, "vmmul.q     M730, M222, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M222"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EB9F, "vmmul.q     M730, M232, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M232"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EC9F, "vmmul.q     M730, M302, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M302"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004ED9F, "vmmul.q     M730, M312, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M312"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EE9F, "vmmul.q     M730, M322, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M322"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004EF9F, "vmmul.q     M730, M332, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M332"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F09F, "vmmul.q     M730, M402, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M402"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F19F, "vmmul.q     M730, M412, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M412"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F29F, "vmmul.q     M730, M422, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M422"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F39F, "vmmul.q     M730, M432, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M432"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F49F, "vmmul.q     M730, M502, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M502"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F59F, "vmmul.q     M730, M512, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M512"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F69F, "vmmul.q     M730, M522, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M522"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F79F, "vmmul.q     M730, M532, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M532"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F89F, "vmmul.q     M730, M602, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M602"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004F99F, "vmmul.q     M730, M612, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M612"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FA9F, "vmmul.q     M730, M622, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M622"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FB9F, "vmmul.q     M730, M632, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M632"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FC9F, "vmmul.q     M730, M702, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M702"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FD9F, "vmmul.q     M730, M712, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M712"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FE9F, "vmmul.q     M730, M722, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M722"), Some("M100"), None, None]),
        TestEntry::new_r4000allegrex(0xF004FF9F, "vmmul.q     M730, M732, M100", Opcode::r4000allegrex_vmmul_q, "vmmul.q", [Some("M730"), Some("M732"), Some("M100"), None, None]),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vrot() {
    const ENTRIES: &[TestEntry] = &[
        /* vrot */
        TestEntry::new_r4000allegrex(
            0xF3A0009F,
            "vrot.p      C730, S000, [C,S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A1009F,
            "vrot.p      C730, S000, [S,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A2009F,
            "vrot.p      C730, S000, [S,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A3009F,
            "vrot.p      C730, S000, [S,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A4009F,
            "vrot.p      C730, S000, [C,S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A5009F,
            "vrot.p      C730, S000, [S,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A6009F,
            "vrot.p      C730, S000, [0,S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A7009F,
            "vrot.p      C730, S000, [0,S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A8009F,
            "vrot.p      C730, S000, [C,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A9009F,
            "vrot.p      C730, S000, [0,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AA009F,
            "vrot.p      C730, S000, [S,S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[S,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AB009F,
            "vrot.p      C730, S000, [0,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AC009F,
            "vrot.p      C730, S000, [C,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AD009F,
            "vrot.p      C730, S000, [0,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AE009F,
            "vrot.p      C730, S000, [0,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AF009F,
            "vrot.p      C730, S000, [S,S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[S,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B0009F,
            "vrot.p      C730, S000, [C,-S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B1009F,
            "vrot.p      C730, S000, [-S,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[-S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B2009F,
            "vrot.p      C730, S000, [-S,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[-S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B3009F,
            "vrot.p      C730, S000, [-S,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[-S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B4009F,
            "vrot.p      C730, S000, [C,-S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B5009F,
            "vrot.p      C730, S000, [-S,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[-S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B6009F,
            "vrot.p      C730, S000, [0,-S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B7009F,
            "vrot.p      C730, S000, [0,-S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B8009F,
            "vrot.p      C730, S000, [C,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B9009F,
            "vrot.p      C730, S000, [0,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BA009F,
            "vrot.p      C730, S000, [-S,-S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[-S,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BB009F,
            "vrot.p      C730, S000, [0,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BC009F,
            "vrot.p      C730, S000, [C,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BD009F,
            "vrot.p      C730, S000, [0,C]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BE009F,
            "vrot.p      C730, S000, [0,0]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BF009F,
            "vrot.p      C730, S000, [-S,-S]",
            Opcode::r4000allegrex_vrot_p,
            "vrot.p",
            [Some("C730"), Some("S000"), Some("[-S,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A0801F,
            "vrot.t      C730, S000, [C,S,S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,S,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A1801F,
            "vrot.t      C730, S000, [S,C,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[S,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A2801F,
            "vrot.t      C730, S000, [S,0,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[S,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A3801F,
            "vrot.t      C730, S000, [S,0,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[S,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A4801F,
            "vrot.t      C730, S000, [C,S,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A5801F,
            "vrot.t      C730, S000, [S,C,S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[S,C,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A6801F,
            "vrot.t      C730, S000, [0,S,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A7801F,
            "vrot.t      C730, S000, [0,S,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A8801F,
            "vrot.t      C730, S000, [C,0,S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,0,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A9801F,
            "vrot.t      C730, S000, [0,C,S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,C,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AA801F,
            "vrot.t      C730, S000, [S,S,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[S,S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AB801F,
            "vrot.t      C730, S000, [0,0,S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,0,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AC801F,
            "vrot.t      C730, S000, [C,0,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AD801F,
            "vrot.t      C730, S000, [0,C,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AE801F,
            "vrot.t      C730, S000, [0,0,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AF801F,
            "vrot.t      C730, S000, [S,S,S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[S,S,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B0801F,
            "vrot.t      C730, S000, [C,-S,-S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,-S,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B1801F,
            "vrot.t      C730, S000, [-S,C,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[-S,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B2801F,
            "vrot.t      C730, S000, [-S,0,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[-S,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B3801F,
            "vrot.t      C730, S000, [-S,0,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[-S,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B4801F,
            "vrot.t      C730, S000, [C,-S,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,-S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B5801F,
            "vrot.t      C730, S000, [-S,C,-S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[-S,C,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B6801F,
            "vrot.t      C730, S000, [0,-S,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,-S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B7801F,
            "vrot.t      C730, S000, [0,-S,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,-S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B8801F,
            "vrot.t      C730, S000, [C,0,-S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,0,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B9801F,
            "vrot.t      C730, S000, [0,C,-S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,C,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BA801F,
            "vrot.t      C730, S000, [-S,-S,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[-S,-S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BB801F,
            "vrot.t      C730, S000, [0,0,-S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,0,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BC801F,
            "vrot.t      C730, S000, [C,0,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[C,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BD801F,
            "vrot.t      C730, S000, [0,C,0]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BE801F,
            "vrot.t      C730, S000, [0,0,C]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[0,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BF801F,
            "vrot.t      C730, S000, [-S,-S,-S]",
            Opcode::r4000allegrex_vrot_t,
            "vrot.t",
            [Some("C730"), Some("S000"), Some("[-S,-S,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A0809F,
            "vrot.q      C730, S000, [C,S,S,S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,S,S,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A1809F,
            "vrot.q      C730, S000, [S,C,0,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[S,C,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A2809F,
            "vrot.q      C730, S000, [S,0,C,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[S,0,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A3809F,
            "vrot.q      C730, S000, [S,0,0,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[S,0,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A4809F,
            "vrot.q      C730, S000, [C,S,0,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,S,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A5809F,
            "vrot.q      C730, S000, [S,C,S,S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[S,C,S,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A6809F,
            "vrot.q      C730, S000, [0,S,C,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,S,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A7809F,
            "vrot.q      C730, S000, [0,S,0,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,S,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A8809F,
            "vrot.q      C730, S000, [C,0,S,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,0,S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3A9809F,
            "vrot.q      C730, S000, [0,C,S,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,C,S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AA809F,
            "vrot.q      C730, S000, [S,S,C,S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[S,S,C,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AB809F,
            "vrot.q      C730, S000, [0,0,S,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,0,S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AC809F,
            "vrot.q      C730, S000, [C,0,0,S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,0,0,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AD809F,
            "vrot.q      C730, S000, [0,C,0,S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,C,0,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AE809F,
            "vrot.q      C730, S000, [0,0,C,S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,0,C,S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3AF809F,
            "vrot.q      C730, S000, [S,S,S,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[S,S,S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B0809F,
            "vrot.q      C730, S000, [C,-S,-S,-S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,-S,-S,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B1809F,
            "vrot.q      C730, S000, [-S,C,0,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[-S,C,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B2809F,
            "vrot.q      C730, S000, [-S,0,C,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[-S,0,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B3809F,
            "vrot.q      C730, S000, [-S,0,0,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[-S,0,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B4809F,
            "vrot.q      C730, S000, [C,-S,0,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,-S,0,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B5809F,
            "vrot.q      C730, S000, [-S,C,-S,-S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[-S,C,-S,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B6809F,
            "vrot.q      C730, S000, [0,-S,C,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,-S,C,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B7809F,
            "vrot.q      C730, S000, [0,-S,0,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,-S,0,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B8809F,
            "vrot.q      C730, S000, [C,0,-S,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,0,-S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3B9809F,
            "vrot.q      C730, S000, [0,C,-S,0]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,C,-S,0]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BA809F,
            "vrot.q      C730, S000, [-S,-S,C,-S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[-S,-S,C,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BB809F,
            "vrot.q      C730, S000, [0,0,-S,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,0,-S,C]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BC809F,
            "vrot.q      C730, S000, [C,0,0,-S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[C,0,0,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BD809F,
            "vrot.q      C730, S000, [0,C,0,-S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,C,0,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BE809F,
            "vrot.q      C730, S000, [0,0,C,-S]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[0,0,C,-S]"), None, None],
        ),
        TestEntry::new_r4000allegrex(
            0xF3BF809F,
            "vrot.q      C730, S000, [-S,-S,-S,C]",
            Opcode::r4000allegrex_vrot_q,
            "vrot.q",
            [Some("C730"), Some("S000"), Some("[-S,-S,-S,C]"), None, None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vpfxs() {
    const ENTRIES: &[TestEntry] = &[
        /* vpfxs */
        TestEntry::new_r4000allegrex(
            0xDC000001,
            "vpfxs       Y, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("Y"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000002,
            "vpfxs       Z, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("Z"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000003,
            "vpfxs       W, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("W"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000100,
            "vpfxs       |X|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("|X|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000101,
            "vpfxs       |Y|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("|Y|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000102,
            "vpfxs       |Z|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("|Z|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000103,
            "vpfxs       |W|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("|W|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001000,
            "vpfxs       0, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("0"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001001,
            "vpfxs       1, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("1"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001002,
            "vpfxs       2, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001003,
            "vpfxs       1/2, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("1/2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001100,
            "vpfxs       3, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001101,
            "vpfxs       1/3, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("1/3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001102,
            "vpfxs       1/4, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("1/4"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC001103,
            "vpfxs       1/6, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("1/6"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010000,
            "vpfxs       -X, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-X"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010001,
            "vpfxs       -Y, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-Y"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010002,
            "vpfxs       -Z, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-Z"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010003,
            "vpfxs       -W, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-W"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010100,
            "vpfxs       -|X|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-|X|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010101,
            "vpfxs       -|Y|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-|Y|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010102,
            "vpfxs       -|Z|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-|Z|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC010103,
            "vpfxs       -|W|, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-|W|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011000,
            "vpfxs       -0, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-0"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011001,
            "vpfxs       -1, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-1"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011002,
            "vpfxs       -2, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011003,
            "vpfxs       -1/2, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-1/2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011100,
            "vpfxs       -3, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011101,
            "vpfxs       -1/3, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-1/3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011102,
            "vpfxs       -1/4, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-1/4"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC011103,
            "vpfxs       -1/6, X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("-1/6"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000004,
            "vpfxs       X, Y, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("Y"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000008,
            "vpfxs       X, Z, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("Z"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC00000C,
            "vpfxs       X, W, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("W"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000200,
            "vpfxs       X, |X|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("|X|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000204,
            "vpfxs       X, |Y|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("|Y|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000208,
            "vpfxs       X, |Z|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("|Z|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC00020C,
            "vpfxs       X, |W|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("|W|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC002000,
            "vpfxs       X, 0, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("0"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC002004,
            "vpfxs       X, 1, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("1"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC002008,
            "vpfxs       X, 2, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC00200C,
            "vpfxs       X, 1/2, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("1/2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC002200,
            "vpfxs       X, 3, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC002204,
            "vpfxs       X, 1/3, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("1/3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC002208,
            "vpfxs       X, 1/4, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("1/4"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC00220C,
            "vpfxs       X, 1/6, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("1/6"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC020000,
            "vpfxs       X, -X, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC020004,
            "vpfxs       X, -Y, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-Y"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC020008,
            "vpfxs       X, -Z, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-Z"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC02000C,
            "vpfxs       X, -W, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-W"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC020200,
            "vpfxs       X, -|X|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-|X|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC020204,
            "vpfxs       X, -|Y|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-|Y|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC020208,
            "vpfxs       X, -|Z|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-|Z|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC02020C,
            "vpfxs       X, -|W|, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-|W|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC022000,
            "vpfxs       X, -0, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-0"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC022004,
            "vpfxs       X, -1, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-1"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC022008,
            "vpfxs       X, -2, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC02200C,
            "vpfxs       X, -1/2, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-1/2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC022200,
            "vpfxs       X, -3, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC022204,
            "vpfxs       X, -1/3, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-1/3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC022208,
            "vpfxs       X, -1/4, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-1/4"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC02220C,
            "vpfxs       X, -1/6, X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("-1/6"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000010,
            "vpfxs       X, X, Y, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("Y"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000020,
            "vpfxs       X, X, Z, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("Z"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000030,
            "vpfxs       X, X, W, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("W"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000400,
            "vpfxs       X, X, |X|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("|X|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000410,
            "vpfxs       X, X, |Y|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("|Y|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000420,
            "vpfxs       X, X, |Z|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("|Z|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000430,
            "vpfxs       X, X, |W|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("|W|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004000,
            "vpfxs       X, X, 0, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("0"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004010,
            "vpfxs       X, X, 1, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("1"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004020,
            "vpfxs       X, X, 2, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004030,
            "vpfxs       X, X, 1/2, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("1/2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004400,
            "vpfxs       X, X, 3, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004410,
            "vpfxs       X, X, 1/3, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("1/3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004420,
            "vpfxs       X, X, 1/4, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("1/4"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC004430,
            "vpfxs       X, X, 1/6, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("1/6"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040000,
            "vpfxs       X, X, -X, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040010,
            "vpfxs       X, X, -Y, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-Y"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040020,
            "vpfxs       X, X, -Z, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-Z"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040030,
            "vpfxs       X, X, -W, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-W"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040400,
            "vpfxs       X, X, -|X|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-|X|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040410,
            "vpfxs       X, X, -|Y|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-|Y|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040420,
            "vpfxs       X, X, -|Z|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-|Z|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC040430,
            "vpfxs       X, X, -|W|, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-|W|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044000,
            "vpfxs       X, X, -0, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-0"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044010,
            "vpfxs       X, X, -1, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-1"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044020,
            "vpfxs       X, X, -2, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044030,
            "vpfxs       X, X, -1/2, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-1/2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044400,
            "vpfxs       X, X, -3, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044410,
            "vpfxs       X, X, -1/3, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-1/3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044420,
            "vpfxs       X, X, -1/4, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-1/4"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC044430,
            "vpfxs       X, X, -1/6, X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("-1/6"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000040,
            "vpfxs       X, X, X, Y",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("Y"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000080,
            "vpfxs       X, X, X, Z",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("Z"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0000C0,
            "vpfxs       X, X, X, W",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("W"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000800,
            "vpfxs       X, X, X, |X|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("|X|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000840,
            "vpfxs       X, X, X, |Y|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("|Y|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC000880,
            "vpfxs       X, X, X, |Z|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("|Z|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0008C0,
            "vpfxs       X, X, X, |W|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("|W|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC008000,
            "vpfxs       X, X, X, 0",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("0"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC008040,
            "vpfxs       X, X, X, 1",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("1"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC008080,
            "vpfxs       X, X, X, 2",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0080C0,
            "vpfxs       X, X, X, 1/2",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("1/2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC008800,
            "vpfxs       X, X, X, 3",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC008840,
            "vpfxs       X, X, X, 1/3",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("1/3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC008880,
            "vpfxs       X, X, X, 1/4",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("1/4"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0088C0,
            "vpfxs       X, X, X, 1/6",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("1/6"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC080000,
            "vpfxs       X, X, X, -X",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC080040,
            "vpfxs       X, X, X, -Y",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-Y"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC080080,
            "vpfxs       X, X, X, -Z",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-Z"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0800C0,
            "vpfxs       X, X, X, -W",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-W"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC080800,
            "vpfxs       X, X, X, -|X|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-|X|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC080840,
            "vpfxs       X, X, X, -|Y|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-|Y|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC080880,
            "vpfxs       X, X, X, -|Z|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-|Z|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0808C0,
            "vpfxs       X, X, X, -|W|",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-|W|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC088000,
            "vpfxs       X, X, X, -0",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-0"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC088040,
            "vpfxs       X, X, X, -1",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-1"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC088080,
            "vpfxs       X, X, X, -2",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0880C0,
            "vpfxs       X, X, X, -1/2",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-1/2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC088800,
            "vpfxs       X, X, X, -3",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC088840,
            "vpfxs       X, X, X, -1/3",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-1/3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC088880,
            "vpfxs       X, X, X, -1/4",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-1/4"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDC0888C0,
            "vpfxs       X, X, X, -1/6",
            Opcode::r4000allegrex_vpfxs,
            "vpfxs",
            [Some("X"), Some("X"), Some("X"), Some("-1/6"), None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vpfxt() {
    const ENTRIES: &[TestEntry] = &[
        /* vpfxt */
        TestEntry::new_r4000allegrex(
            0xDD000001,
            "vpfxt       Y, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("Y"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000002,
            "vpfxt       Z, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("Z"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000003,
            "vpfxt       W, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("W"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000100,
            "vpfxt       |X|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("|X|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000101,
            "vpfxt       |Y|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("|Y|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000102,
            "vpfxt       |Z|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("|Z|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000103,
            "vpfxt       |W|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("|W|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001000,
            "vpfxt       0, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("0"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001001,
            "vpfxt       1, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("1"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001002,
            "vpfxt       2, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001003,
            "vpfxt       1/2, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("1/2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001100,
            "vpfxt       3, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001101,
            "vpfxt       1/3, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("1/3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001102,
            "vpfxt       1/4, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("1/4"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD001103,
            "vpfxt       1/6, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("1/6"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010000,
            "vpfxt       -X, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-X"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010001,
            "vpfxt       -Y, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-Y"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010002,
            "vpfxt       -Z, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-Z"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010003,
            "vpfxt       -W, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-W"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010100,
            "vpfxt       -|X|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-|X|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010101,
            "vpfxt       -|Y|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-|Y|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010102,
            "vpfxt       -|Z|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-|Z|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD010103,
            "vpfxt       -|W|, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-|W|"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011000,
            "vpfxt       -0, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-0"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011001,
            "vpfxt       -1, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-1"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011002,
            "vpfxt       -2, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011003,
            "vpfxt       -1/2, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-1/2"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011100,
            "vpfxt       -3, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011101,
            "vpfxt       -1/3, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-1/3"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011102,
            "vpfxt       -1/4, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-1/4"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD011103,
            "vpfxt       -1/6, X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("-1/6"), Some("X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000004,
            "vpfxt       X, Y, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("Y"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000008,
            "vpfxt       X, Z, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("Z"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD00000C,
            "vpfxt       X, W, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("W"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000200,
            "vpfxt       X, |X|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("|X|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000204,
            "vpfxt       X, |Y|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("|Y|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000208,
            "vpfxt       X, |Z|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("|Z|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD00020C,
            "vpfxt       X, |W|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("|W|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD002000,
            "vpfxt       X, 0, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("0"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD002004,
            "vpfxt       X, 1, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("1"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD002008,
            "vpfxt       X, 2, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD00200C,
            "vpfxt       X, 1/2, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("1/2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD002200,
            "vpfxt       X, 3, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD002204,
            "vpfxt       X, 1/3, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("1/3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD002208,
            "vpfxt       X, 1/4, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("1/4"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD00220C,
            "vpfxt       X, 1/6, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("1/6"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD020000,
            "vpfxt       X, -X, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-X"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD020004,
            "vpfxt       X, -Y, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-Y"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD020008,
            "vpfxt       X, -Z, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-Z"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD02000C,
            "vpfxt       X, -W, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-W"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD020200,
            "vpfxt       X, -|X|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-|X|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD020204,
            "vpfxt       X, -|Y|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-|Y|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD020208,
            "vpfxt       X, -|Z|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-|Z|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD02020C,
            "vpfxt       X, -|W|, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-|W|"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD022000,
            "vpfxt       X, -0, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-0"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD022004,
            "vpfxt       X, -1, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-1"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD022008,
            "vpfxt       X, -2, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD02200C,
            "vpfxt       X, -1/2, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-1/2"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD022200,
            "vpfxt       X, -3, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD022204,
            "vpfxt       X, -1/3, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-1/3"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD022208,
            "vpfxt       X, -1/4, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-1/4"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD02220C,
            "vpfxt       X, -1/6, X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("-1/6"), Some("X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000010,
            "vpfxt       X, X, Y, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("Y"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000020,
            "vpfxt       X, X, Z, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("Z"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000030,
            "vpfxt       X, X, W, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("W"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000400,
            "vpfxt       X, X, |X|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("|X|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000410,
            "vpfxt       X, X, |Y|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("|Y|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000420,
            "vpfxt       X, X, |Z|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("|Z|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000430,
            "vpfxt       X, X, |W|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("|W|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004000,
            "vpfxt       X, X, 0, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("0"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004010,
            "vpfxt       X, X, 1, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("1"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004020,
            "vpfxt       X, X, 2, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004030,
            "vpfxt       X, X, 1/2, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("1/2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004400,
            "vpfxt       X, X, 3, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004410,
            "vpfxt       X, X, 1/3, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("1/3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004420,
            "vpfxt       X, X, 1/4, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("1/4"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD004430,
            "vpfxt       X, X, 1/6, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("1/6"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040000,
            "vpfxt       X, X, -X, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-X"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040010,
            "vpfxt       X, X, -Y, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-Y"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040020,
            "vpfxt       X, X, -Z, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-Z"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040030,
            "vpfxt       X, X, -W, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-W"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040400,
            "vpfxt       X, X, -|X|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-|X|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040410,
            "vpfxt       X, X, -|Y|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-|Y|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040420,
            "vpfxt       X, X, -|Z|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-|Z|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD040430,
            "vpfxt       X, X, -|W|, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-|W|"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044000,
            "vpfxt       X, X, -0, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-0"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044010,
            "vpfxt       X, X, -1, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-1"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044020,
            "vpfxt       X, X, -2, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044030,
            "vpfxt       X, X, -1/2, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-1/2"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044400,
            "vpfxt       X, X, -3, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044410,
            "vpfxt       X, X, -1/3, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-1/3"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044420,
            "vpfxt       X, X, -1/4, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-1/4"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD044430,
            "vpfxt       X, X, -1/6, X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("-1/6"), Some("X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000040,
            "vpfxt       X, X, X, Y",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("Y"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000080,
            "vpfxt       X, X, X, Z",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("Z"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0000C0,
            "vpfxt       X, X, X, W",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("W"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000800,
            "vpfxt       X, X, X, |X|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("|X|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000840,
            "vpfxt       X, X, X, |Y|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("|Y|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD000880,
            "vpfxt       X, X, X, |Z|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("|Z|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0008C0,
            "vpfxt       X, X, X, |W|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("|W|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD008000,
            "vpfxt       X, X, X, 0",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("0"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD008040,
            "vpfxt       X, X, X, 1",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("1"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD008080,
            "vpfxt       X, X, X, 2",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0080C0,
            "vpfxt       X, X, X, 1/2",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("1/2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD008800,
            "vpfxt       X, X, X, 3",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD008840,
            "vpfxt       X, X, X, 1/3",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("1/3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD008880,
            "vpfxt       X, X, X, 1/4",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("1/4"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0088C0,
            "vpfxt       X, X, X, 1/6",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("1/6"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD080000,
            "vpfxt       X, X, X, -X",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-X"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD080040,
            "vpfxt       X, X, X, -Y",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-Y"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD080080,
            "vpfxt       X, X, X, -Z",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-Z"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0800C0,
            "vpfxt       X, X, X, -W",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-W"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD080800,
            "vpfxt       X, X, X, -|X|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-|X|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD080840,
            "vpfxt       X, X, X, -|Y|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-|Y|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD080880,
            "vpfxt       X, X, X, -|Z|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-|Z|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0808C0,
            "vpfxt       X, X, X, -|W|",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-|W|"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD088000,
            "vpfxt       X, X, X, -0",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-0"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD088040,
            "vpfxt       X, X, X, -1",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-1"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD088080,
            "vpfxt       X, X, X, -2",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0880C0,
            "vpfxt       X, X, X, -1/2",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-1/2"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD088800,
            "vpfxt       X, X, X, -3",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD088840,
            "vpfxt       X, X, X, -1/3",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-1/3"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD088880,
            "vpfxt       X, X, X, -1/4",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-1/4"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDD0888C0,
            "vpfxt       X, X, X, -1/6",
            Opcode::r4000allegrex_vpfxt,
            "vpfxt",
            [Some("X"), Some("X"), Some("X"), Some("-1/6"), None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}

#[cfg(feature = "R4000ALLEGREX")]
#[test]
fn check_r4000allegrex_vfpu_instructions_vpfxd() {
    const ENTRIES: &[TestEntry] = &[
        /* vpfxd */
        TestEntry::new_r4000allegrex(
            0xDE000001,
            "vpfxd       0, , , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some("0"), Some(""), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000003,
            "vpfxd       1, , , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some("1"), Some(""), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000100,
            "vpfxd       M, , , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some("M"), Some(""), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000004,
            "vpfxd       , 0, , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some("0"), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE00000C,
            "vpfxd       , 1, , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some("1"), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000200,
            "vpfxd       , M, , ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some("M"), Some(""), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000010,
            "vpfxd       , , 0, ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some("0"), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000030,
            "vpfxd       , , 1, ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some("1"), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000400,
            "vpfxd       , , M, ",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some("M"), Some(""), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000040,
            "vpfxd       , , , 0",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some(""), Some("0"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE0000C0,
            "vpfxd       , , , 1",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some(""), Some("1"), None],
        ),
        TestEntry::new_r4000allegrex(
            0xDE000800,
            "vpfxd       , , , M",
            Opcode::r4000allegrex_vpfxd,
            "vpfxd",
            [Some(""), Some(""), Some(""), Some("M"), None],
        ),
    ];

    assert_eq!(check_test_entries(ENTRIES, true), (0, 0));
}
